
/* //////////////////////////////////////////////////// */
/* This file is a part of the BSTools procedure package */
/* written by Przemyslaw Kiciak.                        */
/* //////////////////////////////////////////////////// */

#include <assert.h>
#include <stdlib.h>   
#include <stdio.h>
#include <math.h>  
#include <malloc.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <GL/gl.h>  
#include <GL/glu.h> 
#include <GL/glx.h>

#include "pkvaria.h" 
#include "pknum.h"
#include "pkgeom.h"
#include "camera.h"
#include "multibs.h"
#include "bsmesh.h"
#include "g2blendingd.h"
#include "egholed.h"
#include "bsfile.h"
#include "xgedit.h"
#include "xgledit.h"

#include "editor.h"
#include "editor_bsp.h"

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _UnpackTheBoundary ( int ntrd, mbs_polycurved *bound,
                                    BSP_TrPolycurve *ebound )
{
  int  i, j, np;
  byte *mkcp;

  memset ( ebound, 0, ntrd*sizeof(BSP_TrPolycurve) );
  for ( i = 0; i < ntrd; i++ ) {
    ebound[i].elem = bound[i];
    if ( bound[i].knots )            /* a B-spline curve */
      np = bound[i].lastknot-bound[i].degree;
    else if ( bound[i].degree > 1 )  /* a Bezier curve */
      np = bound[i].degree+1;
    else                             /* a polyline */
      np = bound[i].lastknot+1;
    ebound[i].mkcp = mkcp = malloc ( np*sizeof(byte) );
    if ( !ebound[i].mkcp )
      goto failure;
    memset ( mkcp, MASK_CP_MOVEABLE, np*sizeof(byte) );
  }
        /* make the lists of elements making closed polylines */
  j = 0;
  for ( i = 0; i < ntrd; i++ ) {
    if ( ebound[i].elem.closing || i == ntrd-1 ) {
      ebound[i].next = j;
      ebound[j].prev = i;
      j = i+1;
    }
    else {
      ebound[i].next = i+1;
      ebound[i+1].prev = i;
    }
  }
  return true;

failure:
  for ( ; i >= 0; i-- )
    free ( ebound[i].mkcp );
  return false;
} /*_UnpackTheBoundary*/

static void _PackTheBoundary ( int ntrd, BSP_TrPolycurve *ebound,
                               mbs_polycurved *bound )
{
  int i, j, k;

  for ( i = 0; i < ntrd; i++ )
    ebound[i].tag = 0;
  k = 0;
  for ( i = 0; i < ntrd; i++ )
    for ( j = i; !ebound[j].tag; j = ebound[j].next ) {
      ebound[j].tag = 1;
      bound[k] = ebound[j].elem;
      bound[k++].closing = (ebound[j].next == i);
    }
} /*_PackTheBoundary*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean GeomObjectBSplinePatchCopyTrimmedDomain ( GO_BSplinePatch *obj,
                                                  GO_BSplinePatch *copy )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  mbs_polycurved    *bound;
  int               i, nelem, dim, deg, lkn, ncp;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH ||
       copy->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trpd )
    return false;
  nelem = obj->trpd->nelem;
  trpd = copy->trpd = malloc ( sizeof(BSP_TrimmedDomain) );
  if ( !trpd )
    return false;
  memset ( trpd, 0, sizeof(BSP_TrimmedDomain) );
  ebound = trpd->ebound = malloc ( nelem*sizeof(BSP_TrPolycurve) );
  bound = trpd->bound = malloc ( nelem*sizeof(mbs_polycurved) );
  if ( !bound || !ebound ) {
    if ( bound )  free ( bound );
    if ( ebound ) free ( ebound );
    free ( trpd );
    return false;
  }
  memset ( bound, 0, nelem*sizeof(mbs_polycurved) );
  memset ( ebound, 0, nelem*sizeof(BSP_TrPolycurve) );
  trpd->nelem = nelem;
  for ( i = 0; i < nelem; i++ ) {
    bound[i].ident    = -1;
    bound[i].closing  = obj->trpd->bound[i].closing;
    bound[i].cpdimen  = dim = obj->trpd->bound[i].cpdimen;
    bound[i].degree   = deg = obj->trpd->bound[i].degree;
    bound[i].lastknot = lkn = obj->trpd->bound[i].lastknot;
    if ( obj->trpd->bound[i].knots ) {  /* a B-spline curve */
      bound[i].knots = malloc ( (lkn+1)*sizeof(double) );
      if ( bound[i].knots ) {
        memcpy ( bound[i].knots, obj->trpd->bound[i].knots,
                 (lkn+1)*sizeof(double) );
        ncp = lkn-deg;
      }
      else
        goto failure;
    }
    else if ( deg > 1 )  /* a Bezier curve */
      ncp = deg+1;
    else                 /* a polyline */
      ncp = lkn+1;
    bound[i].points = malloc ( ncp*dim*sizeof(double) );
    if ( bound[i].points )
      memcpy ( bound[i].points, obj->trpd->bound[i].points,
               ncp*dim*sizeof(double) );
    else
      goto failure;
  }
  _UnpackTheBoundary ( nelem, bound, ebound );
  return true;

failure:
  GeomObjectBSplinePatchDeleteTrimmedDomain ( copy );
  return false;
} /*GeomObjectBSplinePatchCopyTrimmedDomain*/

void GeomObjectBSplinePatchDeleteTrimmedDomain ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  int               i, nelem;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  nelem = trpd->nelem;
  for ( i = 0; i < nelem; i++ ) {
    if ( ebound[i].elem.points ) free ( ebound[i].elem.points );
    if ( ebound[i].elem.knots )  free ( ebound[i].elem.knots );
    if ( ebound[i].savedpoints ) free ( ebound[i].savedpoints );
    if ( ebound[i].mkcp )        free ( ebound[i].mkcp );
  }
  free ( trpd->ebound );
  free ( trpd->bound );
  free ( trpd );
  obj->trpd = NULL;
  obj->trimmed = obj->trimmed_on = false;
} /*GeomObjectBSplinePatchDeleteTrimmedDomain*/

void GeomObjectBSplinePatchEnterTrimmedDomain ( GO_BSplinePatch *obj,
                                                int nelem, mbs_polycurved *elem )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  int               i;


  trpd = NULL;
  ebound = NULL;
  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    goto failure;
  if ( obj->trpd )
    GeomObjectBSplinePatchDeleteTrimmedDomain ( obj );
  trpd = malloc ( sizeof(BSP_TrimmedDomain) );
  ebound = malloc ( nelem*sizeof(BSP_TrPolycurve) );
  if ( !trpd || !ebound )
    goto failure;
  memset ( trpd, 0, sizeof(BSP_TrimmedDomain) );
  memset ( ebound, 0, nelem*sizeof(BSP_TrPolycurve) );
  trpd->nelem = nelem;
  trpd->bound = elem;
  trpd->ebound = ebound;
  obj->trpd = trpd;
  if ( !_UnpackTheBoundary ( nelem, elem, ebound ) )
    goto failure;
  obj->trimmed = true;
  obj->me.dlistmask = 0;
  return;

failure:
        /* deallocate all knot sequences and control points; */
        /* the elem array will be deallocated by the caller. */
  for ( i = 0; i < nelem; i++ ) {
    if ( elem[i].knots )  free ( elem[i].knots );
    if ( elem[i].points ) free ( elem[i].points );
  }
  free ( elem );
  if ( ebound ) free ( ebound );
  if ( trpd )   free ( trpd );
  obj->trpd = NULL;
  obj->trimmed = false;
} /*GeomObjectBSplinePatchEnterTrimmedDomain*/

boolean GeomObjectBSplinePatchMakeItTrimmed ( GO_BSplinePatch *obj, boolean tr )
{
  mbs_polycurved *elem;
  point2d        *p;
  int            degu, degv, lknu, lknv;

  if ( tr ) {
    elem = malloc ( sizeof(mbs_polycurved) );
    p    = malloc ( 5*sizeof(point2d) );
    if ( !elem || !p ) {
      if ( elem ) free ( elem );
      if ( p )    free ( p );
      goto out;
    }
    elem->ident = 0;
    elem->closing = 2;
    elem->cpdimen = 2;
    elem->degree = 1;
    elem->lastknot = 4;
    elem->knots = NULL;
    elem->points = &p[0].x;
    degu = obj->degree_u;
    lknu = obj->lastknot_u;
    degv = obj->degree_v;
    lknv = obj->lastknot_v;
    p[0].x = p[3].x = p[4].x = obj->knots_u[degu];
    p[1].x = p[2].x = obj->knots_u[lknu-degu];
    p[0].y = p[1].y = p[4].y = obj->knots_v[degv];
    p[2].y = p[3].y = obj->knots_v[lknv-degv];
    GeomObjectBSplinePatchEnterTrimmedDomain ( obj, 1, elem );
  }
  else if ( obj->trimmed )
    GeomObjectBSplinePatchDeleteTrimmedDomain ( obj );
out:
  obj->me.dlistmask = 0;
  return obj->trimmed;
} /*GeomObjectBSplinePatchMakeItTrimmed*/

/* ////////////////////////////////////////////////////////////////////////// */
static void _GetClipLines ( CameraRecd *CPos, vector3d cliplines[4] )
{
  SetVector3d ( &cliplines[0], 1.0, 0.0,  (double)(-CPos->xmin) );
  SetVector3d ( &cliplines[1], 0.0, 1.0,  (double)(-CPos->ymin) );
  SetVector3d ( &cliplines[2], -1.0, 0.0, (double)(CPos->xmin+CPos->width) );
  SetVector3d ( &cliplines[3], 0.0, -1.0, (double)(CPos->ymin+CPos->height) );
} /*_GetClipLines*/

static void _DrawControlPolygon ( CameraRecd *CPos, byte dim,
                                  int npts, const double *cp, boolean current )
{
  void     *sp;
  vector3d cliplines[4];
  int      i;
  point2d  p[2], q;

  sp = pkv_GetScratchMemTop ();
  if ( current )
    xgeSetForeground ( xgec_Green2 );
  else
    xgeSetForeground ( xgec_Green4 );
  _GetClipLines ( CPos, cliplines );
  switch ( dim ) {
case 2:
    CameraProjectPoint2d ( CPos, (point2d*)&cp[0], &p[0] );
    for ( i = 1; i < npts; i++ ) {
      CameraProjectPoint2d ( CPos, (point2d*)&cp[2*i], &p[1] );
      mbs_ClipBC2d ( 4, cliplines, 1, p, xge_DrawBC2d );
      p[0] = p[1];
    }
    break;
case 3:
    SetPoint2d ( &q, cp[0]/cp[2], cp[1]/cp[2] );
    CameraProjectPoint2d ( CPos, &q, &p[0] );
    for ( i = 1; i < npts; i++ ) {
      SetPoint2d ( &q, cp[3*i]/cp[3*i+2], cp[3*i+1]/cp[3*i+2] );
      CameraProjectPoint2d ( CPos, &q, &p[1] );
      mbs_ClipBC2d ( 4, cliplines, 1, p, xge_DrawBC2d );
      p[0] = p[1];
    }
    break;
default:
    break;
  }
  pkv_SetScratchMemTop ( sp );
} /*_DrawControlPolygon*/

static void _DrawControlPoints ( CameraRecd *CPos, byte dim,
                                 int npts, const double *cp,
                                 const byte *mkcp, byte mask )
{
  int     i;
  point2d q, r;

  switch ( dim ) {
case 2:
    for ( i = 0; i < npts; i++ )
      if ( CameraClipPoint2d ( CPos, (point2d*)&cp[2*i], &q ) ) {
        xgeSetForeground ( (mkcp[i] & mask) ? xgec_OrangeRed : xgec_Yellow );
        xgeFillRectangle ( 3, 3, (int)q.x-1, (int)q.y-1 );
      }
    break;
case 3:
    for ( i = 0; i < npts; i++ ) {
      Point3to2d ( (point3d*)&cp[3*i], &r );
      if ( CameraClipPoint2d ( CPos, &r, &q ) ) {
        xgeSetForeground ( (mkcp[i] & mask) ? xgec_OrangeRed : xgec_Yellow );
        xgeFillRectangle ( 3, 3, (int)q.x, (int)q.y );
      }
    }
    break;
default:
    break;
  }
} /*_DrawControlPoints*/

static void _DrawTrdBSCurve ( CameraRecd *CPos, int degree, int lastknot,
                              double *knots, int cpdim, double *cp )
{
  void *sp;
  double   *pcp, *bcp, w;
  vector3d cliplines[4];
  int      i, ncp, kpcs;
  point2d  q;

  sp = pkv_GetScratchMemTop ();
  kpcs = mbs_NumKnotIntervalsd ( degree, lastknot, knots );
  if ( kpcs < 1 )
    goto way_out;
  ncp = lastknot-degree;
  _GetClipLines ( CPos, cliplines );
  switch ( cpdim ) {
case 2:
    pcp = pkv_GetScratchMem ( ncp*sizeof(point2d) );
    bcp = pkv_GetScratchMem ( kpcs*(degree+1)*sizeof(point2d) );
    if ( !pcp || !bcp )
      goto way_out;
    for ( i = 0; i < ncp; i++ )
      CameraProjectPoint2d ( CPos, (point2d*)&cp[2*i], (point2d*)&pcp[2*i] );
    mbs_BSToBezC2d ( degree, lastknot, knots, pcp, &kpcs, NULL, NULL, bcp );
    for ( i = 0; i < kpcs; i++ )
      mbs_ClipBC2d ( 4, cliplines, degree, (point2d*)&bcp[i*2*(degree+1)],
                     xge_DrawBC2d );
    break;
case 3:
    pcp = pkv_GetScratchMem ( ncp*sizeof(point3d) );
    bcp = pkv_GetScratchMem ( kpcs*(degree+1)*sizeof(point3d) );
    if ( !pcp || !bcp )
      goto way_out;
    for ( i = 0; i < ncp; i++ ) {
      w = cp[3*i+2];
      SetPoint2d ( &q, cp[3*i]/w, cp[3*i+1]/w );
      CameraProjectPoint2d ( CPos, &q, &q );
      SetPoint3d ( (point3d*)&pcp[3*i], q.x*w, q.y*w, w );
    }
    mbs_BSToBezC3d ( degree, lastknot, knots, pcp, &kpcs, NULL, NULL, bcp );
    for ( i = 0; i < kpcs; i++ )
      mbs_ClipBC2Rd ( 4, cliplines, degree, (point3d*)&bcp[i*3*(degree+1)],
                      xge_DrawBC2Rd );
    break;
default:
    break;
  }

way_out:
  pkv_SetScratchMemTop ( sp );
} /*_DrawTrdBSCurve*/

static void _DrawTrdBezCurve ( CameraRecd *CPos, int degree,
                               int cpdim, double *cp )
{
  void     *sp;
  double   *pcp, w;
  vector3d cliplines[4];
  int      i;
  point2d  q;

  sp = pkv_GetScratchMemTop ();
  _GetClipLines ( CPos, cliplines );

  switch ( cpdim ) {
case 2:
    pcp = pkv_GetScratchMem ( (degree+1)*sizeof(point2d) );
    if ( !pcp )
      goto way_out;
    for ( i = 0; i <= degree; i++ )
      CameraProjectPoint2d ( CPos, (point2d*)&cp[2*i], (point2d*)&pcp[2*i] );
    mbs_ClipBC2d ( 4, cliplines, degree, (point2d*)pcp, xge_DrawBC2d );
    break;
case 3:
    pcp = pkv_GetScratchMem ( (degree+1)*sizeof(point3d) );
    if ( !pcp )
      goto way_out;
    for ( i = 0; i <= degree; i++ ) {
      w = cp[3*i+2];
      SetPoint2d ( &q, cp[3*i]/w, cp[3*i+1]/w );
      CameraProjectPoint2d ( CPos, &q, &q );
      SetPoint3d ( (point3d*)&pcp[3*i], q.x*w, q.y*w, w );
    }
    mbs_ClipBC2Rd ( 4, cliplines, degree, (point3d*)pcp, xge_DrawBC2Rd );
    break;
default:
    break;
  }

way_out:
  pkv_SetScratchMemTop ( sp );
} /*_DrawTrdBezCurve*/

static void _DrawTrdPolyline ( CameraRecd *CPos, int npoints,
                               int cpdim, double *cp )
{
  void    *sp;
  vector3d cliplines[4];
  point2d p, *q;
  int     i;

  sp = pkv_GetScratchMemTop ();
  q = pkv_GetScratchMem ( npoints*sizeof(point2d) );
  if ( !q )
    goto way_out;
  _GetClipLines ( CPos, cliplines );
  switch ( cpdim ) {
case 2:
    for ( i = 0; i < npoints; i++ )
      CameraProjectPoint2d ( CPos, (point2d*)&cp[2*i], &q[i] );
    goto draw_it;
case 3:
    for ( i = 0; i < npoints; i++ ) {
      Point3to2d ( (point3d*)&cp[3*i], &p );
      CameraProjectPoint2d ( CPos, &p, &q[i] );
    }
draw_it:
    for ( i = 0; i < npoints-1; i++ )
      mbs_ClipBC2d ( 4, cliplines, 1, &q[i], xge_DrawBC2d );
    break;
default:
    break;
  }
way_out:
  pkv_SetScratchMemTop ( sp );
} /*_DrawTrdPolyline*/

static void _DrawDivisionPoint ( CameraRecd *CPos, point2d *dpt, boolean last )
{
  point2d q;
  int     x, y;

  if ( CameraClipPoint2d ( CPos, dpt, &q ) ) {
    x = (int)(q.x+0.5);
    y = (int)(q.y+0.5);
    if ( last ) xgeSetForeground ( xgec_GreenYellow );
    else        xgeSetForeground ( xgec_Gold );
    xgeDrawLine ( x-4, y, x+4, y );
    xgeDrawLine ( x, y-4, x, y+4 );
    xgeDrawLine ( x-3, y-3, x+3, y+3 );
    xgeDrawLine ( x-3, y+3, x+3, y-3 );
  }
} /*_DrawDivisionPoint*/

static void _DrawDotsInArea ( CameraRecd *CPos,
                              double u0, double u1, double v0, double v1,
                              BSP_TrimmedDomain *trpd )
{
#define DDIST 10
  int             nelem;
  mbs_polycurved  *bound;
  BSP_TrPolycurve *ebound;
  Box2i           box;
  point2d         p0, p1, q0, q1, p, q;
  int             i, j;
  double          *workspace;

  nelem = trpd->nelem;
  bound = trpd->bound;
  ebound = trpd->ebound;
  _PackTheBoundary ( nelem, ebound, bound );
  workspace = pkv_GetScratchMem ( 1048576 );
  _pkv_InitPixelBuffer ();
  if ( workspace && _pkv_pixbuf ) {
    _pkv_OutputPixels = xge_OutPixels;
    box.x0 = CPos->xmin;  box.x1 = CPos->xmin+CPos->width;
    box.y0 = CPos->ymin;  box.y1 = CPos->ymin+CPos->height;
    SetPoint2d ( &p0, u0, v0 );
    CameraProjectPoint2d ( CPos, &p0, &q0 );
    if ( (int)q0.x > box.x0 ) box.x0 = (int)q0.x;
    if ( (int)q0.y < box.y1 ) box.y1 = (int)q0.y;
    SetPoint2d ( &p1, u1, v1 );
    CameraProjectPoint2d ( CPos, &p1, &q1 );
    if ( (int)q1.x < box.x1 ) box.x1 = (int)q1.x;
    if ( (int)q1.y > box.y0 ) box.y0 = (int)q1.y;
    box.x0 -= (box.x0-(int)q0.x+DDIST-1) % DDIST + DDIST/2;
    box.y1 -= (box.y1-(int)q1.y+DDIST-1) % DDIST - DDIST/2;
    xgeSetForeground ( xgec_Blue1 );
    for ( i = box.y1; i > box.y0; i -= DDIST/2 ) {
      for ( j = box.x0; j < box.x1; j += DDIST ) {
        SetPoint2d ( &q, (double)j, (double)i );
        CameraUnProjectPoint2d ( CPos, &q, &p );
        if ( mbs_PointPolycurveIndexd ( nelem, bound, p.x, p.y, workspace ) == 1 )
          PKV_SETPIXEL ( j, i );
      }
      i -= DDIST/2;
      for ( j = box.x0+DDIST/2; j < box.x1; j += DDIST ) {
        SetPoint2d ( &q, (double)j, (double)i );
        CameraUnProjectPoint2d ( CPos, &q, &p );
        if ( mbs_PointPolycurveIndexd ( nelem, bound, p.x, p.y, workspace ) == 1 )
          PKV_SETPIXEL ( j, i );
      }
    }
    PKV_FLUSH
  }
  _pkv_DestroyPixelBuffer ();
  if ( workspace )
    pkv_SetScratchMemTop ( workspace );
#undef DDIST
} /*_DrawDotsInArea*/

void GeomObjectBSplinePatchDrawTrimmedDomain ( GO_BSplinePatch *obj,
                                               CameraRecd *CPos )
{
  BSP_TrimmedDomain *trpd;
  int               nelem, i;
  BSP_TrPolycurve   *ebound;
  mbs_polycurved    *elem;
  int               degu, degv, lknu, lknv;
  double            *knu, *knv;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ebound = trpd->ebound;

  degu = obj->degree_u;
  degv = obj->degree_v;
  lknu = obj->lastknot_u;
  lknv = obj->lastknot_v;
  knu  = obj->knots_u;
  knv  = obj->knots_v;
  _DrawDotsInArea ( CPos, knu[degu], knu[lknu-degu],
                    knv[degv], knv[lknv-degv], trpd );

  if ( !trpd->show_cpoly )
    xgeSetForeground ( xgec_White );
        /* draw the elements */
  for ( i = 0; i < nelem; i++ ) {
    if ( trpd->show_cpoly ) {
      if ( i == trpd->current_elem )
        xgeSetForeground ( xgec_White );
      else
        xgeSetForeground ( xgec_Grey2 );
    }
    elem = &ebound[i].elem;
    if ( elem->knots )  /* draw a B-spline curve */
      _DrawTrdBSCurve ( CPos, elem->degree, elem->lastknot,
                        elem->knots, elem->cpdimen, elem->points );
    else if ( elem->degree > 1 ) /* draw a Bezier curve */
      _DrawTrdBezCurve ( CPos, elem->degree,
                         elem->cpdimen, elem->points );
    else  /* draw a polyline */
      _DrawTrdPolyline ( CPos, elem->lastknot+1,
                         elem->cpdimen, elem->points );
  }
  if ( trpd->show_cpoly ) {
        /* draw their control polygons */
    for ( i = 0; i < nelem; i++ ) {
      elem = &ebound[i].elem;
      if ( elem->knots ) { /* a B-spline curve */
        _DrawControlPolygon ( CPos, elem->cpdimen, elem->lastknot-elem->degree,
                              elem->points, i == trpd->current_elem );
        _DrawControlPoints ( CPos, elem->cpdimen, elem->lastknot-elem->degree,
                             elem->points, ebound[i].mkcp, MASK_MARKED );
      }
      else if ( elem->degree > 1 ) {
        _DrawControlPolygon ( CPos, elem->cpdimen, elem->degree+1,
                              elem->points, i == trpd->current_elem );
        _DrawControlPoints ( CPos, elem->cpdimen, elem->degree+1,
                             elem->points, ebound[i].mkcp, MASK_MARKED );
      }
      else {
        _DrawControlPoints ( CPos, elem->cpdimen, elem->lastknot+1,
                             elem->points, ebound[i].mkcp, MASK_MARKED );
      }
    }
      /* draw the division points */
    for ( i = 0; i < trpd->ndivpt; i++ )
      _DrawDivisionPoint ( CPos, &trpd->divpt[i], i == trpd->ndivpt-1 );
  }
} /*GeomObjectBSplinePatchDrawTrimmedDomain*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean GeomObjectBSplinePatchTrDFindNearestPoint ( GO_BSplinePatch *obj,
                                                    CameraRecd *CPos,
                                                    short x, short y )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  mbs_polycurved    *elem;
  int               nelem;
  int               i, j, np;
  int               dist, dmin;
  int               imin, jmin;
  byte              dim, deg;
  double            *points;
  point2d           q, r;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ebound = trpd->ebound;
  dmin = 100;
  imin = jmin = -1;
  for ( i = 0; i < nelem; i++ ) {
    elem = &ebound[i].elem;
    dim = elem->cpdimen;
    deg = elem->degree;
    points = elem->points;
    if ( elem->knots )
      np = elem->lastknot-deg;
    else if ( deg > 1 )
      np = deg+1;
    else
      np = elem->lastknot+1;
    switch ( dim ) {
  case 2:
      for ( j = 0; j < np; j++ ) {
        CameraProjectPoint2d ( CPos, (point2d*)&points[2*j], &q );
        dist = (int)(fabs(x-q.x) + fabs(y-q.y));
        if ( dist < dmin )
          { dmin = dist;  imin = i;  jmin = j; }
      }
      break;
  case 3:
      for ( j = 0; j < np; j++ ) {
        Point3to2d ( (point3d*)&points[3*j], &r );
        CameraProjectPoint2d ( CPos, &r, &q );
        dist = (int)(fabs(x-q.x) + fabs(y-q.y));
        if ( dist < dmin )
          { dmin = dist;  imin = i;  jmin = j; }
      }
          /* Farin points are to be added */
      break;
  default:
      break;
    }
  }
  trpd->celem = imin;
  trpd->cpt = jmin;
/*printf ( "imin = %d, jmin = %d, dmin = %d\n", imin, jmin, dmin );*/
  return dmin < 10;
} /*GeomObjectBSplinePatchTrDFindNearestPoint*/

static void _ModifyNeighbourOrigin ( BSP_TrimmedDomain *trpd, int el,
                                     point2d *pos )
{
  mbs_polycurved *elem;
  byte           dim, deg;
  int            i;
  double         *knots, *newkn, *oldkn, *points;

  elem = &trpd->ebound[el].elem;
  dim = elem->cpdimen;
  deg = elem->degree;
  points = elem->points;
  if ( (knots = elem->knots) ) {  /* a B-spline curve, not closed */
    if ( (newkn = pkv_GetScratchMemd ( 2*(deg+1) )) ) {
      oldkn = &newkn[deg+1];
      memcpy ( oldkn, knots, (deg+1)*sizeof(double) );
      for ( i = 0; i <= deg; i++ )
        newkn[i] = oldkn[deg];
      mbs_multiBSChangeLeftKnotsd ( 1, dim, deg, knots, 0, points, newkn );
      if ( dim == 2 )
        memcpy ( points, pos, sizeof(point2d) );
      else
        SetPoint2d ( (point2d*)points, pos->x*points[2], pos->y*points[2] );
      mbs_multiBSChangeLeftKnotsd ( 1, dim, deg, knots, 0, points, oldkn );
      pkv_SetScratchMemTop ( newkn );
    }
  }
  else { /* a Bezier curve or a polyline */
    if ( dim == 2 )
      memcpy ( points, pos, 2*sizeof(double) );
    else   /* preserve the weight */
      SetPoint2d ( (point2d*)points, pos->x*points[2], pos->y*points[2] );
  }
} /*_ModifyNeighbourOrigin*/

static void _ModifyNeighbourEnd ( BSP_TrimmedDomain *trpd, int el,
                                  point2d *pos )
{
  mbs_polycurved *elem;
  byte           dim, deg;
  int            lkn, i, j;
  double         *knots, *newkn, *oldkn, *points;

  elem = &trpd->ebound[el].elem;
  dim = elem->cpdimen;
  deg = elem->degree;
  lkn = elem->lastknot;
  points = elem->points;
  if ( (knots = elem->knots) ) {  /* a B-spline curve, not closed */
    if ( (newkn = pkv_GetScratchMemd ( 2*(deg+1) )) ) {
      oldkn = &newkn[deg+1];
      j = lkn-deg;
      memcpy ( oldkn, &knots[j], (deg+1)*sizeof(double) );
      for ( i = 0; i <= deg; i++ )
        newkn[i] = oldkn[0];
      mbs_multiBSChangeRightKnotsd ( 1, dim, deg, lkn, knots, 0, points, newkn );
      if ( dim == 2 )
        memcpy ( &points[2*(j-1)], pos, sizeof(point2d) );
      else
        SetPoint2d ( (point2d*)&points[3*(j-1)],
                     pos->x*points[3*j-1], pos->y*points[3*j-1] );
      mbs_multiBSChangeRightKnotsd ( 1, dim, deg, lkn, knots, 0, points, oldkn );
      pkv_SetScratchMemTop ( newkn );
    }
  }
  else if ( deg > 1 ) { /* a Bezier curve */
    lkn = deg;
    goto continue_poly;
  }
  else { /* a polyline */
continue_poly:
    if ( dim == 2 )
      memcpy ( &points[2*lkn], pos, 2*sizeof(double) );
    else   /* preserve the weight */
      SetPoint2d ( (point2d*)&points[3*lkn],
                   pos->x*points[3*lkn+2], pos->y*points[3*lkn+2] );
  }
} /*_ModifyNeighbourEnd*/

static void _DeleteDivisionPoint ( GO_BSplinePatch *obj, int i )
{
  BSP_TrimmedDomain *trpd;
  int               j;

  trpd = obj->trpd;
  if ( i >= trpd->ndivpt ) return;
  if ( i < trpd->ndivpt-1 ) {
    j = trpd->ndivpt-i-1;
    memmove ( &trpd->dptel[i], &trpd->dptel[i+1], j*sizeof(int) );
    memmove ( &trpd->dptparam[i], &trpd->dptparam[i+1], j*sizeof(double) );
    memmove ( &trpd->divpt[i], &trpd->divpt[i+1], j*sizeof(point2d) );
  }
  trpd->ndivpt --;
} /*_DeleteDivisionPoint*/

static void _UpdateDivisionPoints ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  mbs_polycurved    *elem;
  int               nelem, ndivpt, i, j;
  int               dim, deg, lkn;
  double            *knots, *points, t;
  point3d           q;

  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ndivpt = trpd->ndivpt;
  for ( i = 0; i < ndivpt; i++ ) {
    if ( trpd->dptel[i] < nelem ) {
      elem = &trpd->ebound[trpd->dptel[i]].elem;
      dim = elem->cpdimen;
      deg = elem->degree;
      lkn = elem->lastknot;
      knots = elem->knots;
      points = elem->points;
      t = trpd->dptparam[i];
      if ( knots ) {
        if ( t < knots[deg] || t > knots[lkn-deg] )
          goto delete;
        if ( dim == 2 )
          mbs_deBoorC2d ( deg, lkn, knots, points, t, &trpd->divpt[i] );
        else
          mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, t, &trpd->divpt[i] );
      }
      else if ( deg > 1 ) {
        if ( t < 0.0 || t > 1.0 )
          goto delete;
        if ( dim == 2 )
          mbs_BCHornerC2d ( deg, points, t, &trpd->divpt[i] );
        else
          mbs_BCHornerC2Rd ( deg, (point3d*)points, t, &trpd->divpt[i] );
      }
      else {
        j = (int)t;
        t -= (double)j;
        if ( j < 0 || j >= lkn )
          goto delete;
        if ( dim == 2 )
          InterPoint2d ( (point2d*)&points[2*j], (point2d*)&points[2*(j+1)],
                         t, &trpd->divpt[i] );
        else {
          InterPoint3d ( (point3d*)&points[3*j], (point3d*)&points[3*(j+1)], t, &q );
          Point3to2d ( &q, &trpd->divpt[i] );
        }
      }
    }
    else {
delete:
      _DeleteDivisionPoint ( obj, i );
    }
  }
} /*_UpdateDivisionPoints*/

void GeomObjectBSplinePatchTrDSetCurrentPoint ( GO_BSplinePatch *obj,
                                                CameraRecd *CPos,
                                                short x, short y )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               nelem, j, np, lkn, clcK;
  mbs_polycurved    *elem;
  byte              dim, deg;
  point2d           q, r;
  double            *knots, *points;
  byte              elemtype;  /* 0  polyline, 1 - Bezier, 2 - B-spline curve */

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  if ( trpd->celem >= 0 && trpd->celem < nelem ) {
    eb = &trpd->ebound[trpd->celem];
    elem = &eb->elem;
    dim = elem->cpdimen;
    deg = elem->degree;
    lkn = elem->lastknot;
    points = elem->points;
    if ( (knots = elem->knots) )
      { np = lkn-deg;  elemtype = 2; }
    else if ( deg > 1 )
      { np = deg+1;  elemtype = 1; }
    else
      { np = lkn+1;  elemtype = 0; }
    j = trpd->cpt;
    if ( j >= 0 && j < np ) {
      SetPoint2d ( &q, x, y );
      CameraUnProjectPoint2d ( CPos, &q, &r );
      switch ( dim ) {
    case 2:
        memcpy ( &points[2*j], &r, sizeof(point2d) );
        break;
    case 3:
        SetPoint2d ( (point2d*)&points[3*j],
                     r.x*points[3*j+2], r.y*points[3*j+2] );
        break;
    default:
        break;
      }
        /* neighbouring elements may have to be modified */
      switch ( elemtype ) {
    case 1: /* Bezier curve */
    case 0: /* polyline */
        if ( j == 0 )
          _ModifyNeighbourEnd ( trpd, eb->prev, &r );
        else if ( j == np-1 )
          _ModifyNeighbourOrigin ( trpd, eb->next, &r );
        break;
    case 2: /* B-spline curve */
        if ( eb->elem.closing == 2 ) { /* a closed B-spline curve */
          clcK = lkn - 2*deg;
          if ( j < deg )
            memcpy ( &points[dim*(j+clcK)], &points[dim*j], dim*sizeof(double) );
          else if ( j >= clcK )
            memcpy ( &points[dim*j], &points[dim*(j-clcK)], dim*sizeof(double) );
        }
        else {
          if ( j < deg ) { /* a non-closed B-spline curve */
            if ( dim == 2 )
              mbs_deBoorC2d ( deg, lkn, knots, points, knots[deg], &r );
            else
              mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[deg], &r );
           _ModifyNeighbourEnd ( trpd, eb->prev, &r );
          }
          if ( j > lkn-2*deg ) {
            if ( dim == 2 )
              mbs_deBoorC2d ( deg, lkn, knots, points, knots[lkn-deg], &r );
            else
              mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[lkn-deg], &r );
            _ModifyNeighbourOrigin ( trpd, eb->next, &r );
          }
        }
        break;
      }
    }
  }
  _UpdateDivisionPoints ( obj );
  obj->me.dlistmask = 0;
} /*GeomObjectBSplinePatchTrDSetCurrentPoint*/

void GeomObjectBSplinePatchTrDSelectPoints ( GO_BSplinePatch *obj,
                              CameraRecd *CPos, Box2s *sel_rect, int action )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  short             dim, deg;
  int               nelem, lkn, np, i;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ebound = trpd->ebound;
  for ( i = 0; i < nelem; i++ ) {
    dim = ebound[i].elem.cpdimen;
    deg = ebound[i].elem.degree;
    lkn = ebound[i].elem.lastknot;
    if ( ebound[i].elem.knots )  np = lkn-deg;
    else if ( deg > 1 )          np = deg+1;
    else                         np = lkn+1;
    GeomObjectMarkPoints ( dim, 2, np, ebound[i].mkcp,
                           ebound[i].elem.points, CPos, sel_rect,
                           MASK_CP_MARKED_0, action );
  }
} /*GeomObjectBSplinePatchTrDSelectPoints*/

void GeomObjectBSplinePatchTrDSavePoints ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  short             dim, deg;
  int               nelem, lkn, np, i;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ebound = trpd->ebound;
  for ( i = 0; i < nelem; i++ )
    if ( ebound[i].savedpoints )
      { free ( ebound[i].savedpoints );  ebound[i].savedpoints = NULL; }
  for ( i = 0; i < nelem; i++ ) {
    dim = ebound[i].elem.cpdimen;
    deg = ebound[i].elem.degree;
    lkn = ebound[i].elem.lastknot;
    if ( ebound[i].elem.knots )  np = lkn-deg;
    else if ( deg > 1 )          np = deg+1;
    else                         np = lkn+1;
    ebound[i].savedpoints = malloc ( dim*np*sizeof(double) );
    if ( ebound[i].savedpoints )
      memcpy ( ebound[i].savedpoints, ebound[i].elem.points,
               dim*np*sizeof(double) );
  }
} /*GeomObjectBSplinePatchTrDSavePoints*/

void GeomObjectBSplinePatchTrDTransformPoints ( GO_BSplinePatch *obj,
                                                trans2d *tr )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  short             dim, deg;
  int               nelem, lkn, np, i, j;
  byte              *mkcp;
  double            *points, *spoints;
  point2d           p, q;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd   = obj->trpd;
  nelem  = trpd->nelem;
  ebound = trpd->ebound;
  for ( i = 0; i < nelem; i++ ) {
    dim = ebound[i].elem.cpdimen;
    deg = ebound[i].elem.degree;
    lkn = ebound[i].elem.lastknot;
    if ( ebound[i].elem.knots )  np = lkn-deg;
    else if ( deg > 1 )          np = deg+1;
    else                         np = lkn+1;
    spoints = ebound[i].savedpoints;
    if ( spoints ) {
      points  = ebound[i].elem.points;
      mkcp    = ebound[i].mkcp;
      switch ( dim ) {
    case 2:
        for ( j = 0; j < np; j++ )
          if ( mkcp[j] & MASK_CP_MARKED_0 )
            TransPoint2d ( tr, (point2d*)&spoints[2*j], (point2d*)&points[2*j] );
        break;
    case 3:
        for ( j = 0; j < np; j++ )
          if (mkcp[j] & MASK_CP_MARKED_0 ) {
            Point3to2d ( (point3d*)&spoints[3*j], &p );
            TransPoint2d ( tr, &p, &q );
            SetPoint2d ( (point2d*)&points[3*j],
                         q.x*spoints[3*j+2], q.y*spoints[3*j+2] );
          }
        break;
    default:
        break;
      }
    }
  }
  _UpdateDivisionPoints ( obj );
  obj->me.dlistmask = 0;
} /*GeomObjectBSplinePatchTrDTransformPoints*/

/* ////////////////////////////////////////////////////////////////////////// */
boolean GeomObjectBSplinePatchTrDSelectElem ( GO_BSplinePatch *obj, int el,
                  int *edim, int *edeg, int *elkn, double **eknots, boolean *closed )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound;
  int               ce;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  ce = trpd->current_elem;
  switch ( el ) {
case BSP_TRD_SELECT_CURRENT:
    goto get_knots;
case BSP_TRD_SELECT_PREVIOUS:
    trpd->current_elem = ce = ebound[ce].prev;
    goto get_knots;
case BSP_TRD_SELECT_NEXT:
    trpd->current_elem = ce = ebound[ce].next;
    goto get_knots;
default:
    if ( el < 0 ) el = trpd->nelem-1;
    if ( el >= trpd->nelem ) el = 0;
    trpd->current_elem = ce = el;
    goto get_knots;
  }
get_knots:
  *edim = ebound[ce].elem.cpdimen;
  *edeg = ebound[ce].elem.degree;
  *elkn = ebound[ce].elem.lastknot;
  *eknots = ebound[ce].elem.knots;
  *closed = ebound[ce].elem.closing == 2;
  return true;
} /*GeomObjectBSplinePatchTrDSelectElem*/

int GeomObjectBSplinePatchTrDGetCurrentElem ( GO_BSplinePatch *obj )
{
  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return 0;
  if ( !obj->trimmed || !obj->trpd )
    return 0;
  return obj->trpd->current_elem;
} /*GeomObjectBSplinePatchTrDGetCurrentElem*/

boolean GeomObjectBSplinePatchTrDSetElemDegree ( GO_BSplinePatch *obj, int deg,
                         int *edeg, int *elkn, double **eknots )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               ce, dim, degree, lastknot;
  int               k, nkn, ncp, nnlkn, nncp;
  double            *knots, *points, *nknots, *npoints;
  byte              *mkcp;
  boolean           closed;
  point2d           pos;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  ce = trpd->current_elem;
  eb = &trpd->ebound[ce];
  dim    = eb->elem.cpdimen;
  degree = eb->elem.degree;
  if ( deg < 1 || deg > MAX_DEGREE || deg == degree )
    return false;
  lastknot = eb->elem.lastknot;
  knots = eb->elem.knots;
  points = eb->elem.points;
  closed = eb->elem.closing == 2;
        /* a polyline or a Bezier curve must be converted to B-spline form */
  if ( !knots ) {
    if ( degree > 1 ) {  /* a Bezier curve */
      knots = eb->elem.knots = malloc ( 2*(degree+1)*sizeof(double) );
      if ( !knots )
        return false;
      for ( k = 0; k <= degree; k++ )
        knots[k] = 0.0;
      for ( ; k <= 2*degree+1; k++ )
        knots[k] = 1.0;
      eb->elem.lastknot = lastknot = 2*degree+1;
    }
    else { /* a polyline */
      knots = eb->elem.knots = malloc ( (lastknot+3)*sizeof(double) );
      if ( !knots )
        return false;
      for ( k = 0; k <= lastknot+2; k++ )
        knots[k] = (double)(k-1);
      eb->elem.lastknot = lastknot = lastknot+2;
    }
  }
        /* now the proper degree elevation or reduction */
  k = mbs_NumKnotIntervalsd ( degree, lastknot, knots );
  if ( deg > degree ) {  /* elevation */
    nkn = lastknot + (k+2)*(deg-degree);
    ncp = nkn - deg;
    nknots = malloc ( (nkn+1)*sizeof(double) );
    npoints = malloc ( ncp*dim*sizeof(double) );
    mkcp = malloc ( ncp*sizeof(byte) );
    if ( !nknots || !npoints || !mkcp )
      goto failure;
    if ( !closed ) {
      if ( !mbs_multiBSDegElevd ( 1, dim, degree, lastknot, knots, 0, points,
                                  deg-degree, &deg, &nnlkn, nknots, 0, npoints,
                                  false ) )
        goto failure;
    }
    else {
      if ( !mbs_multiBSDegElevClosedd ( 1, dim, degree, lastknot, knots, 0, points,
                                  deg-degree, &deg, &nnlkn, nknots, 0, npoints ) )
        goto failure;
    }
  }
  else {  /* reduction */
    nkn = (k+1)*(deg+1);  /* an estimate */
    ncp = nkn-deg;
    nknots = malloc ( (nkn+1)*sizeof(double) );
    npoints = malloc ( ncp*dim*sizeof(double) );
    mkcp = malloc ( ncp*sizeof(byte) );
    if ( !nknots || !npoints || !mkcp )
      goto failure;
    if ( !closed )
      mbs_multiBSDegRedd ( 1, dim, degree, lastknot, knots, 0, points,
                           degree-deg, &deg, &nnlkn, nknots, 0, npoints );
    else
      mbs_multiBSDegRedClosedd ( 1, dim, degree, lastknot, knots, 0, points,
                           degree-deg, &deg, &nnlkn, nknots, 0, npoints );
  }
  nncp = nnlkn - deg;
  free ( knots );
  free ( points );
  free ( eb->mkcp );
  *eknots = eb->elem.knots = nknots;
  eb->elem.points = npoints;
  eb->mkcp = mkcp;
  memset ( mkcp, 0, nncp );
  *edeg = eb->elem.degree = deg;
  *elkn = eb->elem.lastknot = nnlkn;
  if ( deg < degree && !closed ) {
    if ( dim == 2 )
      mbs_deBoorC2d ( deg, nnlkn, nknots, npoints, nknots[nnlkn-deg], &pos );
    else
      mbs_deBoorC2Rd ( deg, nnlkn, nknots, (point3d*)npoints,
                       nknots[nnlkn-deg], &pos );
    _ModifyNeighbourOrigin ( trpd, eb->next, &pos );
    if ( dim == 2 )
      mbs_deBoorC2d ( deg, nnlkn, nknots, npoints, nknots[deg], &pos );
    else
      mbs_deBoorC2Rd ( deg, nnlkn, nknots, (point3d*)npoints,
                       nknots[deg], &pos );
    _ModifyNeighbourEnd ( trpd, eb->prev, &pos );
  }
  _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( nknots )  free ( nknots );
  if ( npoints ) free ( npoints );
  if ( mkcp )    free ( mkcp );
  return false;
} /*GeomObjectBSplinePatchTrDSetElemDegree*/

boolean GeomObjectBSplinePatchTrDNewContour ( GO_BSplinePatch *obj, int *el )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound, *neb;
  mbs_polycurved    *bound, *nb;
  int               nelem;
  double            *points, *svp;
  byte              *mkcp;
  double            u0, u1, v0, v1;
  int               i, np;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  bound = trpd->bound;
  nelem = trpd->nelem;
  neb = malloc ( (nelem+1)*sizeof(BSP_TrPolycurve) );
  nb = malloc ( (nelem+1)*sizeof(mbs_polycurved) );
  points = malloc ( 5*sizeof(point2d) );
  svp = malloc ( 5*sizeof(point2d) );
  mkcp = malloc ( 4*sizeof(byte) );
  if ( !neb || !nb || !points || !svp || !mkcp )
    goto failure;
  memcpy ( neb, ebound, (nelem+1)*sizeof(BSP_TrPolycurve) );
  memcpy ( nb, bound, (nelem+1)*sizeof(mbs_polycurved) );
  free ( ebound );
  free ( bound );
  trpd->ebound = ebound = neb;
  trpd->bound = bound = nb;
  nb = &bound[nelem];
  neb = &ebound[nelem];
  nb->ident = nelem;
  nb->closing = 2;
  nb->cpdimen = 2;
  nb->degree = 1;
  nb->lastknot = 4;
  nb->knots = NULL;
  nb->points = points;
  neb->prev = neb->next = nelem;
  neb->elem = *nb;
  neb->savedpoints = svp;
  neb->mkcp = mkcp;
  neb->tag = 0;
  u0 = obj->knots_u[obj->degree_u];
  u1 = obj->knots_u[obj->lastknot_u-obj->degree_u];
  v0 = obj->knots_v[obj->degree_v];
  v1 = obj->knots_v[obj->lastknot_v-obj->degree_v];
  points[0] = points[6] = points[8] = u0;
  points[2] = points[4] = u1;
  points[1] = points[3] = points[9] = v0;
  points[5] = points[7] = v1;
  memset ( mkcp, MASK_MARKED, 5 );
        /* unmark all other points */
  for ( i = 0; i < nelem; i++ ) {
    nb = &bound[i];
    if ( nb->knots )           np = nb->lastknot-nb->degree;
    else if ( nb->degree > 1 ) np = nb->degree+1;
    else                       np = nb->lastknot+1;
    memset ( ebound[i].mkcp, 0, np );
  }
  if ( el )
    *el = nelem;
  trpd->nelem ++;
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( neb )    free ( neb );
  if ( nb )     free ( nb );
  if ( points ) free ( points );
  if ( svp )
  if ( mkcp )   free ( mkcp );
  return false;
} /*GeomObjectBSplinePatchTrDNewContour*/

boolean GeomObjectBSplinePatchTrDDeleteContour ( GO_BSplinePatch *obj )
{
  void              *sp;
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound, *neb;
  mbs_polycurved    *nb;
  int               nelem, nnelem, i, j;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  sp = pkv_GetScratchMemTop ();
  trpd = obj->trpd;
  ebound = trpd->ebound;
  nelem = trpd->nelem;
  if ( nelem < 1 )
    return false;
  for ( i = 0; i < nelem; i++ )
    ebound[i].tag = 0;
  nnelem = nelem;
  for ( i = trpd->current_elem; ebound[i].tag >= 0; i = ebound[i].next ) {
    ebound[i].tag = -1;
    nnelem --;
  }
  if ( nnelem <= 0 )
    return false;
  neb = malloc ( nnelem*sizeof(BSP_TrPolycurve) );
  nb = malloc ( nnelem*sizeof(mbs_polycurved) );
  if ( !neb || !nb )
    goto failure;
  for ( i = nnelem = 0;  i < nelem;  i++ ) {
    if ( ebound[i].tag == -1 ) {
      if ( ebound[i].elem.knots )  free ( ebound[i].elem.knots );
      if ( ebound[i].elem.points ) free ( ebound[i].elem.points );
      if ( ebound[i].savedpoints ) free ( ebound[i].savedpoints );
      if ( ebound[i].mkcp )        free ( ebound[i].mkcp );
      for ( j = 0; j < trpd->ndivpt; j++ )
        if ( trpd->dptel[j] == i )
          _DeleteDivisionPoint ( obj, j );
    }
    else {
      neb[nnelem].tag = i;
      neb[nnelem].elem = ebound[i].elem;
      neb[nnelem].savedpoints = ebound[i].savedpoints;
      neb[nnelem].mkcp = ebound[i].mkcp;
      for ( j = 0; j < trpd->ndivpt; j++ )
        if ( trpd->dptel[j] == i )
          trpd->dptel[j] = nnelem;
      ebound[i].tag = nnelem ++;
    }
  }
  for ( i = 0; i < nnelem; i++ ) {
    neb[i].prev = ebound[ebound[neb[i].tag].prev].tag;
    neb[i].next = ebound[ebound[neb[i].tag].next].tag;
  }
  free ( ebound );
  free ( trpd->bound );
  trpd->ebound = neb;
  trpd->bound = nb;
  trpd->nelem = nnelem;
  trpd->current_elem = 0;
  _PackTheBoundary ( nnelem, neb, nb );
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( neb ) free ( neb );
  if ( nb )  free ( nb );
  pkv_SetScratchMemTop ( sp );
  return false;
} /*GeomObjectBSplinePatchTrDDeleteContour*/

boolean GeomObjectBSplinePatchTrDReverseContour ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound, *eb;
  int               nelem, dim, deg, lkn, np, i, j;
  double            *points, *knots;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  nelem = trpd->nelem;
  for ( i = 0; i < nelem; i++ )
    ebound[i].tag = 1;
  for ( i = trpd->current_elem; ebound[i].tag; i = ebound[i].next ) {
    eb = &ebound[i];
    eb->tag = 0;
    dim = eb->elem.cpdimen;
    deg = eb->elem.degree;
    lkn = eb->elem.lastknot;
    points = eb->elem.points;
    knots = eb->elem.knots;
    if ( knots ) {
      np = lkn-deg;
      pkv_ReverseMatc ( lkn+1, sizeof(double), sizeof(double), (char*)knots );
      for ( j = 0; j <= lkn; j++ )
        knots[j] = -knots[j];
      for ( j = 0; j < trpd->ndivpt; j++ )
        if ( trpd->dptel[j] == i )
          trpd->dptparam[j] = -trpd->dptparam[j];
    }
    else if ( deg > 1 ) {
      np = deg+1;
      for ( j = 0; j < trpd->ndivpt; j++ )
        if ( trpd->dptel[j] == i )
          trpd->dptparam[j] = 1.0-trpd->dptparam[j];
    }
    else {
      np = lkn+1;
      for ( j = 0; j < trpd->ndivpt; j++ )
        if ( trpd->dptel[j] == i )
          trpd->dptparam[j] = (double)lkn-trpd->dptparam[j];
    }
    pkv_ReverseMatc ( np, dim*sizeof(double), dim*sizeof(double), (char*)points );
    pkv_ReverseMatc ( np, sizeof(byte), sizeof(byte), (char*)eb->mkcp );
    j = eb->prev;
    eb->prev = eb->next;
    eb->next = j;
  }
  _PackTheBoundary ( nelem, ebound, trpd->bound );
  obj->me.dlistmask = 0;
  return true;
} /*GeomObjectBSplinePatchTrDReverseContour*/

boolean GeomObjectBSplinePatchTrDInsertKnot ( GO_BSplinePatch *obj, double knot )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               el, dim, deg, lkn;
  double            *knots, *points;
  byte              *mkcp;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  el = trpd->current_elem;
  if ( el < 0 || el >= trpd->nelem )
    return false;
  eb = &trpd->ebound[el];
  if ( !eb->elem.knots )
    return false;
  dim = eb->elem.cpdimen;
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  knots = malloc ( (lkn+2)*sizeof(double) );
  points = malloc ( (lkn-deg+1)*dim*sizeof(double) );
  mkcp = malloc ( (lkn-deg+1)*sizeof(byte) );
  if ( !points || !knots || !mkcp ) {
    if ( knots )  free ( knots );
    if ( points ) free ( points );
    if ( mkcp )   free ( mkcp );
    return false;
  }
  memcpy ( knots, eb->elem.knots, (lkn+1)*sizeof(double) );
  memcpy ( points, eb->elem.points, (lkn-deg)*dim*sizeof(double) );
  mbs_multiKnotInsd ( deg, &lkn, knots, 1, dim, 0, 0, points, knot );
  memset ( mkcp, 0, (lkn-deg)*sizeof(byte) );
  free ( eb->elem.knots );
  free ( eb->elem.points );
  free ( eb->mkcp );
  eb->elem.knots = knots;
  eb->elem.points = points;
  eb->elem.lastknot = lkn;
  eb->mkcp = mkcp;
  _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
  obj->me.dlistmask = 0;
  return true;
} /*GeomObjectBSplinePatchTrDInsertKnot*/

boolean GeomObjectBSplinePatchTrDRemoveKnot ( GO_BSplinePatch *obj, int ki )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               el, deg, lkn;
  double            *knots, *points;
  byte              *mkcp;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  el = trpd->current_elem;
  if ( el < 0 || el >= trpd->nelem )
    return false;
  eb = &trpd->ebound[el];
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  knots = eb->elem.knots;
  points = eb->elem.points;
  if ( deg <= 0 || deg > MAX_DEGREE || lkn <= 2*deg+1 || !knots || !points )
    return false;
  if ( ki <= deg || ki >= lkn-deg )
    return false;
  if ( eb->elem.closing == 2 )
    mbs_multiKnotRemoveClosedd ( deg, &lkn, knots, 1, eb->elem.cpdimen,
                           0, 0, points, ki );
  else
    mbs_multiKnotRemoved ( deg, &lkn, knots, 1, eb->elem.cpdimen,
                           0, 0, points, ki );
  eb->elem.lastknot = lkn;
  knots = malloc ( (lkn+1)*sizeof(double) );
  points = malloc ( (lkn-deg)*eb->elem.cpdimen*sizeof(double) );
  mkcp = malloc ( (lkn-deg)*sizeof(byte) );
  if ( knots && points && mkcp ) {
    memcpy ( knots, eb->elem.knots, (lkn+1)*sizeof(double) );
    memcpy ( points, eb->elem.points, (lkn-deg)*eb->elem.cpdimen*sizeof(double) );
    memset ( mkcp, 0, (lkn-deg)*sizeof(byte) );
    free ( eb->elem.knots );
    free ( eb->elem.points );
    free ( eb->mkcp );
    eb->elem.knots = knots;
    eb->elem.points = points;
    eb->mkcp = mkcp;
    _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
    obj->me.dlistmask = 0;
    return true;
  }
  else {
    if ( knots )  free ( knots );
    if ( points ) free ( points );
    if ( mkcp )   free ( mkcp );
    return false;
  }
} /*GeomObjectBSplinePatchTrDRemoveKnot*/

boolean GeomObjectBSplinePatchTrDChangeKnot ( GO_BSplinePatch *obj,
                                              int ki, double knot )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               el, deg, lkn;
  double            *knots, *points;
  point2d           pos;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
/*printf ( "ki = %d, knot = %f\n", ki, knot );*/
  trpd = obj->trpd;
  el = trpd->current_elem;
  if ( el < 0 || el >= trpd->nelem )
    return false;
  eb = &trpd->ebound[el];
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  knots = eb->elem.knots;
  points = eb->elem.points;
  if ( deg <= 0 || deg > MAX_DEGREE || lkn <= 2*deg || !knots || !points )
    return false;
  if ( eb->elem.closing != 2 ) {
    if ( ki <= deg+1 ) {
      if ( eb->elem.cpdimen == 2 )
        mbs_deBoorC2d ( deg, lkn, knots, (point2d*)points, knots[deg], &pos );
      else
        mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[deg], &pos );
      _ModifyNeighbourEnd ( trpd, eb->prev, &pos );
    }
    if ( ki >= lkn-deg-1 ) {
      if ( eb->elem.cpdimen == 2 )
        mbs_deBoorC2d ( deg, lkn, knots, (point2d*)points, knots[lkn-deg], &pos );
      else
        mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[lkn-deg], &pos );
      _ModifyNeighbourOrigin ( trpd, eb->next, &pos );
    }
    _UpdateDivisionPoints ( obj );
  }
  obj->me.dlistmask = 0;
  return true;
} /*GeomObjectBSplinePatchTrDChangeKnot*/

boolean GeomObjectBSplinePatchTrDRemoveVertex ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               nelem, ce, cpt, lkn, dim;
  double            *points;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  nelem = trpd->nelem;
  ce = trpd->celem;
  if ( ce < 0 || ce >= nelem )
    return false;
  eb = &trpd->ebound[ce];
        /* can remove a vertex of a polyline */
  if ( eb->elem.degree > 1 )
    return false;
  lkn = eb->elem.lastknot;
  cpt = trpd->cpt;
  if ( lkn < 2 || cpt < 0 || cpt > lkn )
    return false;
        /* cannot remove end vertices */
  if ( eb->elem.closing != 2 && (cpt <= 0 || cpt >= lkn) )
    return false;
        /* for a spline of degree 1 forget the knot sequence */
  if ( eb->elem.knots ) {
    free ( eb->elem.knots );
    eb->elem.knots = NULL;
        /* lastknot for a polyline is the number of line segments */
        /* which is less by 2 than the number of the last knot in the */
        /* B-spline representation */
    eb->elem.lastknot = lkn -= 2;
  }
  points = eb->elem.points;
  dim = eb->elem.cpdimen;
  if ( cpt < lkn )
    memmove ( &points[dim*cpt], &points[dim*(cpt+1)], dim*(lkn-cpt)*sizeof(double) );
  eb->elem.lastknot = --lkn;
  if ( eb->elem.closing == 2 )
    memcpy ( &points[dim*lkn], points, dim*sizeof(double) );
  trpd->celem = trpd->cpt = -1;
  _PackTheBoundary ( nelem, trpd->ebound, trpd->bound );
  obj->me.dlistmask = 0;
  return true;
} /*GeomObjectBSplinePatchTrDRemoveVertex*/

boolean GeomObjectBSplinePatchTrDEnterDivPt ( GO_BSplinePatch *obj, point2d *p,
                                              int *el )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  int               nelem, i, j, minelem, dim, deg, lkn;
  double            *points;
  point2d           dpt, mindpt;
  double            t, dist, mint, mindist;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  nelem = trpd->nelem;
  mindist = dist = mint = 1.0e+300;
  minelem = -1;
  for ( i = 0; i < nelem; i++ ) {
    eb = &trpd->ebound[i];
    dim = eb->elem.cpdimen;
    deg = eb->elem.degree;
    lkn = eb->elem.lastknot;
    points = eb->elem.points;
    if ( eb->elem.knots ) { /* a B-spline curve */
      if ( dim == 2 )
        mbs_FindBS2ClosestPointd ( deg, lkn, eb->elem.knots,
                                  (point2d*)points, p, &t, &dpt, &dist );
      else
        mbs_FindBS2RClosestPointd ( deg, lkn, eb->elem.knots,
                                    (point3d*)points, p, &t, &dpt, &dist );
    }
    else if ( deg > 1 ) { /* a Bezier curve */
      if ( dim == 2 )
        mbs_FindBC2ClosestPointd ( deg, (point2d*)points, p, &t, &dpt, &dist );
      else
        mbs_FindBC2RClosestPointd ( deg, (point3d*)points, p, &t, &dpt, &dist );
    }
    else { /* a polyline */
      for ( j = 0; j < lkn; j++ ) {
        if ( dim == 2 )
          mbs_FindBC2ClosestPointd ( 1, (point2d*)&points[2*j], p, &t, &dpt, &dist );
        else
          mbs_FindBC2RClosestPointd ( 1, (point3d*)&points[3*j], p, &t, &dpt, &dist );
        if ( dist < mindist ) {
          mindist = dist;
          mindpt = dpt;
          mint = (double)j+t;
          minelem = i;
        }
      }
    }
    if ( dist < mindist ) {
      mindist = dist;
      mindpt = dpt;
      mint = t;
      minelem = i;
    }
  }

  if ( mindist < 1.0e+300 ) {
    if ( trpd->ndivpt > 1 ) {  /* reject the oldest point */
      trpd->dptel[0] = trpd->dptel[1];
      trpd->dptparam[0] = trpd->dptparam[1];
      trpd->divpt[0] = trpd->divpt[1];
      trpd->ndivpt = 1;
    }
/*printf ( "ndivpt = %d\n", trpd->ndivpt );*/
    trpd->dptel[trpd->ndivpt] = minelem;
    trpd->dptparam[trpd->ndivpt] = mint;
    trpd->divpt[trpd->ndivpt] = mindpt;
    trpd->ndivpt ++;
    *el = minelem;
  }
  else
    *el = trpd->current_elem;
  return true;
} /*GeomObjectBSplinePatchTrDEnterDivPt*/

static boolean _GOBSPTrDDivideElem ( BSP_TrPolycurve *eb0, BSP_TrPolycurve *eb1,
                                     BSP_TrPolycurve *eb2,
                                     int i0, int i1, double param )
{
  void   *sp;
  int    dim, deg, lkn, lkn1, lkn2, i;
  double *points, *points1, *points2, *knots, *knots1, *knots2;
  byte   *mkcp0, *mkcp1;
  int    np, np0, np1, intv, mult;
  double t;

  sp = pkv_GetScratchMemTop ();
  mkcp0 = mkcp1 = NULL;
  points1 = knots1 = NULL;
  dim = eb0->elem.cpdimen;
  deg = eb0->elem.degree;
  lkn = eb0->elem.lastknot;
  points = eb0->elem.points;
  knots = eb0->elem.knots;
  if ( knots ) { /* a B-spline curve */
    if ( param <= knots[deg] || param >= knots[lkn-deg] )
      return false;
    intv = mbs_FindKnotIntervald ( deg, lkn, knots, param, &mult );
    if ( mult > deg+1 )
      return false;
    knots2 = pkv_GetScratchMemd ( lkn+deg+2-mult );
    points2 = pkv_GetScratchMemd ( (lkn+1-mult)*dim );
    if ( !knots2 || !points2 )
      goto failure;
    memcpy ( knots2, knots, (lkn+1)*sizeof(double) );
    memcpy ( points2, points, (lkn-deg)*dim*sizeof(double) );
    lkn2 = lkn;
    for ( i = 0; i <= deg-mult; i++ )
      mbs_multiKnotInsd ( deg, &lkn2, knots2, 1, dim, 0, 0, points2, param );
    lkn1 = intv+deg+1-mult;
    np0 = lkn1-deg;
    lkn = lkn2-lkn1+deg;
    np1 = lkn-deg;
    points1 = malloc ( dim*np0*sizeof(double) );
    knots1 = malloc ( (lkn1+1)*sizeof(double) );
    mkcp0 = malloc ( np0*sizeof(byte) );
    points = malloc ( dim*np1*sizeof(double) );
    knots = malloc ( (lkn+1)*sizeof(double) );
    mkcp1 = malloc ( np1*sizeof(byte) );
    if ( !points || !knots || !mkcp0 || !points1 || !knots1 || !mkcp1 ) {
      if ( points ) free ( points );
      if ( knots )  free ( knots );
      if ( mkcp0 )  free ( mkcp0 );
      goto failure;
    }
    memcpy ( knots, &knots2[lkn1-deg], (lkn+1)*sizeof(double) );
    memcpy ( points, &points2[dim*np0], dim*np1*sizeof(double) );
    memcpy ( knots1, knots2, (lkn1+1)*sizeof(double) );
    memcpy ( points1, points2, dim*np0*sizeof(double) );
    free ( eb0->elem.knots );
    free ( eb0->elem.points );
    free ( eb0->mkcp );
    eb0->elem.knots = knots1;
    eb0->elem.points = points1;
    eb0->mkcp = mkcp0;
    eb0->elem.lastknot = lkn1;
    eb1->elem.knots = knots;
    eb1->elem.points = points;
    eb1->mkcp = mkcp1;
    eb1->elem.lastknot = lkn;
    pkv_SetScratchMemTop ( sp );
  }
  else if ( deg > 1 ) { /* a Bezier curve */
    np0 = np1 = deg+1;
    if ( param <= 0.0 || param >= 1.0 )
      return false;
    points1 = malloc ( dim*(deg+1)*sizeof(double) );
    mkcp1 = malloc ( (deg+1)*sizeof(byte) );
    if ( !points1 || !mkcp1 )
      goto failure;
    mbs_multiDivideBezCurvesd ( deg, 1, dim, 0, param, points, points1 );
    eb1->elem.knots = NULL;
    eb1->elem.points = points;
    eb0->elem.points = points1;
    eb1->elem.lastknot = 0;
  }
  else { /* a polyline */
    np = lkn+1;
    if ( param <= 0.0 || param >= lkn )
      return false;
    np0 = (int)param+2;
    np1 = np-np0+2;
    points = malloc ( dim*np0*sizeof(double) );
    points1 = malloc ( dim*np1*sizeof(double) );
    mkcp0 = malloc ( np0*sizeof(byte) );
    mkcp1 = malloc ( np1*sizeof(byte) );
    if ( !points || !points1 || !mkcp0 || !mkcp1 ) {
      if ( points ) free ( points );
      if ( mkcp0 )  free ( mkcp0 );
      goto failure;
    }
    memcpy ( points, eb0->elem.points, dim*(np0-1)*sizeof(double) );
    memcpy ( &points1[dim], &eb0->elem.points[(np0-1)*dim], dim*(np1-1)*sizeof(double) );
    t = param-floor(param);
    if ( dim == 2 )
      InterPoint2d ( (point2d*)&eb0->elem.points[2*(np0-2)],
                     (point2d*)&eb0->elem.points[2*(np0-1)], t, (point2d*)points1 );
    else
      InterPoint3d ( (point3d*)&eb0->elem.points[3*(np0-2)],
                     (point3d*)&eb0->elem.points[3*(np0-1)], t, (point3d*)points1 );
    memcpy ( &points[dim*(np0-1)], points1, dim*sizeof(double) );
    free ( eb0->elem.points );
    free ( eb0->mkcp );
    eb0->elem.points = points;
    eb0->elem.lastknot = np0-1;
    eb0->mkcp = mkcp0;
    eb1->elem.points = points1;
    eb1->elem.knots = NULL;
    eb1->elem.lastknot = np1-1;
  }
  eb1->savedpoints = NULL;
  eb1->mkcp = mkcp1;
  memset ( eb0->mkcp, 0, np0 );
  memset ( eb1->mkcp, 0, np1 );

  eb1->elem.ident = eb0->elem.ident;
  eb1->elem.cpdimen = dim;
  eb1->elem.degree = deg;
  eb1->elem.closing = eb0->elem.closing;
  eb0->elem.closing = 0;
  eb1->next = eb0->next;
  eb1->prev = i0;
  eb0->next = eb2->prev = i1;
  return true;

failure:
  if ( knots1 )  free ( knots1 );
  if ( points1 ) free ( points1 );
  if ( mkcp1 )   free ( mkcp1 );
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_GOBSPTrDDivideElem*/

static boolean _GOBSPTrDDivideClosedElem ( BSP_TrPolycurve *eb, double param,
                                           double *param1 )
{
  void   *sp;
  int    dim, deg, lkn, lkn1, i, j;
  double *points, *knots, *points1, *knots1, *points2, *knots2, par;
  int    intv, mult;
  byte   *mkcp;

  sp = pkv_GetScratchMemTop ();
  dim = eb->elem.cpdimen;
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  points = eb->elem.points;
  knots = eb->elem.knots;
  if ( knots ) { /* a closed B-spline curve*/
    knots1 = pkv_GetScratchMemd ( deg+1 );
    if ( !knots1 || param < knots[deg] || param > knots[lkn-deg] )
      return false;
    intv = mbs_FindKnotIntervald ( deg, lkn, knots, param, &mult );
    lkn1 = lkn;
    if ( mult > 0 ) {
      knots2 = malloc ( (lkn+1)*sizeof(double) );
      points2 = malloc ( (lkn-deg)*dim*sizeof(double) );
      mkcp = malloc ( (lkn-deg)*sizeof(byte) );
      if ( !knots2 || !points2 || !mkcp ) {
        if ( knots2 ) free ( knots2 );
        if ( points2 ) free ( points2 );
        if ( mkcp ) free ( mkcp );
        return false;
      }
      mbs_multiRollClosedBSCurvesd ( 1, dim, deg, lkn, knots, 0, points,
                                     intv-deg, knots2, points2 );
      if ( param1 && intv != deg )
        *param1 += knots[lkn-deg]-knots[deg];
    }
    else {
      knots1 = pkv_GetScratchMemd ( lkn+2 );
      points1 = pkv_GetScratchMemd ( (lkn+1-deg)*dim );
      knots2 = malloc ( (lkn+2)*sizeof(double) );
      points2 = malloc ( (lkn+1-deg)*dim*sizeof(double) );
      mkcp = malloc ( (lkn+1-deg)*sizeof(byte) );
      if ( !knots1 || !points1 || !knots2 || !points2 || !mkcp ) {
        if ( knots2 ) free ( knots2 );
        if ( points2 ) free ( points2 );
        if ( mkcp ) free ( mkcp );
        pkv_SetScratchMemTop ( sp );
        return false;
      }
      memcpy ( knots1, knots, (lkn+1)*sizeof(double) );
      memcpy ( points1, points, (lkn-deg)*dim*sizeof(double) );
      intv = mbs_multiKnotInsClosedd ( deg, &lkn1, knots1, 1, dim,
                                       0, 0, points1, param ) + 1;
      mbs_multiRollClosedBSCurvesd ( 1, dim, deg, lkn1, knots1, 0, points1,
                                     intv-deg, knots2, points2 );
      if ( param1 )
        *param1 += knots[lkn-deg]-knots[deg];
    }
        /* modify the knots at the end of spline domain */
    for ( i = 0; i <= deg; i++ )
      knots1[i] = knots2[deg];
    mbs_multiBSChangeLeftKnotsd ( 1, dim, deg, knots2, 0, points2, knots1 );
    for ( i = 0; i <= deg; i++ )
      knots1[i] = knots2[lkn1-deg];
    mbs_multiBSChangeRightKnotsd ( 1, dim, deg, lkn1, knots2, 0, points2, knots1 );
        /* assign the new knots and control points */
    free ( knots );
    free ( points );
    free ( eb->mkcp );
    eb->elem.lastknot = lkn1;
    eb->elem.knots = knots2;
    eb->elem.points = points2;
    eb->mkcp = mkcp;
    memset ( mkcp, 0, (lkn1-deg)*sizeof(byte) );
  }
  else if ( deg == 1 ) { /* a closed polyline */
    if ( param < 0.0 || param > lkn )
      return false;
    i = (int)param;
    param -= (double)i;
    if ( param != 0.0 ) { /* a new point to add */
      points1 = pkv_GetScratchMemd ( (lkn+2)*dim );
      if ( !points1 )
        return false;
      points2 = malloc ( (lkn+2)*dim*sizeof(double) );
      mkcp = malloc ( (lkn+2)*sizeof(byte) );
      if ( !points2 || !mkcp ) {
        if ( points2 ) free ( points2 );
        if ( mkcp )    free ( mkcp );
        pkv_SetScratchMemTop ( sp );
        return false;
      }
      memcpy ( points1, points, (i+1)*dim*sizeof(double) );
      memcpy ( &points1[(i+2)*dim], &points[(i+1)*dim], (lkn-i)*dim*sizeof(double) );
      if ( dim == 2 )
        InterPoint2d ( (point2d*)&points1[2*i], (point2d*)&points1[2*i+4], param,
                       (point2d*)&points1[2*i+2] );
      else
        InterPoint3d ( (point3d*)&points[3*i], (point3d*)&points1[3*i+6], param,
                       (point3d*)&points1[3*i+3] );
      mbs_multiRollClosedBSCurvesd ( 1, dim, 1, lkn+3, NULL, 0, points1, i+1,
                                     NULL, points2 );
      if ( param1 ) {
        j = (int)(*param1);
        if ( j < i )
          *param1 += (double)(lkn-i);
        else {
          par = (*param1)-j;
          *param1 = (double)i+par/param;
        }
      }
      pkv_SetScratchMemTop ( points1 );
      free ( points );
      free ( eb->mkcp );
      eb->elem.points = points2;
      eb->mkcp = mkcp;
      memset ( mkcp, 0, (lkn+2)*sizeof(byte) );
      eb->elem.lastknot ++;
    }
    else if ( i > 0 && i < lkn ) { /* no point to add */
      points1 = pkv_GetScratchMemd ( dim*(lkn+1) );
      if ( !points1 )
        return false;
      memcpy ( points1, points, dim*(lkn+1)*sizeof(double) );
      memcpy ( points, &points1[dim*i], dim*(lkn-i+1)*sizeof(double) );
      memcpy ( &points[dim*(lkn-i+1)], &points1[dim], dim*i*sizeof(double) );
      if ( param1 )
        *param1 += (double)(lkn-i);
      pkv_SetScratchMemTop ( points1 );
    }  /* else nothing to do */
  }
  else {
    pkv_SetScratchMemTop ( sp );
    return false;
  }
  eb->elem.closing = 1;
  pkv_SetScratchMemTop ( sp );
  return true;
} /*_GOBSPTrDDivideClosedElem*/

boolean GeomObjectBSplinePatchDivideContourElem ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound, *nebound;
  mbs_polycurved    *nbound;
  int               nelem, ce, i;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  nebound = NULL;
  nbound = NULL;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  nelem = trpd->nelem;
  i = trpd->ndivpt;
  if ( !i )
    return false;
  ce = trpd->dptel[i-1];
  if ( ce < 0 || ce >= nelem )
    return false;
  if ( ebound[ce].elem.closing < 2 ) {
    nebound = malloc ( (nelem+1)*sizeof(BSP_TrPolycurve) );
    nbound = malloc ( (nelem+1)*sizeof(mbs_polycurved) );
    if ( !nebound || !nbound )
      goto failure;
    memcpy ( nebound, ebound, nelem*sizeof(BSP_TrPolycurve) );
    if ( !_GOBSPTrDDivideElem ( &nebound[ce], &nebound[nelem],
                                &nebound[nebound[ce].next],
                                ce, nelem, trpd->dptparam[i-1] ) )
      goto failure;
    free ( trpd->bound );
    free ( ebound );
    trpd->bound = nbound;
    trpd->ebound = nebound;
    trpd->nelem ++;
    _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
  }
  else {
    if ( !_GOBSPTrDDivideClosedElem ( &ebound[ce], trpd->dptparam[i-1], NULL ) )
      goto failure;
  }
  trpd->ndivpt = 0;
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( nbound )  free ( nbound );
  if ( nebound ) free ( nebound );
  return false;
} /*GeomObjectBSplinePatchDivideContourElem*/

boolean GeomObjectBSplinePatchJoinContourElem ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *ebound, *nebound;
  mbs_polycurved    *nbound;
  int               nelem, ce, ne;
  double            *knots0, *knots1, *knots, *points0, *points1, *points;
  byte              *mkcp;
  int               deg0, deg1, deg, dim0, dim1, dim, lkn0, lkn1, lkn;
  int               i, np;
  double            T;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  nebound = NULL;
  nbound = NULL;
  knots = points = NULL;
  mkcp = NULL;
  trpd = obj->trpd;
  ebound = trpd->ebound;
  nelem = trpd->nelem;
  ce = trpd->current_elem;
  if ( ce < 0 || ce >= nelem )
    return false;
  ne = ebound[ce].next;
  if ( ne < 0 || ne >= nelem )
    return false;
  if ( ne == ce ) {

    return false;
  }
  else {
    nebound = malloc ( (nelem-1)*sizeof(BSP_TrimmedDomain) );
    nbound = malloc ( (nelem-1)*sizeof(mbs_polycurved) );
    if ( !nebound || !nbound )
      goto failure;
    deg0 = ebound[ce].elem.degree;
    dim0 = ebound[ce].elem.cpdimen;
    lkn0 = ebound[ce].elem.lastknot;
    knots0 = ebound[ce].elem.knots;
    points0 = ebound[ce].elem.points;
    deg1 = ebound[ne].elem.degree;
    dim1 = ebound[ne].elem.cpdimen;
    lkn1 = ebound[ne].elem.lastknot;
    knots1 = ebound[ne].elem.knots;
    points1 = ebound[ne].elem.points;
    deg = max ( deg0, deg1 );
    dim = max ( dim0, dim1 );
    if ( deg == 1 ) {  /* join two polylines-knots will be discarded */
      if ( knots0 ) lkn0 -= 2;
      if ( knots1 ) lkn1 -= 2;
      lkn = lkn0+lkn1;
      np = lkn+1;
      points = malloc ( np*dim*sizeof(double) );
      mkcp = malloc ( np*sizeof(byte) );
      if ( !points || !mkcp )
        goto failure;
      if ( dim0 == dim )
        memcpy ( points, points0, lkn0*dim*sizeof(double) );
      else {
        for ( i = 0; i < lkn0-1; i++ )
          Point2to3d ( (point2d*)&points0[2*i], 1.0, (point3d*)&points[3*i] );
      }
      if ( dim1 == dim )
        memcpy ( &points[dim*lkn0], points1, (lkn1+1)*dim*sizeof(double) );
      else {
        for ( i = 0; i < lkn1; i++ )
          Point2to3d ( (point2d*)&points1[2*i], 1.0, (point3d*)&points[3*(lkn0+i)] );
      }
    }
    else {
      if ( deg > deg0 ) {
        if ( !GeomObjectBSplinePatchTrDSetElemDegree ( obj, deg,
                        &deg0, &lkn0, &knots0 ) )
          goto failure;
        points0 = ebound[ce].elem.points;
      }
      if ( deg > deg1 ) {
        trpd->current_elem = ne;
        if ( !GeomObjectBSplinePatchTrDSetElemDegree ( obj, deg,
                        &deg1, &lkn1, &knots1 ) )
          goto failure;
        points1 = ebound[ne].elem.points;
      }
      lkn = lkn0+lkn1-deg-1;
      np = lkn-deg;
      knots = malloc ( (lkn+1)*sizeof(double) );
      points = malloc ( np*dim*sizeof(double) );
      mkcp = malloc ( np*sizeof(byte) );
      if ( !knots || !points || !mkcp )
        goto failure;
      memcpy ( knots, knots0, lkn0*sizeof(double) );
      T = knots0[lkn0-deg]-knots1[deg];
      for ( i = deg+1; i <= lkn1; i++ )
        knots[lkn0-deg-1+i] = knots1[i] + T;
      if ( dim0 == dim )
        memcpy ( points, points0, (lkn0-deg)*dim*sizeof(double) );
      else {
        for ( i = 0; i < lkn0-deg; i++ )
          Point2to3d ( (point2d*)&points0[2*i], 1.0, (point3d*)&points[3*i] );
      }
      if ( dim1 == dim )
        memcpy ( &points[(lkn0-deg)*dim], &points1[dim],
                 (lkn1-deg-1)*dim*sizeof(double) );
      else {
        for ( i = 1; i < lkn1-deg; i++ )
          Point2to3d ( (point2d*)&points1[2*i], 1.0, (point3d*)&points[3*(lkn0+i-1)] );
      }
    }
    memcpy ( nebound, ebound, (nelem-1)*sizeof(BSP_TrimmedDomain) );
    free ( nebound[ce].mkcp );
    free ( nebound[ne].mkcp );
    free ( points0 );
    free ( points1 );
    if ( knots0 ) free ( knots0 );
    if ( knots1 ) free ( knots1 );
    if ( ce == nelem-1 ) {
      nebound[ne] = nebound[ce];
      nebound[ne].prev = ebound[ce].prev;
      nebound[ne].next = ebound[ne].next;
      ce = ne;
    }
    else if ( ne == nelem-1 ) {
      nebound[ce].next = ebound[ne].next;
    }
    else {
      nebound[ce].next = ebound[ne].next == nelem-1 ? ne : ebound[ne].next;
      nebound[ne] = nebound[nelem-1];
    }
    nebound[nebound[ce].prev].next = nebound[nebound[ce].next].prev = ce;
    nebound[ce].elem.knots = knots;
    nebound[ce].elem.points = points;
    nebound[ce].elem.lastknot = lkn;
    nebound[ce].elem.cpdimen = dim;
    nebound[ce].elem.degree = deg;
    nebound[ce].mkcp = mkcp;
    memset ( mkcp, 0, np*sizeof(byte) );
    trpd->ebound = nebound;
    trpd->bound = nbound;
    trpd->nelem = nelem-1;
    trpd->current_elem = ce;
    _PackTheBoundary ( nelem-1, nebound, nbound );
  }
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( nebound ) free ( nebound );
  if ( nbound )  free ( nbound );
  if ( points )  free ( points );
  if ( knots )   free ( knots );
  if ( mkcp )    free ( mkcp );
  return false;
} /*GeomObjectBSplinePatchJoinContourElem*/

static double _GetElemOriginParam ( mbs_polycurved *elem )
{
  if ( elem->knots )
    return elem->knots[elem->degree];
  else
    return 0.0;
} /*_GetElemOriginParam*/

static void _GetElemOrigin ( mbs_polycurved *elem, point2d *p )
{
  int    dim, deg, lkn;
  double *knots, *points;

  dim = elem->cpdimen;
  deg = elem->degree;
  lkn = elem->lastknot;
  knots = elem->knots;
  points = elem->points;
  if ( knots ) {
    if ( dim == 2 )
      mbs_deBoorC2d ( deg, lkn, knots, points, knots[deg], p );
    else
      mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[deg], p );
  }
  else {
    if ( dim == 2 )
      memcpy ( p, points, sizeof(point2d) );
    else
      Point3to2d ( (point3d*)points, p );
  }
} /*_GetElemOrigin*/

#ifdef ELEM_END
static void _GetElemEnd ( mbs_polycurved *elem, point2d *p )
{
  int    dim, deg, lkn;
  double *knots, *points;

  dim = elem->cpdimen;
  deg = elem->degree;
  lkn = elem->lastknot;
  knots = elem->knots;
  points = elem->points;
  if ( knots ) {
    if ( dim == 2 )
      mbs_deBoorC2d ( deg, lkn, knots, points, knots[lkn-deg], p );
    else
      mbs_deBoorC2Rd ( deg, lkn, knots, (point3d*)points, knots[lkn-deg], p );
  }
  else if ( deg > 1 ) {
    if ( dim == 2 )
      memcpy ( p, &points[2*deg], sizeof(point2d) );
    else
      Point3to2d ( (point3d*)&points[3*deg], p );
  }
  else {
    if ( dim == 2 )
      memcpy ( p, &points[2*lkn], sizeof(point2d) );
    else
      Point3to2d ( (point3d*)&points[3*lkn], p );
  }
} /*_GetElemEnd*/
#endif

static boolean _IsElemToDivide ( BSP_TrPolycurve *eb, BSP_TrPolycurve *next,
                                 double *param, boolean *end )
{
  int deg, lkn;

  *end = false;
  if ( eb->elem.closing == 2 )
    return false;
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  if ( eb->elem.knots ) {
    if ( *param >= eb->elem.knots[lkn-deg] ) {
      *param = _GetElemOriginParam ( &next->elem );
      *end = true;
      return false;
    }
    return *param > eb->elem.knots[deg];
  }
  else if ( eb->elem.degree > 1 ) {
    if ( *param >= 1.0 ) {
      *param = _GetElemOriginParam ( &next->elem );
      *end = true;
      return false;
    }
    return *param > 0.0;
  }
  else {
    if ( *param >= (double)lkn ) {
      *end = true;
      return false;
    }
    return *param > 0.0;
  }
} /*_IsElemToDivide*/

boolean GeomObjectBSplinePatchBridgeContour ( GO_BSplinePatch *obj )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb0, *eb1, *eb2, *eb3, *ebound, *nebound;
  mbs_polycurved    *nbound;
  int               nelem, nelem1, i0, i1, i2, i3;
  boolean           div0, div1, end;
  double            param0, param1, t, *line0, *line1;
  byte              *mkcp0, *mkcp1;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  if ( trpd->ndivpt < 2 )
    return false;
  ebound = trpd->ebound;
  nebound = NULL;
  nbound = NULL;
  line0 = line1 = NULL;
  mkcp0 = mkcp1 = NULL;
  nelem = trpd->nelem;
  i0 = trpd->dptel[0];
  i1 = trpd->dptel[1];
  if ( i0 < 0 || i0 >= nelem || i1 < 0 || i1 >= nelem )
    return false;
        /* determine the number of arcs to divide */
  if ( i0 == i1 && ebound[i0].elem.closing == 2 ) { 
    nelem1 = nelem+3;
    div0 = div1 = true;
    param0 = trpd->dptparam[0];
    param1 = trpd->dptparam[1];
  }
  else {
    nelem1 = nelem+2;
    param0 = trpd->dptparam[0];
    if ( (div0 = _IsElemToDivide ( &ebound[i0], &ebound[ebound[i0].next],
                                   &param0, &end )) )
      nelem1 ++;
    if ( end )
      i0 = ebound[i0].next;
    param1 = trpd->dptparam[1];
    if ( (div1 = _IsElemToDivide ( &ebound[i1], &ebound[ebound[i1].next],
                                   &param1, &end )) )
      nelem1 ++;
    if ( end )
      i1 = ebound[i1].next;
  }
  nebound = malloc ( nelem1*sizeof(BSP_TrPolycurve) );
  nbound = malloc ( nelem1*sizeof(mbs_polycurved) );
  line0 = malloc ( 4*sizeof(double) );
  line1 = malloc ( 4*sizeof(double) );
  mkcp0 = malloc ( 2*sizeof(byte) );
  mkcp1 = malloc ( 2*sizeof(byte) );
  if ( !nebound || !nbound || !line0 || !line1 || !mkcp0 || !mkcp1 )
    goto failure;
  memcpy ( nebound, ebound, nelem*sizeof(BSP_TrPolycurve) );
  eb0 = &nebound[i0];
  eb1 = &nebound[i1];
  if ( i0 == i1 ) {
    if ( param0 == param1 )
      goto failure;
    else if ( param0 < param1 )
      { t = param0;  param0 = param1;  param1 = t; }
  }
  if ( eb0->elem.closing == 2 ) {
    if ( !_GOBSPTrDDivideClosedElem ( eb0, param0,
                                      i0 == i1 ? &param1 : NULL ) )
      goto failure;
  }
  else if ( div0 ) {
    if ( !_GOBSPTrDDivideElem ( eb0, &nebound[nelem], &nebound[eb0->next],
                                i0, nelem, param0 ) )
      goto failure;
    nelem ++;
    i0 = eb0->next;
    eb0 = &nebound[i0];
  }
  if ( eb1->elem.closing == 2 ) {
    if ( !_GOBSPTrDDivideClosedElem ( eb1, param1, NULL ) )
      goto failure;
  }
  else if ( div1 ) {
    if ( !_GOBSPTrDDivideElem ( eb1, &nebound[nelem], &nebound[eb1->next],
                                i1, nelem, param1 ) )
      goto failure;
    nelem ++;
    i1 = eb1->next;
    eb1 = &nebound[i1];
  }
  i2 = nelem;
  i3 = nelem+1;
  assert ( i3+1 == nelem1 );
  eb2 = &nebound[i2];
  eb3 = &nebound[i3];
  eb2->elem.ident = 0;
  eb2->elem.cpdimen = 2;
  eb2->elem.degree = 1;
  eb2->elem.lastknot = 1;
  eb2->elem.knots = NULL;
  eb2->elem.points = line0;
  eb2->elem.closing = 0;
  eb2->savedpoints = NULL;
  eb2->mkcp = mkcp0;
  eb2->tag = 0;
  memset ( mkcp0, 0, 2*sizeof(byte) );
  eb3->elem.ident = 0;
  eb3->elem.cpdimen = 2;
  eb3->elem.degree = 1;
  eb3->elem.lastknot = 1;
  eb3->elem.knots = NULL;
  eb3->elem.points = line1;
  eb3->elem.closing = 0;
  eb3->savedpoints = NULL;
  eb3->mkcp = mkcp1;
  eb3->tag = 0;
  memset ( mkcp1, 0, 2*sizeof(byte) );
  eb2->prev = eb1->prev;
  nebound[eb2->prev].next = i2;
  eb3->prev = eb0->prev;
  nebound[eb3->prev].next = i3;
  eb2->next = i0;
  eb0->prev = i2;
  eb3->next = i1;
  eb1->prev = i3;
  _GetElemOrigin ( &eb1->elem, (point2d*)line0 );
  _GetElemOrigin ( &eb0->elem, (point2d*)line1 );
  memcpy ( &line0[2], line1, sizeof(point2d) );
  memcpy ( &line1[2], line0, sizeof(point2d) );

  free ( trpd->bound );
  free ( ebound );
  trpd->bound = nbound;
  trpd->ebound = nebound;
  trpd->nelem = nelem1;
  _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
  trpd->ndivpt = 0;
  obj->me.dlistmask = 0;
  return true;

failure:
  if ( nebound ) free ( nebound );
  if ( nbound )  free ( nbound );
  if ( line0 )   free ( line0 );
  if ( line1 )   free ( line1 );
  if ( mkcp0 )   free ( mkcp0 );
  if ( mkcp1 )   free ( mkcp1 );
  return false;
} /*GeomObjectBSplinePatchBridgeContour*/

boolean GeomObjectBSplinePatchTrDSplineElem ( GO_BSplinePatch *obj, boolean spline )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  double            *knots;
  int               lkn, deg, i;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  eb = &trpd->ebound[trpd->current_elem];
  deg = eb->elem.degree;
  if ( spline && !eb->elem.knots ) {
        /* add knots to a polyline or a Bezier curve */
    if ( deg > 1 ) {  /* a Bezier curve */
      knots = malloc ( 2*(deg+1)*sizeof(double) );
      if ( !knots )
        goto way_out;
      for ( i = 0; i <= deg; i++ )
        knots[i] = 0.0;
      for ( ; i <= 2*deg+1; i++ )
        knots[i] = 1.0;
      eb->elem.lastknot = 2*deg+1;
    }
    else {  /* a polyline */
      lkn = eb->elem.lastknot;
      knots = malloc ( (lkn+3)*sizeof(double) );
      if ( !knots )
        goto way_out;
      for ( i = 0; i <= lkn+2; i++ )
        knots[i] = (double)(i-1);
      eb->elem.lastknot = lkn+2;
    }
    eb->elem.knots = knots;
  }
  else if ( !spline && eb->elem.knots ) {
        /* convert a spline curve to a Bezier curve or a polyline */
    if ( deg > 1 ) {  /* a Bezier curve */

    }
    else {  /* a polyline */
      free ( eb->elem.knots );
      eb->elem.knots = NULL;
      eb->elem.lastknot -= 2;
    }
  }
way_out:
  _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );
  obj->me.dlistmask = 0;
  return eb->elem.knots != NULL;
} /*GeomObjectBSplinePatchTrDSplineElem*/

boolean GeomObjectBSplinePatchTrDRationalElem ( GO_BSplinePatch *obj,
                                                boolean rational )
{
  BSP_TrimmedDomain *trpd;
  BSP_TrPolycurve   *eb;
  double            *newpt;
  int               np, lkn, deg, i;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return false;
  if ( !obj->trimmed || !obj->trpd )
    return false;
  trpd = obj->trpd;
  eb = &trpd->ebound[trpd->current_elem];
  deg = eb->elem.degree;
  lkn = eb->elem.lastknot;
  if ( eb->elem.knots )
    np = lkn-deg;
  else if ( deg > 1 )
    np = deg+1;
  else
    np = lkn+1;
  if ( rational && eb->elem.cpdimen == 2 ) {  /* convert to rational */
    newpt = malloc ( 3*np*sizeof(double) );
    if ( !newpt )
      goto way_out;
    pkv_Selectd ( np, 2, 2, 3, eb->elem.points, newpt );
    for ( i = 0; i < np; i++ )
      newpt[3*i+2] = 1.0;
    eb->elem.cpdimen = 3;
    free ( eb->elem.points );
    eb->elem.points = newpt;
  }
  else if ( !rational && eb->elem.cpdimen == 3 ) {  /* convert to nonrational */
    newpt = malloc ( 2*np*sizeof(double) );
    if ( !newpt )
      goto way_out;
    for ( i = 0; i < np; i++ )
      Point3to2d ( (point3d*)&eb->elem.points[3*i], (point2d*)&newpt[2*i] );
    eb->elem.cpdimen = 2;
    if ( eb->next != trpd->current_elem ) {
      /* end points may need to be modified */

    }
    free ( eb->elem.points );
    eb->elem.points = newpt;
  }
  _PackTheBoundary ( trpd->nelem, trpd->ebound, trpd->bound );

way_out:
  obj->me.dlistmask = 0;
  return eb->elem.cpdimen == 3;
} /*GeomObjectBSplinePatchTrDRationalElem*/

/* ////////////////////////////////////////////////////////////////////////// */
static void _FlipXYPoints ( int dim, int npoints, double *points )
{
  int    i, k;
  double s;

  for ( i = k = 0;  i < npoints;  i++, k += dim )
    { s = points[k];  points[k] = points[k+1];  points[k+1] = s; }
} /*_FlipXYPoints*/

void GeomObjectBSplinePatchFlipTrimmedDomain ( GO_BSplinePatch *obj )
  
{
  BSP_TrimmedDomain *trpd;
  int               nelem, i;
  mbs_polycurved    *elem;

  if ( obj->me.obj_type != GO_BSPLINE_PATCH )
    return;
  if ( !obj->trimmed || !obj->trpd )
    return;
  trpd  = obj->trpd;
  elem  = trpd->bound;
  nelem = trpd->nelem;
  for ( i = 0; i < nelem; i++ ) {
    if ( elem[i].knots )  /* flip a B-spline curve */
      _FlipXYPoints ( elem[i].cpdimen, elem[i].lastknot-elem[i].degree,
                      elem[i].points );
    else if ( elem[i].degree > 1 ) /* flip a Bezier curve */
      _FlipXYPoints ( elem[i].cpdimen, elem[i].degree+1, elem[i].points );
    else /* flip a polyline */
      _FlipXYPoints ( elem[i].cpdimen, elem[i].lastknot+1, elem[i].points );
  }
  obj->me.dlistmask = 0;
} /*GeomObjectBSplinePatchFlipTrimmedDomain*/

