
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */ 
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2009, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

boolean bsf_ReadPointd ( pks_scanner *sc,
                         int maxcpdimen, double *point, int *cpdimen )
{
  int         i;
  pks_calcval x;

  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    return false;
  memset ( point, 0, maxcpdimen*sizeof(double) );
  if ( maxcpdimen == 4)
    point[3] = 1.0;
  sc->GetNextSymbol ( sc );
  for ( i = 0; i < maxcpdimen; i++ ) {
    if ( sc->nextsymbol == PKS_SYMB_RBRACE )
      break;
    if ( !pkscalc_Expression ( sc, &x ) )
      return false;
    point[i] = x.floatval;
    if ( sc->nextsymbol == PKS_SYMB_ERROR )
      return false;
    if ( sc->nextsymbol != PKS_SYMB_COMMA )
      break;
    sc->GetNextSymbol ( sc );
  }
  if ( sc->nextsymbol != PKS_SYMB_RBRACE )
    return false;
  *cpdimen = min ( 4, i+1 );
  sc->GetNextSymbol ( sc );
  return true;
} /*bsf_ReadPointd*/

int bsf_ReadPointsd ( pks_scanner *sc, int maxcpdimen, int maxnpoints,
                      double *points, int *cpdimen )
{
  int    ncp, spd0, spd1;
  double *p;

  ncp = spd0 = 0;
  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    return 0;
  sc->GetNextSymbol ( sc );
  for ( p = points; ; p += maxcpdimen ) {
    if ( ncp >= maxnpoints )
      return 0;
    if ( bsf_ReadPointd ( sc, maxcpdimen, p, &spd1 ) ) {
      ncp ++;
      spd0 = max ( spd0, spd1 );
    }
    else
      return 0;
    if ( sc->nextsymbol != PKS_SYMB_COMMA )
      break;
    sc->GetNextSymbol ( sc );
  }
  if ( sc->nextsymbol != PKS_SYMB_RBRACE )
    return 0;
  sc->GetNextSymbol ( sc );
  *cpdimen = spd0;
  return ncp;
} /*bsf_ReadPointsd*/

boolean bsf_ReadTrans3d ( pks_scanner *sc, trans3d *tr )
{
  int i, j;
  pks_calcval x;

  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    return false;
  memset ( tr, 0, sizeof(trans3d) );
  for ( i = 0; i < 3; i++ )
    for ( j = 0; j < 4; j++ ) {
      sc->GetNextSymbol ( sc );
      if ( !pkscalc_Expression ( sc, &x ) )
        return false;
      tr->U1.a[i][j] = x.floatval;
    }
  if ( sc->nextsymbol != PKS_SYMB_RBRACE )
    return false;
  sc->GetNextSymbol ( sc );
  return true;
} /*bsf_ReadTrans3d*/

static int _bsf_ReadMK ( pks_scanner *sc, int maxnk, unsigned int *mk )
{
  int np;
  int num;

  sc->GetNextSymbol ( sc );
  if ( sc->nextsymbol != PKS_SYMB_LBRACE )
    return 0;
  sc->GetNextSymbol ( sc );
  for ( np = 0; np < maxnk; ) {
    if ( !bsf_ReadIntNumber ( sc, &num ) )
      return 0;
    if ( mk )  /* NULL is allowed here */
      mk[np++] = num;
    else
      np ++;
    if ( sc->nextsymbol != PKS_SYMB_COMMA )
      break;
    sc->GetNextSymbol ( sc );
  }
  if ( sc->nextsymbol != PKS_SYMB_RBRACE )
    return 0;
  sc->GetNextSymbol ( sc );
  return np;
} /*_bsf_ReadMK*/

int bsf_ReadPointsMK ( pks_scanner *sc, int maxnpoints, unsigned int *mk )
{
  if ( sc->nextinteger != BSF_KEY_CPOINTSMK )
    return 0;
  return _bsf_ReadMK ( sc, maxnpoints, mk );
} /*bsf_ReadPointsMK*/

int bsf_ReadHEdgeMK ( pks_scanner *sc, int maxnhe, unsigned int *hemk )
{
  if ( sc->nextinteger != BSF_KEY_HALFEDGEMK )
    return 0;
  return _bsf_ReadMK ( sc, maxnhe, hemk );
} /*bsf_ReadHEdgeMK*/

int bsf_ReadFacetMK ( pks_scanner *sc, int maxnfac, unsigned int *fmk )
{
  if ( sc->nextinteger != BSF_KEY_FACETMK )
    return 0;
  return _bsf_ReadMK ( sc, maxnfac, fmk );
} /*bsf_ReadFacetMK*/

