
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2010, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */
/* changes:                                                                  */
/* 22.07.2013, R. Putanowicz - static pointers to application reading        */
/*   procedures replaced by pointers in the bsf_Reader structure passed      */
/*   by a parameter to the reading procedure.                                */
/* 24.07.2013, P. Kiciak - changes related with integration of the above     */
/*   with the package.                                                       */
/* 27.12.2013, P. Kiciak - adding procedures reading colours and cameras.    */
/* 26.05.2017, P. Kiciak - adding procedures reading kinematic linkages.     */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "egholed.h"
#include "bsmesh.h"
#include "pkscanner.h"
#include "bsfile.h"

#include "bsfprivate.h"

/* ////////////////////////////////////////////////////////////////////////// */
static boolean _bsf_ReadBCurve ( bsf_Reader *reader )
{
  void    *sp;
  char    *name;
  point4d *cp;
  int     degree, spdimen;
  int     ident;
  boolean rational;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( (reader->bc_maxdeg+1)*sizeof(point4d) );
  if ( !name || !cp )
    goto failure;
  ident = -1;
  if ( !bsf_ReadBezierCurve4d ( reader, reader->bc_maxdeg, &degree, cp,
                                &spdimen, &rational, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BezierCurveReader )
      reader->BezierCurveReader ( reader->userData, name, ident, degree, cp,
                                   spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBCurve*/

static boolean _bsf_ReadBPatch ( bsf_Reader *reader )
{
  void    *sp;
  char    *name;
  point4d *cp;
  int     udeg, vdeg, pitch, spdimen, maxdeg;
  int     ident;
  boolean rational;

  sp = pkv_GetScratchMemTop ();
  maxdeg = reader->bp_maxdeg;
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( (maxdeg+1)*(maxdeg+1)*sizeof(point4d) );
  if ( !name || !cp )
    goto failure;
   ident = -1;
  if ( !bsf_ReadBezierPatch4d ( reader, maxdeg, &udeg, &vdeg, &pitch, cp,
                                &spdimen, &rational, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BezierPatchReader )
      reader->BezierPatchReader ( reader->userData, name, ident, udeg, vdeg, 
                                   pitch, cp, spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBPatch*/

static boolean _bsf_ReadBSCurve ( bsf_Reader *reader )
{
  void    *sp;
  char    *name;
  point4d *cp;
  double  *knots;
  int     degree, lastknot, spdimen, maxdeg, maxlkn;
  int     ident;
  boolean rational, closed;

  sp = pkv_GetScratchMemTop ();
  maxlkn = reader->bsc_maxlkn;
  maxdeg = reader->bsc_maxdeg;
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( maxlkn*sizeof(point4d) );
  knots = pkv_GetScratchMemd ( maxlkn+1 );
  if ( !name || !cp || !knots )
    goto failure;
   ident = -1;
  if ( !bsf_ReadBSplineCurve4d ( reader, maxdeg, maxlkn, maxlkn,
                                 &degree, &lastknot, knots, &closed, cp,
                                 &spdimen, &rational, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BSplineCurveReader )
      reader->BSplineCurveReader ( reader->userData, name, ident, degree, lastknot,
                                    knots, closed, cp, spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSCurve*/

static boolean _bsf_ReadBSPatch ( bsf_Reader *reader )
{
  void    *sp;
  char    *name;
  point4d *cp;
  double  *knotsu, *knotsv;
  int     spdimen, udeg, vdeg, lastknotu, lastknotv, pitch, maxdeg, maxlkn;
  int     ident;
  boolean closed_u, closed_v, rational;

  sp = pkv_GetScratchMemTop ();
  maxlkn = reader->bsp_maxlkn;
  maxdeg = reader->bsp_maxdeg;
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  cp = pkv_GetScratchMem ( maxlkn*maxlkn*sizeof(point4d) );
  knotsu = pkv_GetScratchMemd ( maxlkn );
  knotsv = pkv_GetScratchMemd ( maxlkn );
  if ( !name || !cp || !knotsu || !knotsv )
    goto failure;
  ident = -1;
  if ( !bsf_ReadBSplinePatch4d ( reader, maxdeg, maxlkn, maxlkn*maxlkn,
                                 &udeg, &lastknotu, knotsu,
                                 &vdeg, &lastknotv, knotsv,
                                 &closed_u, &closed_v, &pitch, cp,
                                 &spdimen, &rational, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BSplinePatchReader )
      reader->BSplinePatchReader ( reader->userData,
               name, ident, udeg, lastknotu, knotsu, vdeg, lastknotv, knotsv,
               closed_u, closed_v, pitch, cp, spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSPatch*/

static boolean _bsf_ReadBSMesh ( bsf_Reader *reader )
{
  void        *sp;
  char        *name;
  BSMvertex   *mv;
  BSMhalfedge *mhe;
  BSMfacet    *mfac;
  int         *mvhei, *mfhei;
  point4d     *vc;
  int         nv, nhe, nfac, spdimen, degree, maxnv, maxnhe, maxnfac;
  int         ident;
  boolean     rational;

  sp = pkv_GetScratchMemTop ();
  maxnv = reader->bsm_maxnv;
  maxnhe = reader->bsm_maxnhe;
  maxnfac = reader->bsm_maxnfac;
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  mv = pkv_GetScratchMem ( maxnv*sizeof(BSMvertex) );
  mhe = pkv_GetScratchMem ( maxnhe*sizeof(BSMhalfedge) );
  mfac = pkv_GetScratchMem ( maxnfac*sizeof(BSMfacet) );
  vc = pkv_GetScratchMem ( maxnv*sizeof(point4d) );
  mvhei = pkv_GetScratchMemi ( maxnhe );
  mfhei = pkv_GetScratchMemi ( maxnhe );
  if ( !name || !mv || !mhe || !mfac || !vc || !mvhei || !mfhei )
    goto failure;
  ident = -1;
  if ( !bsf_ReadBSMesh4d ( reader, maxnv, maxnhe, maxnfac,
                           &degree, &nv, mv, mvhei, vc,
                           &nhe, mhe, &nfac, mfac, mfhei, &spdimen, &rational,
                           name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BSMeshReader )
      reader->BSMeshReader ( reader->userData, name, ident,
                              degree, nv, mv, mvhei, 
                              vc, nhe, mhe, nfac, mfac, mfhei,
                              spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSMesh*/

static boolean _bsf_ReadBSHole ( bsf_Reader *reader )
{
  void    *sp;
  int     hole_k, spdimen;
  char    *name;
  double  *knots;
  point2d *domcp;
  point4d *holecp;
  int     ident;
  boolean rational;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  knots = pkv_GetScratchMemd ( 11*GH_MAX_K );
  domcp = pkv_GetScratchMem ( (12*GH_MAX_K+1)*sizeof(point2d) );
  holecp = pkv_GetScratchMem ( (12*GH_MAX_K+1)*sizeof(point4d) );
  if ( !knots || !domcp || !holecp )
    goto failure;
  ident = -1;
  if ( !bsf_ReadBSplineHoled ( reader, GH_MAX_K, &hole_k, knots, domcp, holecp,
                               &spdimen, &rational, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->BSplineHoleReader )
      reader->BSplineHoleReader ( reader->userData, name, ident,
                                  hole_k, knots, 
                                  domcp, holecp, spdimen, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadBSHole*/

static boolean _bsf_ReadKLinkage ( bsf_Reader *reader )
{
  void         *sp;
  pkkl_linkage lkg;
  int          *objvert;
  char         *name;
  int          ident;

  sp = pkv_GetScratchMemTop ();
  memset ( &lkg, 0, sizeof(pkkl_linkage) );
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  objvert = pkv_GetScratchMem ( reader->klk_maxobjvert*sizeof(int) );
  lkg.obj  = pkv_GetScratchMem ( reader->klk_maxobj*sizeof(pkkl_geom_obj) );
  lkg.oref = pkv_GetScratchMem ( reader->klk_maxoref*sizeof(pkkl_obj_ref) );
  lkg.link = pkv_GetScratchMem ( reader->klk_maxlinks*sizeof(pkkl_link) );
  lkg.hj   = pkv_GetScratchMem ( 2*reader->klk_maxjoints*sizeof(pkkl_halfjoint) );
  lkg.artp = pkv_GetScratchMem ( 4*reader->klk_maxartp*sizeof(double) );
  if ( !objvert || !lkg.obj || !lkg.oref || !lkg.link || !lkg.hj || !lkg.artp )
    goto failure;
  lkg.prevartp = &lkg.artp[reader->klk_maxartp];
  lkg.artprange = &lkg.prevartp[reader->klk_maxartp];
  ident = -1;
  if ( !bsf_ReadKLinkage ( reader, &lkg, objvert, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->KLinkageReader )
      reader->KLinkageReader ( reader->userData, name, ident, &lkg );
  }

  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadKLinkage*/

static boolean _bsf_ReadPolyline ( bsf_Reader *reader )
{
  void    *sp;
  int     maxvert, nvert, spdimen;
  char    *name;
  point4d *vert;
  int     ident;
  boolean closed, rational;

  sp = pkv_GetScratchMemTop ();
  name = pkv_GetScratchMem ( BSF_MAX_NAME_LENGTH+1 );
  maxvert = reader->poly_maxvert;
  vert = pkv_GetScratchMem ( maxvert*sizeof(point4d) );
  if ( !name || !vert )
    goto failure;
  ident = -1;
  if ( !bsf_ReadPolyline4d ( reader, maxvert, &nvert, vert, &spdimen,
                             &rational, &closed, name, &ident ) )
    goto failure;
  if ( reader ) {
    if ( reader->PolylineReader )
      reader->PolylineReader ( reader->userData, name, ident,
                                nvert, vert, spdimen, closed, rational );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadPolyline*/

boolean _bsf_ReadCPMark ( bsf_Reader *reader, int maxnpoints )
{
  void         *sp;
  unsigned int *mk;
  int          nmk;

  sp = pkv_GetScratchMemTop ();
  mk = (unsigned int*)pkv_GetScratchMemi ( maxnpoints );
  if ( !mk )
    goto failure;
  memset ( mk, 0, maxnpoints*sizeof(int) );
  nmk = bsf_ReadPointsMK ( &reader->sc, maxnpoints, mk );
  if ( nmk <= 0 )
    goto failure;
  if ( reader ) {
    if ( reader->CPMarkReader )
      reader->CPMarkReader ( reader->userData, nmk, mk );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadCPMark*/

boolean _bsf_ReadHEdgeMark ( bsf_Reader *reader )
{
  void         *sp;
  unsigned int *mk;
  int          nmk;

  sp = pkv_GetScratchMemTop ();
  mk = (unsigned int*)pkv_GetScratchMemi ( reader->bsm_maxnhe );
  if ( !mk )
    goto failure;
  memset ( mk, 0, reader->bsm_maxnhe*sizeof(int) );
  nmk = bsf_ReadHEdgeMK ( &reader->sc, reader->bsm_maxnhe, mk );
  if ( nmk <= 0 )
    goto failure;
  if ( reader ) {
    if ( reader->HEdgeMarkReader )
      reader->HEdgeMarkReader ( reader->userData, nmk, mk );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadHEdgeMark*/

boolean _bsf_ReadFacetMark ( bsf_Reader *reader )
{
  void         *sp;
  unsigned int *mk;
  int          nmk;

  sp = pkv_GetScratchMemTop ();
  mk = (unsigned int*)pkv_GetScratchMemi ( reader->bsm_maxnfac );
  if ( !mk )
    goto failure;
  memset ( mk, 0, reader->bsm_maxnfac*sizeof(int) );
  nmk = bsf_ReadFacetMK ( &reader->sc, reader->bsm_maxnfac, mk );
  if ( nmk <= 0 )
    goto failure;
  if ( reader ) {
    if ( reader->FacetMarkReader )
      reader->FacetMarkReader ( reader->userData, nmk, mk );
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_bsf_ReadFacetMark*/

boolean _bsf_ReadColour ( bsf_Reader *reader )
{
  point3d colour;

  if ( !bsf_ReadColour ( reader, &colour ) )
    return false;
  if ( reader ) {
    if ( reader->ColourReader )
      reader->ColourReader ( reader->userData, &colour );
  }
  return true;
} /*_bsf_ReadColour*/

static boolean _bsf_ReadCamera ( bsf_Reader *reader )
{
  CameraRecd Camera;
  int        ident;

  ident = -1;
  if ( !bsf_ReadCamera ( reader, &Camera, &ident ) )
    return false;
  if ( reader ) {
    if ( reader->CameraReader )
      reader->CameraReader ( reader->userData, ident, &Camera );
  }
  return true;
} /*_bsf_ReadCamera*/

/* procedure of reading an entire file; for each item the application */
/* routine is called in order to enter the data into the application */
/* data structures */
boolean bsf_ReadBSFiled ( const char *filename, bsf_Reader *reader )
{
  void    *sp;
  boolean signal_end;
  int     obj_type;

  sp = pkv_GetScratchMemTop ();
  if ( !_bsf_OpenInputFile ( filename, reader ) )
    return false;
  reader->done = false;
  for (;;) {
    obj_type = BSF_NONE;
    signal_end = false;
    if ( reader->sc.nextsymbol == PKS_SYMB_EOF )
      goto finish;
    if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
         reader->sc.id_space != ID_SPACE_BSFILE )
      goto failure;
    if ( reader ) {
      if ( reader->BeginReader ) {
        switch ( reader->sc.nextinteger ) {
    case BSF_KEY_BCURVE:
          obj_type = BSF_BEZIER_CURVE;
          goto notify;
    case BSF_KEY_BPATCH:
          obj_type = BSF_BEZIER_PATCH;
          goto notify;
    case BSF_KEY_BSCURVE:
          obj_type = BSF_BSPLINE_CURVE;
          goto notify;
    case BSF_KEY_BSPATCH:
          obj_type = BSF_BSPLINE_PATCH;
          goto notify;
    case BSF_KEY_BSMESH:
          obj_type = BSF_BSPLINE_MESH;
          goto notify;
    case BSF_KEY_BSHOLE:
          obj_type = BSF_BSPLINE_HOLE;
          goto notify;
    case BSF_KEY_KLINKAGE:
          obj_type = BSF_KINEMATIC_LINKAGE;
          goto notify;
    case BSF_KEY_POLYLINE:
          obj_type = BSF_POLYLINE;
          goto notify;
    case BSF_KEY_CAMERA:
          obj_type = BSF_CAMERA;
notify:
          reader->BeginReader ( reader->userData, obj_type );
          signal_end = true;
          break;
    default:
          break;
        }
      }
    }

    switch ( reader->sc.nextinteger ) {
case BSF_KEY_BCURVE:
      if ( !_bsf_ReadBCurve ( reader ) )
        goto failure;
      break;

case BSF_KEY_BPATCH:
      if ( !_bsf_ReadBPatch ( reader ) )
        goto failure;
      break;

case BSF_KEY_BSCURVE:
      if ( !_bsf_ReadBSCurve ( reader ) )
        goto failure;
      break;

case BSF_KEY_BSPATCH:
      if ( !_bsf_ReadBSPatch ( reader ) )
        goto failure;
      break;

case BSF_KEY_BSMESH:
      if ( !_bsf_ReadBSMesh ( reader ) )
        goto failure;
      break;

case BSF_KEY_BSHOLE:
      if ( !_bsf_ReadBSHole ( reader ) )
        goto failure;
      break;

case BSF_KEY_KLINKAGE:
      if ( !_bsf_ReadKLinkage ( reader ) )
        goto failure;
      break;

case BSF_KEY_POLYLINE:
      if ( !_bsf_ReadPolyline ( reader ) )
        goto failure;
      break;

case BSF_KEY_COLOR:
case BSF_KEY_COLOUR:
      if ( !_bsf_ReadColour ( reader ) )
        goto failure;
      break;

case BSF_KEY_CAMERA:
      if ( !_bsf_ReadCamera ( reader ) )
        goto failure;
      break;

default:
      goto failure;
    }
    if ( signal_end && reader->EndReader )
      reader->EndReader ( reader->userData, obj_type, true );
    if ( reader->done )
      goto finish;
  }

finish:
  pks_ShutDownPreprocessor ( &reader->sc );
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  if ( signal_end && reader->EndReader )
    reader->EndReader ( reader->userData, obj_type, false );
  pks_ShutDownPreprocessor ( &reader->sc );
  pkv_SetScratchMemTop ( sp );
  return false;
} /*bsf_ReadBSFiled*/

void bsf_ClearAppReaders ( bsf_Reader *reader )
{
  if ( reader ) {
    reader->BeginReader        = NULL;
    reader->EndReader          = NULL;
    reader->BezierCurveReader  = NULL;
    reader->BSplineCurveReader = NULL;
    reader->BezierPatchReader  = NULL;
    reader->BSplinePatchReader = NULL;
    reader->BSMeshReader       = NULL;
    reader->BSplineHoleReader  = NULL;
    reader->KLinkageReader     = NULL;
    reader->PolylineReader     = NULL;
    reader->DepReader          = NULL;
    reader->TrimmedReader      = NULL;
    reader->CPMarkReader       = NULL;
    reader->HEdgeMarkReader    = NULL;
    reader->FacetMarkReader    = NULL;
    reader->CameraReader       = NULL;
    reader->ColourReader       = NULL;
    reader->userData           = NULL;
    reader->done = false;
        /* arbitrary default values */
    reader->bc_maxdeg      = 10;
    reader->bsc_maxdeg     = 10;
    reader->bsc_maxlkn     = 100;
    reader->bp_maxdeg      = 10;
    reader->bsp_maxdeg     = 10;
    reader->bsp_maxlkn     = 100;
    reader->bsm_maxdeg     = 10;
    reader->bsm_maxnv      = 1000;
    reader->bsm_maxnhe     = 2000;
    reader->bsm_maxnfac    = 1000;
    reader->poly_maxvert   = 10;
    reader->maxdep         = 10;
    reader->klk_maxobj     = 10;
    reader->klk_maxobjvert = 1000;
    reader->klk_maxoref    = 100;
    reader->klk_maxlinks   = 100;
    reader->klk_maxjoints  = 100;
    reader->klk_maxartp    = 100;
  }
} /*bsf_ClearAppReaders*/

void bsf_BeginReadingFuncd ( bsf_Reader *reader,
                             bsf_BeginRead_fptr BeginReader )
{
  reader->BeginReader = BeginReader;
} /*bsf_BeginReadingFuncd*/

void bsf_EndReadingFuncd ( bsf_Reader *reader,
                           bsf_EndRead_fptr EndReader )
{
  reader->EndReader = EndReader;
} /*bsf_EndReadingFuncd*/

void bsf_BC4ReadFuncd ( bsf_Reader *reader, bsf_BC_fptr BCReader,
                        int maxdeg )
{
  reader->BezierCurveReader = BCReader;
  reader->bc_maxdeg = maxdeg;
} /*bsf_BC4ReadFuncd*/

void bsf_BSC4ReadFuncd ( bsf_Reader *reader, bsf_BSC_fptr BSCReader,
                         int maxdeg, int maxlastknot )
{
  reader->BSplineCurveReader = BSCReader;
  reader->bsc_maxdeg = maxdeg;
  reader->bsc_maxlkn = maxlastknot;
} /*bsf_BSC4ReadFuncd*/

void bsf_BP4ReadFuncd ( bsf_Reader *reader, bsf_BP_fptr BPReader,
                        int maxdeg )
{
  reader->BezierPatchReader = BPReader;
  reader->bp_maxdeg = maxdeg;
} /*bsf_BP4ReadFuncd*/

void bsf_BSP4ReadFuncd ( bsf_Reader *reader, bsf_BSP_fptr BSPReader,
                         int maxdeg, int maxlastknot )
{
  reader->BSplinePatchReader = BSPReader;
  reader->bsp_maxdeg = maxdeg;
  reader->bsp_maxlkn = maxlastknot;
} /*bsf_BSP4ReadFuncd*/

void bsf_BSM4ReadFuncd ( bsf_Reader *reader, bsf_BSM_fptr BSMReader,
                         int maxdeg, int maxnv, int maxnhe, int maxnfac )
{
  reader->BSMeshReader = BSMReader;
  reader->bsm_maxdeg = maxdeg;
  reader->bsm_maxnv = maxnv;
  reader->bsm_maxnhe = maxnhe;
  reader->bsm_maxnfac = maxnfac;
} /*bsf_BSM4ReadFuncd*/

void bsf_BSH4ReadFuncd ( bsf_Reader *reader, bsf_BSH_fptr BSHReader )
{
  reader->BSplineHoleReader = BSHReader;
} /*bsf_BSH4ReadFuncd*/

void bsf_KLinkageReadFuncd ( bsf_Reader *reader, bsf_KLK_fptr KLReader,
                             int maxobj, int maxobjvert, int maxoref,
                             int maxlinks, int maxjoints, int maxartpar  )
{
  reader->KLinkageReader = KLReader;
  reader->klk_maxobj = maxobj;
  reader->klk_maxobjvert = maxobjvert;
  reader->klk_maxoref = maxoref;
  reader->klk_maxlinks = maxlinks;
  reader->klk_maxjoints = maxjoints;
  reader->klk_maxartp = maxartpar;
} /*bsf_KLinkageReadFuncd*/

void bsf_Polyline4ReadFuncd ( bsf_Reader *reader, bsf_polyline_fptr PReader,
                              int maxvert )
{
  reader->PolylineReader = PReader;
  reader->poly_maxvert = maxvert;
} /*bsf_Polyline4ReadFuncd*/

void bsf_DependencyReadFunc ( bsf_Reader *reader,
                              bsf_dependency_fptr DepReader, int maxdep )
{
  reader->DepReader = DepReader;
  reader->maxdep = maxdep;
} /*bsf_DependencyReadFunc*/

void bsf_TrimmedReadFuncd ( bsf_Reader *reader,
                            bsf_trimmed_fptr TrimmedReader )
{
  reader->TrimmedReader = TrimmedReader;
} /*bsf_TrimmedReadFuncd*/

void bsf_CPMarkReadFunc ( bsf_Reader *reader,
                          bsf_CPMark_fptr CPMarkReader )
{
  reader->CPMarkReader = CPMarkReader;
} /*bsf_CPMarkReadFunc*/

void bsf_HalfedgeMarkReadFunc ( bsf_Reader *reader,
                                bsf_HEMark_fptr HEdgeMarkReader )
{
  reader->HEdgeMarkReader = HEdgeMarkReader;
} /*bsf_HEdgeMarkReadFunc*/

void bsf_FacetMarkReadFunc ( bsf_Reader *reader,
                             bsf_FacetMark_fptr FacetMarkReader )
{
  reader->FacetMarkReader = FacetMarkReader;
} /*bsf_FacetMarkReadFunc*/

void bsf_CameraReadFuncd ( bsf_Reader *reader,
                           bsf_Camera_fptr CameraReader )
{
  reader->CameraReader = CameraReader;
} /*bsf_CameraReadFuncd*/

void bsf_ColourReadFuncd ( bsf_Reader *reader,
                           bsf_Colour_fptr ColourReader )
{
  reader->ColourReader = ColourReader;
} /*bsf_ColourReadFuncd*/

