
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2017, 2018                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "camerad.h"
#include "multibs.h"
#include "pkscanner.h"
#include "pklinkage.h"
#include "bsfile.h"

#include "bsfprivate.h"

static boolean _bsf_ReadKLObject ( bsf_Reader *reader, pkkl_linkage *lkg )
{
  int     objn, l;
  boolean _name, _filename, _dim, _nvert, _type, _trans;

  _name = _filename = _dim = _nvert = _type = _trans = false;
  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  objn = reader->sc.nextinteger;
  if ( objn < 0 || objn >= lkg->maxobj || lkg->obj[objn].type != PKKL_NONE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  pkkl_obj_init_stub ( &lkg->obj[objn], NULL );

  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
   case BSF_KEY_BSCURVE:
         if ( _type )
           goto failure;
         lkg->obj[objn].type = PKKL_BSPLINE_CURVE;
         reader->sc.GetNextSymbol ( &reader->sc );
         _type = true;
         if ( reader->sc.nextsymbol == PKS_SYMB_STRING )
           goto get_obj_name;
         break;

   case BSF_KEY_BSPATCH:
         if ( _type )
           goto failure;
         lkg->obj[objn].type = PKKL_BSPLINE_PATCH;
         reader->sc.GetNextSymbol ( &reader->sc );
         _type = true;
         if ( reader->sc.nextsymbol == PKS_SYMB_STRING )
           goto get_obj_name;
         break;

   case BSF_KEY_BSMESH:
         if ( _type )
           goto failure;
         lkg->obj[objn].type = PKKL_BSPLINE_MESH;
         reader->sc.GetNextSymbol ( &reader->sc );
         _type = true;
         if ( reader->sc.nextsymbol == PKS_SYMB_STRING )
           goto get_obj_name;
         break;

   case BSF_KEY_NAME:
         if ( _name )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         if ( reader->sc.nextsymbol != PKS_SYMB_STRING )
           goto failure;
get_obj_name:
         l = strlen ( reader->sc.nextname );
         if ( !(lkg->obj[objn].name = pkv_GetScratchMem ( (l+1)*sizeof(char) )) )
           goto failure;
         strcpy ( lkg->obj[objn].name, reader->sc.nextname );
         reader->sc.GetNextSymbol ( &reader->sc );
         _name = true;
         break;

   case BSF_KEY_FILENAME:
         if ( _filename )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         if ( reader->sc.nextsymbol != PKS_SYMB_STRING )
           goto failure;
         l = strlen ( reader->sc.nextname );
         lkg->obj[objn].filename = pkv_GetScratchMem ( l+1 );
         strcpy ( lkg->obj[objn].filename, reader->sc.nextname );
         reader->sc.GetNextSymbol ( &reader->sc );
         _filename = true;
         break;

   case BSF_KEY_DIM:
         if ( _dim )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
           goto failure;
         lkg->obj[objn].nvc = reader->sc.nextinteger;
         if ( lkg->obj[objn].nvc < 1 || lkg->obj[objn].nvc > 4 )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         _dim = true;
         break;

   case BSF_KEY_NUM_VERTICES:
         if ( _nvert )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
           goto failure;
         lkg->obj[objn].nvert = reader->sc.nextinteger;
         if ( lkg->obj[objn].nvert < 1 || lkg->obj[objn].nvert > reader->klk_maxobjvert )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         _nvert = true;
         break;

   case BSF_KEY_ETRANS:
         if ( _trans )
           goto failure;
         reader->sc.GetNextSymbol ( &reader->sc );
         if ( !bsf_ReadTrans3d ( &reader->sc, &lkg->obj[objn].Etr ) )
           goto failure;
         _trans = true;
         break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_RBRACE:
      lkg->nobj ++;
      if ( lkg->nobj > lkg->maxobj )
        goto failure;
      reader->sc.GetNextSymbol ( &reader->sc );
        /* assign NULL data pointers and stub procedures */
      return true;

default:
      goto failure;
    }
  }

failure:
  return false;
} /*_bsf_ReadKLObject*/

static boolean _bsf_ReadKLObjectRef ( bsf_Reader *reader, pkkl_linkage *lkg,
                                      int *objvert, int *vcnt )
{
  int refn, objn, cnt;

  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  refn = reader->sc.nextinteger;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  objn = reader->sc.nextinteger;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( objn < 0 || objn >= lkg->maxobj ||
       refn < 0 || refn >= lkg->maxorefs || lkg->oref[refn].on != -1 )
    goto failure;
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  lkg->oref[refn].on = objn;
  lkg->oref[refn].nv = 0;
  if ( reader->sc.nextsymbol == PKS_SYMB_RBRACE ) {
    reader->sc.GetNextSymbol ( &reader->sc );
    lkg->oref[refn].vn = NULL;
    lkg->norefs ++;
    return true;
  }
  cnt = *vcnt;
  lkg->oref[refn].vn = &objvert[cnt];
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_INTEGER:
      if ( cnt < reader->klk_maxobjvert ) {
        objvert[cnt++] = reader->sc.nextinteger;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_COMMA )
          reader->sc.GetNextSymbol ( &reader->sc );
        else if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
          goto failure;
      }
      else
        goto failure;
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->oref[refn].nv = cnt - *vcnt;
      *vcnt = cnt;
      lkg->norefs ++;
      return true;

default:
      goto failure;
    }
  }

failure:
  return false;
} /*_bsf_ReadKLObjectRef*/

static boolean _bsf_ReadKLLink ( bsf_Reader *reader, pkkl_linkage *lkg )
{
  int     lkn, orefn, l;
  boolean _name;

  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  lkn = reader->sc.nextinteger;
  if ( lkn < 0 || lkn >= lkg->maxlinks ||
       lkg->link[lkn].tag != 0 )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  _name = false;
  lkg->link[lkn].name = NULL;
  IdentTrans3d ( &lkg->link[lkn].Mtr );
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol != PKS_SYMB_STRING )
          goto failure;
        l = strlen ( reader->sc.nextname );
        if ( !(lkg->link[lkn].name = pkv_GetScratchMem ( (l+1)*sizeof(char) )) )
          goto failure;
        strcpy ( lkg->link[lkn].name, reader->sc.nextname );
        reader->sc.GetNextSymbol ( &reader->sc );
        _name = true;
        break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_INTEGER:  /* object reference number */
      orefn = reader->sc.nextinteger;
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->oref[orefn].nextr = lkg->link[lkn].fref;
      lkg->link[lkn].fref = orefn;
      if ( reader->sc.nextsymbol == PKS_SYMB_COMMA )
        reader->sc.GetNextSymbol ( &reader->sc );
      else if ( reader->sc.nextsymbol != PKS_SYMB_RBRACE )
        goto failure;
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->link[lkn].tag = 1;
      lkg->nlinks ++;
      return true;

default:
      goto failure;
    }
  }

failure:
  return false;
} /*_bsf_ReadKLLink*/

static boolean _bsf_ReadKLJoint ( bsf_Reader *reader, pkkl_linkage *lkg )
{
  int     jn, l0, l1, art;
  boolean _ftr, _btr;

  jn = lkg->nhj;
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  lkg->hj[jn].l0 = l0 = reader->sc.nextinteger;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  lkg->hj[jn].l1 = l1 = reader->sc.nextinteger;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( l0 < 0 || l0 >= lkg->maxlinks ||
       l1 < 0 || l1 >= lkg->maxlinks || l0 == l1 )
    goto failure;
  switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
    if ( reader->sc.id_space != ID_SPACE_BSFILE )
      goto failure;
    switch ( reader->sc.nextinteger ) {
  case BSF_KEY_ROTATION_V:    art = PKKL_ART_ROT_V;     break;
  case BSF_KEY_ROTATION_X:    art = PKKL_ART_ROT_X;     break;
  case BSF_KEY_ROTATION_Y:    art = PKKL_ART_ROT_Y;     break;
  case BSF_KEY_ROTATION_Z:    art = PKKL_ART_ROT_Z;     break;
  case BSF_KEY_SCALING_X:     art = PKKL_ART_SCALE_X;   break;
  case BSF_KEY_SCALING_XYZ:   art = PKKL_ART_SCALE_XYZ; break;
  case BSF_KEY_SCALING_Y:     art = PKKL_ART_SCALE_Y;   break;
  case BSF_KEY_SCALING_Z:     art = PKKL_ART_SCALE_Z;   break;
  case BSF_KEY_TRANSLATION_V: art = PKKL_ART_TRANS_V;   break;
  case BSF_KEY_TRANSLATION_X: art = PKKL_ART_TRANS_X;   break;
  case BSF_KEY_TRANSLATION_Y: art = PKKL_ART_TRANS_Y;   break;
  case BSF_KEY_TRANSLATION_Z: art = PKKL_ART_TRANS_Z;   break;
  default: goto failure;
    }
    break;
default: goto failure;
  }
  lkg->hj[jn].art = art;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_INTEGER )
    goto failure;
  lkg->hj[jn].pnum = reader->sc.nextinteger;
  reader->sc.GetNextSymbol ( &reader->sc );
  _ftr = _btr = false;
  IdentTrans3d ( &lkg->hj[jn].Ftr );
  lkg->hj[jn].Btr = lkg->hj[jn+1].Ftr = lkg->hj[jn+1].Btr = lkg->hj[jn].Ftr;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_FTRANS:
        if ( _ftr )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadTrans3d ( &reader->sc, &lkg->hj[jn].Ftr ) )
          goto failure;
        _ftr = true;
        break;

  case BSF_KEY_BTRANS:
        if ( _btr )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !bsf_ReadTrans3d ( &reader->sc, &lkg->hj[jn].Btr ) )
          goto failure;
        _btr = true;
        break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->hj[jn].otherhalf = jn+1;
      lkg->hj[jn+1].otherhalf = jn;
      lkg->hj[jn+1].l0 = lkg->hj[jn].l1;
      lkg->hj[jn+1].l1 = lkg->hj[jn].l0;
      lkg->hj[jn+1].pnum = -1;
      lkg->hj[jn+1].art = PKKL_ART_OTHERHALF;
      lkg->hj[jn+1].Ftr = lkg->hj[jn].Btr;  InvertTrans3d ( &lkg->hj[jn+1].Ftr );
      lkg->hj[jn+1].Btr = lkg->hj[jn].Ftr;  InvertTrans3d ( &lkg->hj[jn+1].Btr );
      lkg->hj[jn].nexthj = lkg->link[l0].fhj;
      lkg->link[l0].fhj = jn;
      lkg->hj[jn+1].nexthj = lkg->link[l1].fhj;
      lkg->link[l1].fhj = jn+1;
      lkg->nhj += 2;
      return true;

default:
      goto failure;
    }
  }

failure:
  return false;
} /*_bsf_ReadKLJoint*/

static boolean _bsf_ReadKLArtParam ( bsf_Reader *reader, pkkl_linkage *lkg )
{
  int i;

  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    return false;
  reader->sc.GetNextSymbol ( &reader->sc );
  for ( i = 0; i < lkg->maxartpar; ) {
    if ( !bsf_ReadDoubleNumber ( &reader->sc, &lkg->artp[i] ) )
      return false;
    i++;
    switch ( reader->sc.nextsymbol ) {
  case PKS_SYMB_SEMICOLON:
      if ( i >= lkg->maxartpar )
        return false;
      reader->sc.GetNextSymbol ( &reader->sc );
      break;

  case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->nartpar = i;
      return true;

  default:
      return false;
    }
  } 
  return false;
} /*_bsf_ReadKLArtParam*/

static boolean _bsf_ReadKLArtParamRange ( bsf_Reader *reader, pkkl_linkage *lkg )
{
  int i;

  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    return false;
  reader->sc.GetNextSymbol ( &reader->sc );
  for ( i = 0; i < 2*lkg->maxartpar; ) {
    if ( !bsf_ReadDoubleNumber ( &reader->sc, &lkg->artprange[i] ) )
      return false;
    i++;
    if ( reader->sc.nextsymbol != PKS_SYMB_COMMA )
      return false;
    reader->sc.GetNextSymbol ( &reader->sc );
    if ( !bsf_ReadDoubleNumber ( &reader->sc, &lkg->artprange[i] ) )
      return false;
    i++;
    switch ( reader->sc.nextsymbol ) {
  case PKS_SYMB_SEMICOLON:
      if ( i >= 2*lkg->maxartpar )
        return false;
      reader->sc.GetNextSymbol ( &reader->sc );
      break;

  case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      lkg->nartpar = i;
      return true;

  default:
      return false;
    }
  } 
  return false;
} /*_bsf_ReadKLArtParamRange*/

boolean bsf_ReadKLinkage ( bsf_Reader *reader, pkkl_linkage *lkg,
                           int *objvert, char *name, int *ident )
{
  void    *sp;
  boolean _name, _nobj, _norefs, _nlinks, _njoints,
          _nartp, _artp, _artprange, _id, _root;
  int     vcnt;
  int     i, n;

  sp = pkv_GetScratchMemTop ();
        /* skip the initial greetings */
  if ( reader->sc.nextsymbol != PKS_SYMB_IDENT ||
       reader->sc.id_space != ID_SPACE_BSFILE ||
       reader->sc.nextinteger != BSF_KEY_KLINKAGE )  
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );
  if ( reader->sc.nextsymbol != PKS_SYMB_LBRACE )
    goto failure;
  reader->sc.GetNextSymbol ( &reader->sc );

  _name = _nobj = _norefs = _nlinks = _njoints = _nartp =
  _artp = _artprange = _id = _root = false;
  if ( name )
    *name = 0;
  vcnt = 0;
  for (;;) {
    switch ( reader->sc.nextsymbol ) {
case PKS_SYMB_IDENT:
      if ( reader->sc.id_space != ID_SPACE_BSFILE )
        goto failure;
      switch ( reader->sc.nextinteger ) {
  case BSF_KEY_NAME:
        if ( _name )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_STRING ) {
          strcpy ( name, reader->sc.nextname );
          reader->sc.GetNextSymbol ( &reader->sc );
        }
        else
          goto failure;
        _name = true;
        break;

  case BSF_KEY_IDENT:
        if ( _id )
          goto failure;
        if ( !bsf_ReadIdent ( &reader->sc, ident ) )
          goto failure;
        _id = true;
        break;

  case BSF_KEY_ART_PARAM:
        if ( _artp || !_nartp )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLArtParam ( reader, lkg ) )
          goto failure;
        _artp = true;
        break;

  case BSF_KEY_ART_PARAM_RANGE:
        if ( _artprange || !_nartp )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLArtParamRange ( reader, lkg ) )
          goto failure;
        _artprange = true;
        break;

  case BSF_KEY_NUM_ART_PAR:
        if ( _nartp )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->maxartpar = n = reader->sc.nextinteger;
          if ( n > reader->klk_maxartp )
            goto failure;
          lkg->nartpar = 0;
          reader->sc.GetNextSymbol ( &reader->sc );
          memset ( lkg->artp, 0, n*sizeof(double) );
          memset ( lkg->prevartp, 0, n*sizeof(double) );
          memset ( lkg->artprange, 0, 2*n*sizeof(double) );
          for ( i = 0; i < n; i++ )
            lkg->artprange[2*i] = 1.0;
          _nartp = true;
        }
        else
          goto failure;
        break;

  case BSF_KEY_NUM_JOINTS:
        if ( _njoints )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->maxnhj = n = 2*reader->sc.nextinteger;
          if ( n > 2*reader->klk_maxjoints )
            goto failure;
          for ( i = 0; i < n; i++ )
            lkg->hj[i].l0 = lkg->hj[i].otherhalf = lkg->hj[i].nexthj =
                lkg->hj[i].pnum = -1;
          lkg->nhj = 0;
          reader->sc.GetNextSymbol ( &reader->sc );
          _njoints = true;
        }
        else
          goto failure;
        break;

  case BSF_KEY_NUM_LINKS:
        if ( _nlinks )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->maxlinks = n = reader->sc.nextinteger;
          if ( n > reader->klk_maxlinks )
            goto failure;
          for ( i = 0; i < n; i++ ) {
            lkg->link[i].fref = lkg->link[i].fhj = -1;
            lkg->link[i].tag = 0;
          }
          lkg->nlinks = 0;
          reader->sc.GetNextSymbol ( &reader->sc );
          _nlinks = true;
        }
        else
          goto failure;
        break;

  case BSF_KEY_NUM_OBJ_REF:
        if ( _norefs )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->maxorefs = n = reader->sc.nextinteger;
          if ( n > reader->klk_maxoref )
            goto failure;
          for ( i = 0; i < n; i++ )
            lkg->oref[i].on = lkg->oref[i].nextr = -1;
          lkg->norefs = 0;
          reader->sc.GetNextSymbol ( &reader->sc );
          _norefs = true;
        }
        else
          goto failure;
        break;

  case BSF_KEY_NUM_OBJECTS:
        if ( _nobj )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->maxobj = n = reader->sc.nextinteger;
          if ( n > reader->klk_maxobj )
            goto failure;
          for ( i = 0; i < n; i++ )
            lkg->obj[i].type = PKKL_NONE;
          lkg->nobj = 0;
          reader->sc.GetNextSymbol ( &reader->sc );
          _nobj = true;
        }
        else
          goto failure;
        break;

  case BSF_KEY_OBJECT_REF:
        if ( !_norefs )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLObjectRef ( reader, lkg, objvert, &vcnt ) )
          goto failure;
        break;

  case BSF_KEY_OBJECT:
        if ( !_nobj )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLObject ( reader, lkg ) )
          goto failure;
        break;

  case BSF_KEY_LINK:
        if ( !_nlinks )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLLink ( reader, lkg ) )
          goto failure;
        break;

  case BSF_KEY_JOINT:
        if ( !_njoints )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( !_bsf_ReadKLJoint ( reader, lkg ) )
          goto failure;
        break;

  case BSF_KEY_ROOT:
        if ( _root )
          goto failure;
        reader->sc.GetNextSymbol ( &reader->sc );
        if ( reader->sc.nextsymbol == PKS_SYMB_INTEGER ) {
          lkg->current_root = reader->sc.nextinteger;
          reader->sc.GetNextSymbol ( &reader->sc );
          _nobj = true;
        }
        else
          goto failure;
        break;

  default:
        goto failure;
      }
      break;

case PKS_SYMB_RBRACE:
      reader->sc.GetNextSymbol ( &reader->sc );
      for ( i = 0; i < lkg->norefs; i++ )
        if ( !lkg->oref[i].vn )
          lkg->oref[i].nv = lkg->obj[lkg->oref[i].on].nvert;
        /* check if everything has been read in */
      if ( lkg->nobj < lkg->maxobj )
{
printf ( "ERROR: too little objects\n" );
        goto failure;
}
      if ( lkg->norefs < lkg->maxorefs )
{
printf ( "ERROR: too little object references\n" );
        goto failure;
}
      if ( lkg->nlinks < lkg->maxlinks )
{
printf ( "ERROR: too little links\n" );
        goto failure;
}
      if ( lkg->nhj < lkg->maxnhj )
{
printf ( "ERROR: too little joints\n" );
        goto failure;
}
      if ( !pkkl_LinkageValid ( lkg ) )
{
printf ( "ERROR: invalid linkage read in\n" );
        goto failure;
}
      pkv_SetScratchMemTop ( sp );
      return true;

default:
      goto failure;
    }
  }

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*bsf_ReadKLinkage*/

