
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <limits.h>
#include <string.h>
#include <math.h>

#include "pkvaria.h"
#include "pknum.h"
#include "pkgeom.h"
#include "multibs.h"

#define EPS 1.0e-6

boolean mbs_FindBC2ClosestPointf ( int degree, const point2f *cp,
                                   const point2f *q,
                                   float *t, point2f *p, float *dist )
{
  int      n, i, j, nz;
  float    *c, *z, d, dmin;
  vector2f *a, *b, v;
  point2f  s;
  
  n = 2*degree-1;
  c = pkv_GetScratchMemf ( 6*degree+3 );
  if ( !c )
    return false;
  a = (vector2f*)&c[n+1];
  b = &a[degree+1];
        /* compute the polynomial coefficients */
  for ( i = 0; i <= degree; i++ )
    SubtractPoints2f ( q, &cp[i], &a[i] );
  dmin = DotProduct2f ( a, a );  *t = 0.0;  *p = cp[0];
  d = DotProduct2f ( &a[degree], &a[degree] );
  if ( d < dmin ) { dmin = d;  *t = 1.0; *p = cp[degree]; }
  mbs_multiBezScalef ( degree, 1, 1, 2, 0, &a[0].x );
  for ( i = 0; i < degree; i++ )
    SubtractPoints2f ( &cp[i+1], &cp[i], &b[i] );
  mbs_multiBezScalef ( degree-1, 1, 1, 2, 0, &b[0].x );
  memset ( c, 0, (n+1)*sizeof(float) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < degree; j++ )
      c[i+j] += DotProduct2f ( &a[i], &b[j] );
  mbs_multiBezUnscalef ( n, 1, 1, 1, 0, c );
  z = &a[0].x;
        /* find the closest point */
  if ( mbs_FindPolynomialZerosf ( n, c, &nz, z, EPS ) ) {
    for ( i = 0; i < nz; i++ ) {
      mbs_BCHornerC2f ( degree, cp, z[i], &s );
      SubtractPoints2f ( q, &s, &v );
      d = DotProduct2f ( &v, &v );
      if ( d < dmin )
        { dmin = d;  *t = z[i];  *p = s; }
    }
  }
  *dist = sqrt ( dmin );
  pkv_SetScratchMemTop ( c );
  return true;
} /*mbs_FindBC2ClosestPointf*/

boolean mbs_FindBS2ClosestPointf ( int degree, int lkn, const float *knots,
                                   const point2f *cpoints,
                                   const point2f *q,
                                   float *t, point2f *p, float *dist )
{
  void    *sp;
  int     ku, lknu;
  float   *kn, s, d;
  point2f *bcp, r;
  int     i;

  sp = pkv_GetScratchMemTop ();
  ku = mbs_NumKnotIntervalsf ( degree, lkn, knots );
  lknu = (degree+1)*(ku+1)-1;
  kn = pkv_GetScratchMemf ( lknu+1 );
  bcp = pkv_GetScratchMem ( (lknu-degree)*sizeof(point2f) );
  if ( !kn || !bcp )
    goto failure;
  if ( !mbs_BSToBezC2f ( degree, lkn, knots, cpoints, &ku, &lknu, kn, bcp ) )
    goto failure;
  if ( !mbs_FindBC2ClosestPointf ( degree, bcp, q, &s, p, dist ) )
    goto failure;
  *t = kn[degree] + s*(kn[degree+1]-kn[degree]);
  for ( i = 1; i < ku; i++ ) {
    if ( !mbs_FindBC2ClosestPointf ( degree, &bcp[i*(degree+1)], q,
                                     &s, &r, &d ) )
      goto failure;
    if ( d < *dist ) {
      *t = kn[(i+1)*(degree+1)-1] + s*(kn[(i+1)*(degree+1)]-kn[(i+1)*(degree+1)-1]);
      *p = r;
      *dist = d;
    }
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
} /*mbs_FindBS2ClosestPointf*/

boolean mbs_FindBC2RClosestPointf ( int degree, const point3f *cp,
                                    const point2f *q,
                                    float *t, point2f *p, float *dist )
{
  int      n, i, j, nz;
  float    *c, *z, d, dmin;
  vector2f *a, *e, v;
  vector3f *r, *b;
  point2f  s;
  
  n = 3*degree-1;
  c = pkv_GetScratchMemf ( 15*degree+5 );
  if ( !c )
    return false;
  a = (vector2f*)&c[n+1];
  b = (vector3f*)&a[degree+1];
  r = &b[degree];
  e = (vector2f*)&r[degree+1];
        /* compute the polynomial coefficients */
  memcpy ( r, cp, (degree+1)*sizeof(point3f) );
  for ( i = 0; i <= degree; i++ )
    SetVector2f ( &a[i], cp[i].z*q->x-cp[i].x, cp[i].z*q->y-cp[i].y );
  mbs_multiBezScalef ( degree, 1, 1, 3, 0, &r[0].x );
  mbs_multiBezScalef ( degree, 1, 1, 2, 0, &a[0].x );
  for ( i = 0; i < degree; i++ )
    SubtractPoints3f ( &cp[i+1], &cp[i], &b[i] );
  mbs_multiBezScalef ( degree-1, 1, 1, 3, 0, &b[0].x );
  memset ( e, 0, 4*degree*sizeof(float) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < degree; j++ ) {
      e[i+j].x += r[i].z*b[j].x-r[i].x*b[j].z;
      e[i+j].y += r[i].z*b[j].y-r[i].y*b[j].z;
    }
  memset ( c, 0, (n+1)*sizeof(float) );
  for ( i = 0; i <= degree; i++ )
    for ( j = 0; j < 2*degree; j++ )
      c[i+j] += DotProduct2f ( &a[i], &e[j] );
  mbs_multiBezUnscalef ( n, 1, 1, 1, 0, c );
  z = &a[0].x;
        /* find the closest point */
  Point3to2f ( &cp[0], &s );
  SubtractPoints2f ( q, &s, &v );
  dmin = DotProduct2f ( &v, &v );  *t = 0.0;  *p = s;
  Point3to2f ( &cp[degree], &s );
  SubtractPoints2f ( q, &s, &v );
  d = DotProduct2f ( &v, &v );
  if ( d < dmin )
    { dmin = d;  *t = 1.0;  *p = s; }
  if ( mbs_FindPolynomialZerosf ( n, c, &nz, z, EPS ) ) {
    for ( i = 0; i < nz; i++ ) {
      mbs_BCHornerC2Rf ( degree, cp, z[i], &s );
      SubtractPoints2f ( q, &s, &v );
      d = DotProduct2f ( &v, &v );
      if ( d < dmin )
        { dmin = d;  *t = z[i];  *p = s; }
    }
  }
  *dist = sqrt ( dmin );
  pkv_SetScratchMemTop ( c );
  return true;
} /*mbs_FindBC2RClosestPointf*/

boolean mbs_FindBS2RClosestPointf ( int degree, int lkn, const float *knots,
                                    const point3f *cpoints,
                                    const point2f *q,
                                    float *t, point2f *p, float *dist )
{
  void    *sp;
  int     ku, lknu;
  float   *kn, s, d;
  point3f *bcp;
  point2f r;
  int     i;

  sp = pkv_GetScratchMemTop ();
  ku = mbs_NumKnotIntervalsf ( degree, lkn, knots );
  lknu = (degree+1)*(ku+1)-1;
  kn = pkv_GetScratchMemf ( lknu+1 );
  bcp = pkv_GetScratchMem ( (lknu-degree)*sizeof(point3f) );
  if ( !kn || !bcp )
    goto failure;
  if ( !mbs_BSToBezC3f ( degree, lkn, knots, cpoints, &ku, &lknu, kn, bcp ) )
    goto failure;
  if ( !mbs_FindBC2RClosestPointf ( degree, bcp, q, &s, p, dist ) )
    goto failure;
  *t = kn[degree] + s*(kn[degree+1]-kn[degree]);
  for ( i = 1; i < ku; i++ ) {
    if ( !mbs_FindBC2RClosestPointf ( degree, &bcp[i*(degree+1)], q,
                                      &s, &r, &d ) )
      goto failure;
    if ( d < *dist ) {
      *t = kn[(i+1)*(degree+1)-1] + s*(kn[(i+1)*(degree+1)]-kn[(i+1)*(degree+1)-1]);
      *p = r;
      *dist = d;
    }
  }
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  return false;
}  /*mbs_FindBS2RClosestPointf*/

