
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015                                  */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */
/* This file was originally written by Przemyslaw Kiciak                     */
/* and modified by Radoslaw Piorkowski, who implemented an enhanced          */
/* algorithm of building a binary tree of objects (recursive, using buckets) */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include <assert.h>

#include "pkvaria.h"
#include "pkvthreads.h"
#include "pknum.h"
#include "pkgeom.h"
#include "multibs.h"
#include "camera.h"
#include "raybez.h"
#include "pkrender.h"

#include "pkrenderprivate.h"

#define BUCKET_COUNT 27

/* ///////////////////////////////////////////////////////////////////////// */
rendertnode* NewNode ( int k, rendertnode *left, rendertnode *right )
{
  rendertnode *node;

  node = malloc ( sizeof(rendertnode) );
  if ( !node )
    exit ( 1 );
  node->left = left;
  node->right = right;
  node->k = k;
  return node;
} /*NewNode*/

rendertnode* NewLeafNode ( pkRenderer *rend, int k, 
                           rendertnode *left, rendertnode *right )
{
  rendertnode *node;

  node = NewNode ( k, left, right );

  switch ( rend->obj_tab[k].type ) {
case obj_TRIANGLE:
    memcpy ( &node->bbox, &rend->obj_tab[k].triang.trdata->bbox, sizeof(Box3d) );
    break;
case obj_BSPATCH:
    memcpy ( &node->bbox, &rend->obj_tab[k].bsp.ptree->root->bbox, sizeof(Box3d) );
    break;
case obj_RBSPATCH:
    memcpy ( &node->bbox, &rend->obj_tab[k].rbsp.ptree->root->bbox, sizeof(Box3d) );
    break;
case obj_BEZCURVE:
    memcpy ( &node->bbox, &rend->obj_tab[k].bezc.ctree->root->bbox, sizeof(Box3d) );
    break;
case obj_RBEZCURVE:
    memcpy ( &node->bbox, &rend->obj_tab[k].rbezc.ctree->root->bbox, sizeof(Box3d) );
    break;
default:
    free ( node );
    return NULL;
  }
  return node;
} /*NewLeafNode*/

rendertnode* NewInnerNode ( rendertnode *left, rendertnode *right )
{
  rendertnode *node;

  node = NewNode ( TNODE_INNER, left, right );
  rbez_FindSumBBoxd ( &left->bbox, &right->bbox, &node->bbox );
  return node;
} /*NewInnerNode*/

double BoxDiameter ( Box3d *box )
{
  double a, b, c;

  a = box->x1 - box->x0;
  b = box->y1 - box->y0;
  c = box->z1 - box->z0;
  return a*a + b*b + c*c;
} /*BoxDiameter*/

boolean CompNodePrio ( void *node1, void *node2 )
{
  double d1, d2;

  d1 = BoxDiameter ( &((rendertnode*)node1)->bbox );
  d2 = BoxDiameter ( &((rendertnode*)node2)->bbox );
  return d1 < d2;
} /*CompNodePrio*/

void ComputeSeparator ( Box3d *box, Box3d *sep )
{
  sep->x0 = box->x0 * TAU + box->x1 * (1.0-TAU);
  sep->y0 = box->y0 * TAU + box->y1 * (1.0-TAU);
  sep->z0 = box->z0 * TAU + box->z1 * (1.0-TAU);
  sep->x1 = box->x0 * (1.0-TAU) + box->x1 * TAU;
  sep->y1 = box->y0 * (1.0-TAU) + box->y1 * TAU;
  sep->z1 = box->z0 * (1.0-TAU) + box->z1 * TAU;
} /*ComputeSeparator*/

/* Function ChooseBucket will allways return a number from set {-1,..., 26}. */
/* Return value -1 means that the item is too large to fit any of the        */
/* buckets, all other possible values are indices of buckets.                */
/*                                                                           */
/* Please note that the value of the expression:                             */
/* ┌────────────────────────────────────────────────────┐                    */
/* │ ((cur->x1 > sep->x1) << 1)  |  (cur->x0 < sep->x0) │                    */
/* └────────────────────────────────────────────────────┘                    */
/* can be only 0, 1, 2 or 3.                                                 */
/*                                                                           */
/* The illustration bellow shows all possible relative positions of points   */
/* sep->x0, sep->x1, cbox->x0 and cbox->x1.                                  */
/*                                                                           */
/*                sep.x0       sep.x1                                        */
/* ┌─────────────────┼────────────┼─────────────────┬───┐                    */
/* │                 │ ╞════════╡ │                 │ 0 │                    */
/* ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌┤                    */
/* │ ╞═══════════════╪══════════╡ │                 │ 1 │                    */
/* │ ╞═════════════╡ │            │                 │   │                    */
/* ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌┤                    */
/* │                 │            │ ╞═════════════╡ │ 2 │                    */
/* │                 │ ╞══════════╪═══════════════╡ │   │                    */
/* ├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┼╌╌╌┤                    */
/* │ ╞═══════════════╪════════════╪═══════════════╡ │ 3 │                    */
/* └─────────────────┴────────────┴─────────────────┴───┘                    */
int ChooseBucket ( Box3d *sep, Box3d *cur )
{
  int bucket_idx, cmp_res;

  bucket_idx = cmp_res = ((cur->x1 > sep->x1) << 1) | (cur->x0 < sep->x0);
  if (cmp_res == 3)
    return -1;
  bucket_idx *= 3;
  bucket_idx += cmp_res = ((cur->y1 > sep->y1) << 1) | (cur->y0 < sep->y0);
  if (cmp_res == 3)
    return -1;
  bucket_idx *= 3;
  bucket_idx += cmp_res = ((cur->z1 > sep->z1) << 1) | (cur->z0 < sep->z0);
  if (cmp_res == 3)
    return -1;
  return bucket_idx;
} /* ChooseBucket */

void ExpandNode ( rendertnode *node, void **pqueue )
{
  /* Filter children and and create subtrees. */
  rendertnode *buckets[BUCKET_COUNT];
  rendertnode *too_big, **bucket, *item, *next_item, *subtrees;
  Box3d       bucket_bbox[BUCKET_COUNT];
  Box3d       sep;
  int         i, bucket_idx;

  assert ( node != NULL && node->k == TNODE_NOT_EXPANDED );

  /* Initialize local variables. */
  too_big = subtrees = NULL;
  for ( i = 0; i < BUCKET_COUNT; ++i ) {
    buckets[i] = NULL;
    bucket_bbox[i].x0 = bucket_bbox[i].y0 = bucket_bbox[i].z0 = DBL_MAX;
    bucket_bbox[i].x1 = bucket_bbox[i].y1 = bucket_bbox[i].z1 = -DBL_MAX;
  }
  ComputeSeparator ( &node->bbox, &sep );

  /* Put all items from list into appropriate buckets. */
  for ( item = node->left; item != NULL; item = next_item ) {
    next_item = item->right;
    bucket_idx = ChooseBucket ( &sep, &item->bbox );
    if ( bucket_idx == -1 )
      bucket = &too_big;
    else {
      bucket = &buckets[bucket_idx];
      rbez_FindSumBBoxd ( &bucket_bbox[bucket_idx], 
                          &item->bbox, &bucket_bbox[bucket_idx] );
    }
    item->right = *bucket; /* When bucket is empty, item->right is NULL. */
    *bucket = item;
  }

  /* Create child nodes. */
  for ( bucket_idx = 0; bucket_idx < BUCKET_COUNT; ++bucket_idx ) {
    if (buckets[bucket_idx] == NULL)
      continue;
    if (buckets[bucket_idx]->right == NULL) { 
      /* If there is only one element, it can be moved to too_big list. */
      buckets[bucket_idx]->right = too_big;
      too_big = buckets[bucket_idx];
    }
    else {
      subtrees = NewNode ( TNODE_NOT_EXPANDED, buckets[bucket_idx], subtrees );
      subtrees->bbox = bucket_bbox[bucket_idx];
    }
  }

  /* Number of nonempty buckets shouldn't be 1. */
  assert ( too_big != NULL || (subtrees != NULL && subtrees->right != NULL) );

  /* Construct a subtree using the old n^2 algorithm. */
  BuildInnerObjectTree ( node, subtrees, too_big, pqueue );
  node->k = TNODE_INNER;
} /*ExpandNode*/

void BuildInnerObjectTree ( rendertnode *root, rendertnode *subtrees, 
                            rendertnode *leaves, void **pqueue )
{
  int         nobjects, i, j, k, last;
  rendertnode *next_leaf, *tn0, *tn1, *tn2;
  Box3d       box;
  double      d1, d2;

  for ( nobjects = 0;  subtrees != NULL;  subtrees = subtrees->right, nobjects++ )
    pqueue[nobjects] = (void*)subtrees;
  for ( ;  leaves != NULL;  leaves = next_leaf, nobjects++ ) {
    pqueue[nobjects] = (void*)leaves;
    next_leaf = leaves->right;
    leaves->right = NULL;
    assert ( leaves->left == NULL );
  }
  assert ( nobjects >= 2 );
  pkv_HeapOrder ( pqueue, nobjects, CompNodePrio );
  for ( i = nobjects-1; i > 1; i-- ) {
    tn0 = pqueue[0];
    last = i;
    pkv_HeapRemove ( pqueue, &last, 0, CompNodePrio );
    tn1 = pqueue[0];
    rbez_FindSumBBoxd ( &tn0->bbox, &tn1->bbox, &box );
    d1 = BoxDiameter ( &box );
    k = 0;
    for ( j = 1; j <= last; j++ ) {
      tn2 = pqueue[j];
      rbez_FindSumBBoxd ( &tn0->bbox, &tn2->bbox, &box );
      d2 = BoxDiameter ( &box );
      if ( d2 < d1 ) {
        k = j;
        d1 = d2;
        tn1 = tn2;
      }
    }
    pkv_HeapRemove ( pqueue, &last, k, CompNodePrio );
    tn2 = NewInnerNode ( tn0, tn1 );
    pkv_HeapInsert ( pqueue, &last, (void*)tn2, CompNodePrio );
  }
  root->left = pqueue[0];
  root->right = pqueue[1];
} /*BuildInnerObjectTree*/

boolean BuildObjectTree ( pkRenderer *rend )
{
  int         nobjects;
  void        **pqueue;
  int         i, depth;
  rendertnode *children, *next_node, *current_node, *prev_node;
  Box3d       root_bbox;
  boolean     down;

  children = NULL;
  root_bbox.x0 = root_bbox.y0 = root_bbox.z0 = DBL_MAX;
  root_bbox.x1 = root_bbox.y1 = root_bbox.z1 = -DBL_MAX;
  nobjects = rend->nobjects;
  if ( !nobjects )
    return false;
  if ( nobjects == 1 ) {
    rend->root = NewLeafNode ( rend, 0, NULL, NULL );
    return true;
  }

  /* The end of pqueue will be used to store current path in tree. Total */
  /* number of pointers written to pqueue by expand() and this procedure */
  /* won't be greater than nobjects.                                     */
  pqueue = malloc ( (nobjects+1)*sizeof(void*) );
  if ( !pqueue )
    return false;

  /* Create root node and attach to it a list of nodes for all objects */
  for ( i = 0; i < nobjects; i++ ) {
    children = NewLeafNode ( rend, i, NULL, children );
    rbez_FindSumBBoxd ( &root_bbox, &children->bbox, &root_bbox );
  }
  rend->root = NewNode ( TNODE_NOT_EXPANDED, children, NULL );
  rend->root->bbox = root_bbox;

  /* Visit all vertices in tree and expand them if necessary */
  depth = 1;
  memset ( pqueue, 0, (nobjects+1)*sizeof(void*) );
  current_node = NULL;
  next_node = rend->root;

  while ( next_node != NULL ) {
    prev_node = current_node;
    current_node = next_node;

    if ( current_node->k == TNODE_NOT_EXPANDED )
      ExpandNode ( current_node, pqueue );

    if ( current_node->k == TNODE_INNER ) { /* Visit inner node */
      if ( prev_node == pqueue[nobjects - depth] ) {
        next_node = current_node->left;
        down = true;
      }
      else if ( prev_node == current_node->left ) {
        next_node = current_node->right;
        down = true;
      }
      else if ( current_node->right ) {
        next_node = pqueue[nobjects - depth];
        pqueue[nobjects-depth] = NULL;
        down = false;
      }
      else
        assert ( false );
    }
    else { /* Visit leaf */
      next_node = pqueue[nobjects - depth];
      assert ( current_node->left == NULL && current_node->right == NULL );
      down = false;
    }

    if ( down ) {
      depth ++;
      pqueue[nobjects - depth] = current_node;
    }
    else
      depth --;
  }
  assert ( !down && depth <= 0 );
  free ( pqueue );

  return true;
}

void DestroyTNodeTree ( rendertnode *node )
{
  if ( node ) {
    DestroyTNodeTree ( node->left );
    DestroyTNodeTree ( node->right );
    free ( node );
  }
} /*DestroyTNodeTree*/

