
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#define _ISOC99_SOURCE /* for cbrt, floor, ceil */
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <math.h>

#ifndef _SYS_TIMES_H
#include <sys/times.h>
#endif

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

#define MAXPARAMS 10  /* maximal number of function parameters */

 /* this array must be sorted alphabetically */
const char *pkscalc_fnames[pkscalc_NUMNAMES] =
  { "PI",
    "abs",
    "acosh",
    "asinh",
    "atanh",
    "arccos",
    "arcsin",
    "arctan",
    "cbrt",
    "ceiling",
    "clamp",
    "cos",
    "cosh",
    "exp",
    "floor",
    "ln",
    "max",
    "min",
    "pow",
    "round",
    "sign",
    "sin",
    "sinh",
    "sqrt",
    "tan",
    "tanh" };
static const short funcpnum[pkscalc_NUMNAMES] =
  {0,1,1,1,1,1,1,-1,1,1,3,1,1,1,1,1,-1,-1,2,1,1,1,1,1,1,1};

int pkscalc_GetNextSymbol ( pks_scanner *sc )
{
  pks_GetNextSymbol ( sc );
  if ( sc->nextsymbol == PKS_SYMB_IDENT ) {
    sc->nextinteger = pks_BinSearchNameTable ( pkscalc_NUMNAMES, 0, pkscalc_fnames,
                                               sc->nextname );
    sc->id_space = ID_SPACE_CALC;
  }
  return sc->nextsymbol;
} /*pkscalc_GetNextSymbol*/

static boolean _QCLogExpression ( pks_scanner *sc, boolean ignore,
                                  pks_calcval *value );

static boolean _EvaluateFunction ( pks_scanner *sc, boolean ignore,
                                   pks_calcval *value )
{
  int         fname, nparams;
  pks_calcval param[MAXPARAMS];

  if ( sc->id_space != ID_SPACE_CALC || sc->nextinteger < 0 )
    return false;
  fname = sc->nextinteger;
  sc->GetNextSymbol ( sc );
        /* compute the parameters */
  if ( funcpnum[fname] < 0 ) {  /* functions min, max and atan, */
                                /* with varying numbers of parameters */
    if ( sc->nextsymbol != PKS_SYMB_LPAREN )
      return false;
    sc->GetNextSymbol ( sc );
    switch ( fname ) {
  case pkscalc_f_ARCTAN:  /* one or two parameters */
      if ( !_QCLogExpression ( sc, ignore, &param[0] ) )
        return false;
      if ( !ignore )
        value->valtype = PKS_SYMB_FLOAT;
      if ( sc->nextsymbol == PKS_SYMB_RPAREN ) {
        if ( !ignore )
          value->floatval = atan ( param[0].floatval );
      }
      else if ( sc->nextsymbol == PKS_SYMB_COMMA ) {
        sc->GetNextSymbol ( sc );
        if ( !_QCLogExpression ( sc, ignore, &param[1] ) )
          return false;
        if ( !ignore )
          value->floatval = atan2 ( param[0].floatval, param[1].floatval );
      }
      else
        return false;
      break;
  case pkscalc_f_MAX:
      if ( !_QCLogExpression ( sc, ignore, value ) )
        return false;
      while ( sc->nextsymbol == PKS_SYMB_COMMA ) {
        sc->GetNextSymbol ( sc );
        if ( !_QCLogExpression ( sc, ignore, &param[0] ) )
          return false;
        if ( !ignore ) {
          if ( value->valtype == PKS_SYMB_INTEGER &&
               param[0].valtype == PKS_SYMB_INTEGER ) {
            if ( param[0].intval > value->intval )
              *value = param[0];
          }
          else {
            if ( param[0].floatval > value->floatval )
              *value = param[0];
          }
        }
      }
      break;
  case pkscalc_f_MIN:
      if ( !_QCLogExpression ( sc, ignore, value ) )
        return false;
      while ( sc->nextsymbol == PKS_SYMB_COMMA ) {
        sc->GetNextSymbol ( sc );
        if ( !_QCLogExpression ( sc, ignore, &param[0] ) )
          return false;
        if ( !ignore ) {
          if ( value->valtype == PKS_SYMB_INTEGER &&
               param[0].valtype == PKS_SYMB_INTEGER ) {
            if ( param[0].intval < value->intval )
              *value = param[0];
          }
          else {
            if ( param[0].floatval < value->floatval )
              *value = param[0];
          }
        }
      }
      break;
  default:
      return false;
    }
    if ( sc->nextsymbol != PKS_SYMB_RPAREN )
      return false;
    sc->GetNextSymbol ( sc );
    return true;
  }
  else {
    nparams = 0;
    memset ( param, 0, MAXPARAMS*sizeof(pks_calcval) );
    if ( sc->nextsymbol == PKS_SYMB_LPAREN ) {
      for (;;) {
        sc->GetNextSymbol ( sc );  /* skip the opening parenthesis or comma */
        if (!_QCLogExpression ( sc, ignore, &param[nparams] ) )
          return false;
        if ( param[nparams].valtype == PKS_SYMB_INTEGER )
          param[nparams].floatval = (double)param[nparams].intval;
        nparams ++;
        if ( sc->nextsymbol == PKS_SYMB_RPAREN )
          break;
        else if ( sc->nextsymbol != PKS_SYMB_COMMA || nparams >= MAXPARAMS ) {
          sc->nextsymbol = value->valtype = PKS_SYMB_ERROR;
          return false;
        }
      }
      sc->GetNextSymbol ( sc );  /* skip the closing parenthesis */
    }
    if ( nparams != funcpnum[fname] )  /* invalid number of parameters */
      return false;
    if ( ignore )
      return true;
        /* compute the proper function value */
    value->valtype = PKS_SYMB_FLOAT;
    switch ( fname ) {
  case pkscalc_f_PI:
      value->floatval = PI;
      return true;
  case pkscalc_f_ABS:
      value->valtype  = param[0].valtype;
      value->intval   = abs ( param[0].intval );
      value->floatval = fabs ( param[0].floatval );
      return true;
  case pkscalc_f_ACOSH:
      value->floatval = acosh ( param[0].floatval );
      return true;
  case pkscalc_f_ASINH:
      value->floatval = asinh ( param[0].floatval );
      return true;
  case pkscalc_f_ATANH:
      value->floatval = atanh ( param[0].floatval );
      return true;
  case pkscalc_f_ARCCOS:
      value->floatval = acos ( param[0].floatval );
      return true;
  case pkscalc_f_ARCSIN:
      value->floatval = asin ( param[0].floatval );
      return true;
  case pkscalc_f_CBRT:
      value->floatval = cbrt ( param[0].floatval );
      return true;
  case pkscalc_f_CEILING:
      value->floatval = ceil ( param[0].floatval );
      goto setint;
  case pkscalc_f_CLAMP:
      if ( param[0].floatval < param[1].floatval )
        *value = param[1];
      else if ( param[0].floatval < param[2].floatval )
        *value = param[2];
      else
        *value = param[0];
      return true;
  case pkscalc_f_COS:
      value->floatval = cos ( param[0].floatval );
      return true;
  case pkscalc_f_COSH:
      value->floatval = cosh ( param[0].floatval );
      return true;
  case pkscalc_f_EXP:
      value->floatval = exp ( param[0].floatval );
      return true;
  case pkscalc_f_FLOOR:
      value->floatval = floor ( param[0].floatval );
      goto setint;
  case pkscalc_f_LN:
      value->floatval = log ( param[0].floatval );
      return true;
  case pkscalc_f_POW:
      if ( param[1].valtype == PKS_SYMB_FLOAT )
        value->floatval = pow ( param[0].floatval, param[1].floatval );
      else if ( param[0].valtype == PKS_SYMB_FLOAT || param[1].floatval < 0 )
        value->floatval = pkv_rpower ( param[0].floatval, param[1].intval );
      else {
        value->intval = pkv_ipower ( param[0].floatval, param[1].intval );
        value->floatval = (double)value->intval;
        value->valtype = PKS_SYMB_INTEGER;
      }
      return true;
  case pkscalc_f_ROUND:
      value->floatval = floor ( param[0].floatval );
setint:
      if ( value->floatval >= INT_MIN && value->floatval <= INT_MAX ) {
        value->intval = (int)value->floatval;
        value->valtype = PKS_SYMB_INTEGER;
      }
      return true;
  case pkscalc_f_SIGN:
      if ( param[0].floatval > 0.0 )
        value->intval = 1;
      else if ( param[0].floatval < 0.0 )
        value->intval = -1;
      else
        value->intval = 0;
      value->floatval = (double)value->intval;
      value->valtype = PKS_SYMB_INTEGER;
      return true;
  case pkscalc_f_SIN:
      value->floatval = sin  ( param[0].floatval );
      return true;
  case pkscalc_f_SINH:
      value->floatval = sinh ( param[0].floatval );
      return true;
  case pkscalc_f_SQRT:
      value->floatval = sqrt ( param[0].floatval );
      return true;
  case pkscalc_f_TAN:
      value->floatval = tan  ( param[0].floatval );
      return true;
  case pkscalc_f_TANH:
      value->floatval = tanh  ( param[0].floatval );
      return true;
  default:
      sc->nextsymbol = value->valtype = PKS_SYMB_ERROR;
      return false;
    }
  }
} /*_EvaluateFunction*/

static boolean _NumFactor ( pks_scanner *sc, boolean ignore,
                            pks_calcval *value )
{
  boolean neg;

  if ( sc->nextsymbol == PKS_SYMB_EXCLAM ) {
    neg = true;
    sc->GetNextSymbol ( sc );
  }
  else
    neg = false;
  switch ( sc->nextsymbol ) {
case PKS_SYMB_INTEGER:
    value->valtype  = PKS_SYMB_INTEGER;
    value->intval   = sc->nextinteger;
    value->floatval = sc->nextfloat;
    sc->GetNextSymbol ( sc );
    goto neg_it;
case PKS_SYMB_FLOAT:
    value->valtype  = PKS_SYMB_FLOAT;
    value->floatval = sc->nextfloat;
    sc->GetNextSymbol ( sc );
    goto neg_it;
case PKS_SYMB_IDENT:
    if ( sc->id_space == ID_SPACE_NONE ) {  /* if not found yet */
      sc->nextinteger = pks_BinSearchNameTable ( pkscalc_NUMNAMES,
                            pkscalc_f_PI, pkscalc_fnames, sc->nextname );
      if ( sc->nextinteger >= 0 )
        sc->id_space = ID_SPACE_CALC;
    }
    if ( sc->id_space == ID_SPACE_CALC ) {
      if ( !_EvaluateFunction ( sc, ignore, value ) )
        return false;
    }
    else {
      if ( !sc->AppIdentCalc ( sc, ignore, value ) )
        return false;
    }
    goto neg_it;
case PKS_SYMB_LPAREN:
    sc->GetNextSymbol ( sc );
    _QCLogExpression ( sc, ignore, value );
    if ( sc->nextsymbol == PKS_SYMB_RPAREN ) {
      sc->GetNextSymbol ( sc );
      goto neg_it;
    }
    else
      return false;
default:
    return false;
  }
neg_it:
  if ( neg ) {
    if ( value->valtype == PKS_SYMB_INTEGER )
      value->intval = !(value->intval != 0);
    else {
      value->intval = !(value->floatval != 0.0);
      value->valtype = PKS_SYMB_INTEGER;
    }
  }
  return true;
} /*_NumFactor*/

static boolean _SgnFactor ( pks_scanner *sc, boolean ignore,
                            pks_calcval *value )
{
  int op;;

  switch ( (op = sc->nextsymbol) ) {
case PKS_SYMB_PLUS:
case PKS_SYMB_MINUS:
case PKS_SYMB_TILDE:
case PKS_SYMB_EXCLAM:
    sc->GetNextSymbol ( sc );
    break;
default:
    break;
  }
  if ( !_NumFactor ( sc, ignore, value ) )
    return false;
  if ( !ignore )
    switch ( op ) {
  case PKS_SYMB_MINUS:
      if ( value->valtype == PKS_SYMB_INTEGER )
        value->intval = -value->intval;
      value->floatval = -value->floatval;
      break;
  case PKS_SYMB_TILDE:
      if ( value->valtype != PKS_SYMB_INTEGER )
        return false;
      value->intval = ~value->intval;
      break;
  case PKS_SYMB_EXCLAM:
      if ( value->valtype == PKS_SYMB_INTEGER )
        value->intval = !value->intval;
      else {
        value->intval = !value->floatval;
        value->valtype = PKS_SYMB_INTEGER;
      }
      break;
    }
  return true;
} /*_SgnFactor*/

static boolean _NumTerm ( pks_scanner *sc, boolean ignore, pks_calcval *value )
{
  pks_calcval s;

  if ( !_SgnFactor ( sc, ignore, value ) )
    return false;
  for (;;) {
    switch ( sc->nextsymbol ) {
  case PKS_SYMB_STAR:
      sc->GetNextSymbol ( sc );
      if ( !_SgnFactor ( sc, ignore, &s ) )
        return false;
      if ( !ignore ) {
        if ( value->valtype == PKS_SYMB_INTEGER && s.valtype == PKS_SYMB_INTEGER )
          value->intval *= s.intval;
        else
          value->valtype = PKS_SYMB_FLOAT;
        value->floatval *= s.floatval;
      }
      break;
  case PKS_SYMB_SLASH:
      sc->GetNextSymbol ( sc );
      if ( !_SgnFactor ( sc, ignore, &s ) )
        return false;
      if ( !ignore ) {
        if ( value->valtype == PKS_SYMB_INTEGER && s.valtype == PKS_SYMB_INTEGER )
          value->intval /= s.intval;
        else
          value->valtype = PKS_SYMB_FLOAT;
        value->floatval /= s.floatval;
      }
      break;
  case PKS_SYMB_PERCENT:
      sc->GetNextSymbol ( sc );
      if ( !_SgnFactor ( sc, ignore, &s ) )
        return false;
      if ( value->valtype == PKS_SYMB_INTEGER && s.valtype == PKS_SYMB_INTEGER )
        value->intval %= s.intval;
      else {
        value->valtype = PKS_SYMB_ERROR;
        return false;
      }
      break;
  default:
      return value->valtype != PKS_SYMB_ERROR;
    }
  }
} /*_NumTerm*/

static boolean _NumExpression ( pks_scanner *sc, boolean ignore,
                                pks_calcval *value )
{
  pks_calcval a;

  if ( !_NumTerm ( sc, ignore, value ) )
    return false;
  for (;;) {
    switch ( sc->nextsymbol ) {
  case PKS_SYMB_PLUS:
      sc->GetNextSymbol ( sc );
      if ( !_NumTerm ( sc, ignore, &a ) )
        return false;
      if ( !ignore ) {
        if ( value->valtype == PKS_SYMB_INTEGER && a.valtype == PKS_SYMB_INTEGER )
          value->intval += a.intval;
        else
          value->valtype = PKS_SYMB_FLOAT;
        value->floatval += a.floatval;
      }
      break;
  case PKS_SYMB_MINUS:
      sc->GetNextSymbol ( sc );
      if ( !_NumTerm ( sc, ignore, &a ) )
        return false;
      if ( !ignore ) {
        if ( value->valtype == PKS_SYMB_INTEGER && a.valtype == PKS_SYMB_INTEGER )
          value->intval -= a.intval;
        else
          value->valtype = PKS_SYMB_FLOAT;
        value->floatval -= a.floatval;
      }
      break;
  default:
      return value->valtype != PKS_SYMB_ERROR;
    }
  }
} /*_NumExpression*/

static boolean _ShiftExpression ( pks_scanner *sc, boolean ignore,
                                  pks_calcval *value )
{
  pks_calcval a;
  int         op;

  if ( !_NumExpression ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_LLANGLE ||
          sc->nextsymbol == PKS_SYMB_RRANGLE ) {
    if ( !ignore && value->valtype != PKS_SYMB_INTEGER )
      return false;
    op = sc->nextsymbol;
    sc->GetNextSymbol ( sc );
    if ( !_NumExpression ( sc, ignore, &a ) )
      return false;
    if ( !ignore ) {
      if ( a.valtype != PKS_SYMB_INTEGER )
        return false;
      if ( op == PKS_SYMB_LLANGLE )
        value->intval <<= a.intval;
      else
        value->intval >>= a.intval;
    }
  }
  return true;
 
} /*_ShiftExpression*/

static boolean _RelExpression ( pks_scanner *sc, boolean ignore,
                                pks_calcval *value )
{
  pks_calcval a1, a2;
  short int   relop;
  boolean     s;

  if ( !_ShiftExpression ( sc, ignore, value ) )
    return false;
  a1 = *value;
  for (;;) {
    switch ( (relop = sc->nextsymbol) ) {
  case PKS_SYMB_EQEQ:
  case PKS_SYMB_LANGLE:
  case PKS_SYMB_RANGLE:
  case PKS_SYMB_LEQ:
  case PKS_SYMB_GEQ:
  case PKS_SYMB_NEQ:
      sc->GetNextSymbol ( sc );
      break;
  default:
      return true;
    }
    if ( !_ShiftExpression ( sc, ignore, &a2 ) )
      return false;
    if ( !ignore ) {
      value->valtype = PKS_SYMB_INTEGER;
      value->intval = 1;
      if ( a1.valtype == PKS_SYMB_FLOAT || a2.valtype == PKS_SYMB_FLOAT ) {
        switch ( relop ) {
    case PKS_SYMB_EQEQ:   s = a1.floatval == a2.floatval;  break;
    case PKS_SYMB_LANGLE: s = a1.floatval < a2.floatval;   break;
    case PKS_SYMB_RANGLE: s = a1.floatval > a2.floatval;   break;
    case PKS_SYMB_LEQ:    s = a1.floatval <= a2.floatval;  break;
    case PKS_SYMB_GEQ:    s = a1.floatval >= a2.floatval;  break;
    case PKS_SYMB_NEQ:    s = a1.floatval != a2.floatval;  break;
    default:              s = false;
        }
      }
      else {
        switch ( relop ) {
    case PKS_SYMB_EQEQ:   s = a1.intval == a2.intval;  break;
    case PKS_SYMB_LANGLE: s = a1.intval < a2.intval;   break;
    case PKS_SYMB_RANGLE: s = a1.intval > a2.intval;   break;
    case PKS_SYMB_LEQ:    s = a1.intval <= a2.intval;  break;
    case PKS_SYMB_GEQ:    s = a1.intval >= a2.intval;  break;
    case PKS_SYMB_NEQ:    s = a1.intval != a2.intval;  break;
    default:              s = false;
        }
      }
      if ( !s ) {
        value->intval = 0;
        ignore = true;
      }
    }
  }
} /*_RelExpression*/

static boolean _BinFactor ( pks_scanner *sc, boolean ignore,
                            pks_calcval *value )
{
  pks_calcval a;

  if ( !_RelExpression ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_ET ) {
    if ( !ignore && value->valtype != PKS_SYMB_INTEGER )
      return false;
    sc->GetNextSymbol ( sc );
    if ( !_RelExpression ( sc, ignore, &a ) )
      return false;
    if ( !ignore ) {
      if ( a.valtype != PKS_SYMB_INTEGER )
        return false;
      value->intval &= a.intval;
    }
  }
  return true;
} /*_BinFactor*/

static boolean _BinTerm ( pks_scanner *sc, boolean ignore,
                          pks_calcval *value )
{
  pks_calcval a;

  if ( !_BinFactor ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_WEDGE ) {
    if ( !ignore && value->valtype != PKS_SYMB_INTEGER )
      return false;
    sc->GetNextSymbol ( sc );
    if ( !_BinFactor ( sc, ignore, &a ) )
      return false;
    if ( !ignore ) {
      if ( a.valtype != PKS_SYMB_INTEGER )
        return false;
      value->intval ^= a.intval;
    }
  }
  return true;
} /*_BinTerm*/

static boolean _BinExpression ( pks_scanner *sc, boolean ignore,
                                pks_calcval *value )
{
  pks_calcval a;

  if ( !_BinTerm ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_VERTBAR ) {
    if ( !ignore && value->valtype != PKS_SYMB_INTEGER )
      return false;
    sc->GetNextSymbol ( sc );
    if ( !_BinTerm ( sc, ignore, &a ) )
      return false;
    if ( !ignore ) {
      if ( a.valtype != PKS_SYMB_INTEGER )
        return false;
      value->intval |= a.intval;
    }
  }
  return true;
} /*_BinExpression*/

static boolean _LogTerm ( pks_scanner *sc, boolean ignore,
                          pks_calcval *value )
{
  pks_calcval a;

  if ( !_BinExpression ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_ETET ) {
    sc->GetNextSymbol ( sc );
    if ( !ignore ) {
      ignore = value->valtype == PKS_SYMB_INTEGER ?
                    value->intval == 0 : value->floatval == 0.0;
    }
    if ( !_BinExpression ( sc, ignore, &a ) )
      return false;
    if ( !ignore )
      value->intval = a.intval;
  }
  return true;
} /*_LogTerm*/

static boolean _LogExpression ( pks_scanner *sc, boolean ignore,
                                pks_calcval *value )
{
  pks_calcval a;

  if ( !_LogTerm ( sc, ignore, value ) )
    return false;
  while ( sc->nextsymbol == PKS_SYMB_VERTBARBAR ) {
    sc->GetNextSymbol ( sc );
    if ( !ignore ) {
      ignore = value->valtype == PKS_SYMB_INTEGER ?
                    value->intval != 0 : value->floatval != 0.0;
    }
    if ( !_LogTerm ( sc, ignore, &a ) )
      return false;
    if ( !ignore )
      value->intval = a.intval;
  }
  return true;
} /*_LogExpression*/

static boolean _QCLogExpression ( pks_scanner *sc, boolean ignore,
                                  pks_calcval *value )
{
  boolean     getfirst;
  pks_calcval ignval;

  if ( !_LogExpression ( sc, ignore, value ) )
    return false;
  if ( sc->nextsymbol == PKS_SYMB_QUESTION ) {
    sc->GetNextSymbol ( sc );
    getfirst = value->valtype == PKS_SYMB_INTEGER ?
                      value->intval != 0 : value->floatval != 0.0;
    if ( !_LogExpression ( sc, ignore || !getfirst,
                           getfirst ? value : &ignval ) )
      return false;
    if ( sc->nextsymbol != PKS_SYMB_COLON )
      return false;
    sc->GetNextSymbol ( sc );
    if ( !_LogExpression ( sc, ignore || getfirst,
                           getfirst ? &ignval : value ) )
      return false;
  }
  return true;
} /*_QCLogExpression*/

boolean pkscalc_LogExpression ( pks_scanner *sc, pks_calcval *value )
{
  if ( _LogExpression ( sc, false, value ) )
    return true;
  else {
    value->valtype = PKS_SYMB_ERROR;
    return false;
  }
} /*pkscalc_LogExpression*/

boolean pkscalc_Expression ( pks_scanner *sc, pks_calcval *value )
{
  if ( _QCLogExpression ( sc, false, value ) )
    return true;
  else {
    value->valtype = PKS_SYMB_ERROR;
    return false;
  }
} /*pkscalc_Expression*/

