
/* ///////////////////////////////////////////////////////////////////////// */
/* This file is a part of the BSTools package                                */
/* written by Przemyslaw Kiciak                                              */
/* ///////////////////////////////////////////////////////////////////////// */
/* (C) Copyright by Przemyslaw Kiciak, 2015, 2017                            */
/* this package is distributed under the terms of the                        */
/* Lesser GNU Public License, see the file COPYING.LIB                       */
/* ///////////////////////////////////////////////////////////////////////// */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>
#include <math.h>

#ifndef _SYS_TIMES_H
#include <sys/times.h>
#endif

#include "pkvaria.h"
#include "avltree.h"
#include "pkscanner.h"

#include "pksprivate.h"

/*#define DEBUG*/

/* ///////////////////////////////////////////////////////////////////////// */
#define HASH_CONST        0
#define HASH_DEF          1
#define HASH_DEFINE       2
#define HASH_ELSE         3
#define HASH_ELSEIF       4
#define HASH_ENDIF        5
#define HASH_ERROR        6
#define HASH_IF           7
#define HASH_INCLUDE      8
#define HASH_UNDEF        9
#define HASH_SYMB_LENGTH 10

static const char *hashnames[HASH_SYMB_LENGTH] =
  { "const",
    "def",
    "define",
    "else",
    "elseif",
    "endif",
    "error",
    "if",
    "include",
    "undef" };

/* ///////////////////////////////////////////////////////////////////////// */
static char _PreprocCompKeys ( AVLtnode *node, void *data )
{
  pks_macro *ncr, *mcr;
  char      *nname, *mname;
  int       l;

  ncr = (pks_macro*)node;
  nname = &((char*)ncr)[sizeof(pks_macro)];
  mcr = (pks_macro*)data;
  mname = &((char*)mcr)[sizeof(pks_macro)];
  l = strcmp ( nname, mname );
  if ( l )
    return l > 0 ? +1 : -1;
  else
    return 0;
} /*_PreprocCompKeys*/

static AVLtnode *_PreprocAllocNode ( void *data )
{
  return (AVLtnode*)data;
} /*_PreprocAllocNode*/

static void _PreprocFreeNode ( AVLtnode *node )
{
  pks_macro *mcr, **pmtab;
  int       i;

  mcr = (pks_macro*)node;
  if ( mcr->nparams > 0 ) {
    pmtab = (pks_macro**)&((char*)mcr)[mcr->params];
    for ( i = 0; i < mcr->nparams; i++ )
      free ( pmtab[i] );
  }
  free ( node );
} /*_PreprocFreeNode*/

/* ///////////////////////////////////////////////////////////////////////// */
boolean pks_InitPreprocessor ( pks_scanner *psc,
                unsigned int options,
                int inbuflength, int maxnamelength,
                const char *bcomment, const char *ecomment,
                int (*InputData)(pks_scanner*),
                int (*GetNextSymbol)(pks_scanner*),
                boolean (*AppIdentFind)(pks_scanner*),
                boolean (*AppIdentCalc)(pks_scanner*,boolean,pks_calcval*),
                void *userdata )
{
  if ( !GetNextSymbol )
    GetNextSymbol = pks_GetNextPSymbol;
  if ( !pks_InitScanner ( psc, options, inbuflength, NULL, maxnamelength,
             bcomment, ecomment,
             InputData, GetNextSymbol, AppIdentFind, AppIdentCalc, userdata ) )
    return false;
  AVLInit ( &psc->macros, true,
            _PreprocCompKeys, _PreprocAllocNode, _PreprocFreeNode );
  psc->mstack = NULL;
  psc->ifsp = 0;
  return true;
} /*pks_InitPreprocessor*/

void pks_ShutDownPreprocessor ( pks_scanner *psc )
{
  AVLDestroy ( &psc->macros );
  pks_ShutDownScanner ( psc );
} /*pks_ShutDownPreprocessor*/

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _pks_ProcessHashConst ( pks_scanner *psc )
{
  pks_calcval x;

  pks_GetNextSymbol ( psc ); /* skip #const */
  if ( psc->nextsymbol != PKS_SYMB_LPAREN )
    goto failure;
  pks_GetNextSymbol ( psc );
  if ( !pkscalc_Expression ( psc, &x ) )
    goto failure;
  if ( psc->nextsymbol != PKS_SYMB_RPAREN )
    goto failure;
     /* do not skip the right parenthesis, replace it by the expression value */
  if ( (psc->nextsymbol = x.valtype) == PKS_SYMB_INTEGER )
    psc->nextinteger = x.intval;
  psc->nextfloat = x.floatval;
  return true;

failure:
  return false;
} /*_pks_ProcessHashConst*/

/* ///////////////////////////////////////////////////////////////////////// */
boolean _pks_AddTheMacroToken ( pks_scanner *psc, char *mbuf, int *bufp )
{
  pks_token   token;
  int         length, l0, l1;

  switch ( (token.any.symbol = psc->nextsymbol) ) {
case PKS_SYMB_IDENT:
case PKS_SYMB_STRING:
    l0 = (char*)&token.name.name[0]-(char*)&token;
    l1 = strlen ( psc->nextname );
    token.name.intvalue = psc->nextinteger;
    token.name.id_space = psc->id_space;
    token.name.length = l0+l1+1;
    if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, bufp, (char*)&token, l0 ) )
      return false;
    return pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, bufp,
                            psc->nextname, l1+1 );

case PKS_SYMB_INTEGER:
case PKS_SYMB_CHAR:
case PKS_SYMB_OTHER:
case PKS_SYMB_MACROPAR:
    length = sizeof(pks_token_int);
    token.integer.intvalue = psc->nextinteger;
    goto attach_it;

case PKS_SYMB_FLOAT:
    length = sizeof(pks_token_float);
    token.real.floatvalue = psc->nextfloat;
    goto attach_it;

default:
    length = sizeof(pks_token_any);
attach_it:
    token.any.length = length;
    return pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, bufp,
                            (char*)&token, length );
  }
} /*_pks_AddTheMacroToken*/

static char _PreprocCompMacroNames ( AVLtnode *node, void *data )
{
  pks_macropname *ncr, *pcr;
  int            l;

  ncr = (pks_macropname*)node;
  pcr = (pks_macropname*)data;
  l = strcmp ( ncr->name, pcr->name );
  if ( l )
    return l > 0 ? +1 : -1;
  else
    return 0;
} /*_PreprocCompMacroNames*/

static boolean _pks_ProcessHashDefine ( pks_scanner *psc )
{
  void           *sp;
  pks_macro      *mcr, *mcri;
  char           *mbuf;
  int            bufp;
  pks_token_any  end;
  AVLtree        ptree;
  pks_macropname *pname, *pn;
  int            npar;

  sp = pkv_GetScratchMemTop ();
  psc->bkslnl = true;
  mbuf = pkv_GetScratchMem ( PKS_MACRO_BUF_LENGTH );
  if ( !mbuf ) {
    psc->error_code = PRERROR_NO_MEMORY;
    goto failure;
  }
  pname = NULL;
        /* initially empty name, no parameters */
  mcri = (pks_macro*)mbuf;
  bufp = sizeof(pks_macro);
  memset ( mcri, 0, bufp+1 );

  npar = 0;
  pks_GetNextSymbol ( psc );  /* skip #define */
  if ( psc->nextsymbol == PKS_SYMB_IDENT ) {
    if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                           psc->nextname, strlen(psc->nextname)+1 ) ) {
      psc->error_code = PRERROR_BUFFER_OVERFLOW;
      goto failure;
    }
    mcri->params = bufp;
    if ( psc->nextchar == '(' ) {  /* there is a parameter list */
      AVLInit ( &ptree, true, _PreprocCompMacroNames,
                _PreprocAllocNode, NULL );
      pks_GetNextSymbol ( psc );    /* skip the macro name */
      if ( psc->nextsymbol != PKS_SYMB_LPAREN ) {
        psc->error_code = PRERROR_INTERNAL;
        goto failure;            /* this should never happen */
      }
      pks_GetNextSymbol ( psc );    /* skip the parenthesis */
      if ( psc->nextsymbol == PKS_SYMB_RPAREN ) /* empty parameter list */
        goto get_tokens;
      for (;;) {
        if ( psc->nextsymbol != PKS_SYMB_IDENT ) {
          psc->error_code = PRERROR_WRONG_PARAMETER;
          goto failure;
        }
        pname = pkv_GetScratchMem ( sizeof(pks_macropname) +
                                    strlen(psc->nextname) );
        if ( !pname ) {
          psc->error_code = PRERROR_NO_MEMORY;
          goto failure;
        }
        pname->number = npar;
        strcpy ( pname->name, psc->nextname );
        AVLInsert ( &ptree, (void*)pname );
        if ( ptree.last != (AVLtnode*)pname ) {
          psc->error_code = PRERROR_PARAM_REDEF;
          goto failure;
        }
        mcri->nparams = ++npar;
        pks_GetNextSymbol ( psc );  /* skip the parameter name */
        if ( psc->nextsymbol == PKS_SYMB_COMMA )
          pks_GetNextSymbol ( psc );
        else if ( psc->nextsymbol == PKS_SYMB_RPAREN ) /* end of parameter list */
          break;
      }
      pname = pkv_GetScratchMem ( sizeof(pks_macropname) +
                                  psc->namebuflength );
      if ( !pname ) {
        psc->error_code = PRERROR_NO_MEMORY;
        goto failure;
      }
            /* allocate space for macro parameter macros */
      bufp += npar*sizeof(pks_macro*);
      if ( bufp >= PKS_MACRO_BUF_LENGTH ) {
        psc->error_code = PRERROR_BUFFER_OVERFLOW;
        goto failure;
      }
      else
        memset ( &mbuf[mcri->params], 0, npar*sizeof(pks_macro*) );
    }

get_tokens:
    mcri->tokens = bufp;
        /* now the text of the macro - any sequence of tokens */
    for (;;) {
      while ( psc->nextchar == ' ' || psc->nextchar == '\t' ||
              psc->nextchar == '\r' )
        pks_GetNextChar ( psc );
      if ( psc->nextchar == psc->bcomment1[0] ) {
        if ( _pks_SkipComment ( psc, psc->bcomment1, psc->ecomment1 ) )
          continue;
      }
      if ( psc->nextchar == psc->bcomment2[0] ) {
        if ( _pks_SkipComment ( psc, psc->bcomment2, psc->ecomment2 ) )
          continue;
      }
      switch ( psc->nextchar ) {
    case '\n':
        goto enter_the_macro;
    case EOF:   /* enter the macro */
        psc->nextsymbol = PKS_SYMB_EOF;
        goto enter_the_macro;
    case '\\':  /* backslash extends the macro to the next line */
        do {
          pks_GetNextChar ( psc );
        } while ( psc->nextchar != '\n' && psc->nextchar != EOF );
        if ( psc->nextchar == '\n' )
          pks_GetNextChar ( psc );
        break;
    case '#':  /* #const(constant-expression) is allowed here */
        pks_GetNextChar ( psc );
        if ( !isalpha ( psc->nextchar ) )
          goto wrong_const;
        pks_GetNextSymbol ( psc );
        if ( psc->nextsymbol != PKS_SYMB_IDENT )
          goto wrong_const;
        if ( strcmp ( psc->nextname, hashnames[HASH_CONST] ) )
          goto wrong_const;
        if ( !_pks_ProcessHashConst ( psc ) ) {
wrong_const:
          psc->nextsymbol = PKS_SYMB_ERROR;
          psc->error_code = PRERROR_WRONG_TOKEN;
          goto failure;
        }
        else
          goto add_next_token;
    default:    /* a token of the macro text */
        pks_GetNextSymbol ( psc );
        if ( psc->nextsymbol == PKS_SYMB_ERROR ) {
          psc->error_code = PRERROR_WRONG_TOKEN;
          goto failure;
        }
        if ( psc->nextsymbol == PKS_SYMB_IDENT && npar > 0 ) {
                    /* it may be a macro parameter, search the tree */
          strcpy ( pname->name, psc->nextname );
          pn = (pks_macropname*)AVLFind ( &ptree, pname );
          if ( pn ) {
            psc->nextsymbol = PKS_SYMB_MACROPAR;
            psc->nextinteger = pn->number;
          }
        }
add_next_token:
        if ( !_pks_AddTheMacroToken ( psc, mbuf, &bufp ) ) {
          psc->error_code = PRERROR_BUFFER_OVERFLOW;
          goto failure;
        }
        break;
      }
    }

enter_the_macro:
        /* now terminate the macro and enter it to the tree */
    end.symbol = PKS_SYMB_NONE;
    end.length = sizeof(pks_token_any);
    if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                           (char*)&end, end.length ) ) {
      psc->error_code = PRERROR_BUFFER_OVERFLOW;
      goto failure;
    }
    mcr = malloc ( bufp+1 );
    if ( !mcr ) {
      psc->error_code = PRERROR_NO_MEMORY;
      goto failure;
    }
    memcpy ( mcr, mbuf, bufp );
    ((char*)mcr)[bufp] = 0;
    AVLInsert ( &psc->macros, (void*)mcr );
    if ( psc->macros.last != (AVLtnode*)mcr ) {
      psc->error_code = PRERROR_MACRO_REDEF;
      free ( mcr );
      goto failure;
    }
    psc->bkslnl = false;
    pks_GetNextSymbol ( psc );  /* skip the last symbol of the macro */
    pkv_SetScratchMemTop ( sp );
#ifdef DEBUG
printf ( "#define %s(%d)\n", &((char*)mbuf)[sizeof(pks_macro)], mcri->nparams );
#endif
    return true;
  }
  else {
failure:
    psc->nextsymbol = PKS_SYMB_ERROR;
    psc->bkslnl = false;
    pkv_SetScratchMemTop ( sp );
    return false;
  }
} /*_pks_ProcessHashDefine*/

static boolean _pks_ProcessHashUndef ( pks_scanner *psc )
{
  void        *sp;
  char        *mcr;

  sp = pkv_GetScratchMemTop ();
  pks_GetNextSymbol ( psc );  /* skip #undef */
  if ( psc->nextsymbol == PKS_SYMB_IDENT ) {
    mcr = pkv_GetScratchMem ( sizeof(pks_macro)+strlen(psc->nextname)+1 );
    if ( !mcr ) {
      psc->error_code = PRERROR_NO_MEMORY;
      goto failure;
    }
    memset ( mcr, 0, sizeof(pks_macro) );
    strcpy ( &mcr[sizeof(pks_macro)], psc->nextname );
    AVLDelete ( &psc->macros, mcr );
    if ( psc->macros.error_code == AVLERROR_DELETE ) {
      psc->error_code = PRERROR_MACRO_UNDEF;
      goto failure;
    }
    else {
#ifdef DEBUG
printf ( "#undef %s\n", psc->nextname );
#endif
      pks_GetNextSymbol ( psc );  /* skip the deleted macro name */
      pkv_SetScratchMemTop ( sp );
      return true;
    }
  }
  else {
    psc->error_code = PRERROR_MACRO_UNDEF;
failure:
    psc->nextsymbol = PKS_SYMB_ERROR;
    pkv_SetScratchMemTop ( sp );
    return false;
  }
} /*_pks_ProcessHashUndef*/

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _pks_SkipUntilHashElseEndif ( pks_scanner *psc, short l0,
                                             short *hashn )
{
  short level;
  short symbol, hashname;

  symbol = psc->nextsymbol;
  level = l0;
  for (;;) {
    if ( symbol == PKS_SYMB_HASH && isalpha ( psc->nextchar ) ) {
      symbol = pks_GetNextSymbol ( psc );
      if ( symbol != PKS_SYMB_IDENT ) {
        psc->error_code = PRERROR_WRONG_TOKEN;
        goto failure;
      }
      else {
        *hashn = hashname =
            pks_BinSearchNameTable ( HASH_SYMB_LENGTH, 0, hashnames, psc->nextname );
        switch ( hashname ) {
      case HASH_IF:
          level ++;
          break;
      case HASH_ENDIF:
          if ( --level <= 0 )
            goto way_out;
          else
            break;
      case HASH_ELSEIF:
      case HASH_ELSE:
          if ( level == 0 )
            goto way_out;
          else
            break;
      case HASH_ERROR:
          psc->error_code = PRERROR_WRONG_TOKEN;
          goto failure;
      default:
          break;
        }
      }
    }
    else if ( symbol == PKS_SYMB_EOF ) {
      psc->error_code = PRERROR_HASHIF;
      goto failure;
    }
    else if ( symbol == PKS_SYMB_ERROR ) {
      psc->error_code = PRERROR_WRONG_TOKEN;
      goto failure;
    }
    symbol = pks_GetNextSymbol ( psc );
  }

way_out:
  symbol = pks_GetNextSymbol ( psc );
  return true;

failure:
  psc->nextsymbol = PKS_SYMB_ERROR;
  return false;
} /*_pks_SkipUntilHashElseEndif*/

static boolean _pks_ProcessHashDef ( pks_scanner *psc )
{
  void        *sp;
  char        *mbuf;

  sp = pkv_GetScratchMemTop ();
  pks_GetNextSymbol ( psc );  /* skip #def */

  if ( psc->nextsymbol != PKS_SYMB_IDENT ) {
    psc->error_code = PRERROR_IFSYNTAX;
    goto failure;
  }
  mbuf = pkv_GetScratchMem ( sizeof(pks_macro)+strlen(psc->nextname)+1 );
  if ( !mbuf ) {
    psc->error_code = PRERROR_NO_MEMORY;
    goto failure;
  }
  memset ( mbuf, 0, sizeof(pks_macro) );
  strcpy ( &mbuf[sizeof(pks_macro)], psc->nextname );
        /* do not skip the macro name */
  psc->nextsymbol = PKS_SYMB_INTEGER;
  psc->nextinteger = AVLFind ( &psc->macros, (void*)mbuf ) != NULL;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  pkv_SetScratchMemTop ( sp );
  psc->nextsymbol = PKS_SYMB_ERROR;
  return false;
} /*_pks_ProcessHashDef*/

static boolean _pks_ProcessHashIf ( pks_scanner *psc )
{
  pks_calcval x;
  boolean     satisfied;
  short       hashn;

  if ( psc->mstack ) { /* #if not allowed in a macro expansion */
    psc->error_code = PRERROR_IFSYNTAX;
    goto failure;
  }
  pks_GetNextSymbol ( psc );  /* skip #if */
  psc->ifsp ++;
  for (;;) {
    if ( psc->nextsymbol != PKS_SYMB_LPAREN )
      goto failure;
    pks_GetNextPSymbol ( psc );
    if ( !pkscalc_Expression ( psc, &x ) )
      goto failure;
    if ( psc->nextsymbol != PKS_SYMB_RPAREN )
      goto failure;
    satisfied = x.valtype == PKS_SYMB_INTEGER ? x.intval != 0 : x.floatval != 0.0;
    if ( satisfied ) {
          /* in this case the right parenthesis must be skipped using  */
          /* the preprocessor, as the next symbol may be obtained from */
          /* a macro, and there also may be a #define, which has to be */
          /* processed in order to enter the new macro to the tree     */
      pks_GetNextPSymbol ( psc );
      goto way_out;
    }
    else {
          /* any macros or other tokens here must be skipped without   */
          /* further ado; hence, here the scanner procedure is called  */
          /* directly, without calling the preprocessor                */
      pks_GetNextSymbol ( psc );
      if ( !_pks_SkipUntilHashElseEndif ( psc, 0, &hashn ) )
        goto failure;
      switch ( hashn ) {
  case HASH_ELSE:
        goto way_out;
  case HASH_ELSEIF:
        continue;
  case HASH_ENDIF:
        psc->ifsp --;
        goto way_out;
  default:
        goto failure;
      }
    }
  }
way_out:
  return true;

failure:
  psc->nextsymbol = PKS_SYMB_ERROR;
  return false;
} /*_pks_ProcessHashIf*/

static boolean _pks_ProcessHashElse ( pks_scanner *psc )
{
  short hashn;

  pks_GetNextSymbol ( psc );  /* skip #else */
  _pks_SkipUntilHashElseEndif ( psc, 1, &hashn );
  return psc->nextsymbol != PKS_SYMB_ERROR;
} /*_pks_ProcessHashElse*/

static boolean _pks_ProcessHashEndif ( pks_scanner *psc )
{
  if ( psc->ifsp > 0 ) {
    psc->ifsp --;
    pks_GetNextSymbol ( psc );  /* skip #endif */
    return true;
  }
  else {
    psc->nextsymbol = PKS_SYMB_ERROR;
    psc->error_code = PRERROR_HASHIF;
    return false;
  }
} /*_pks_ProcessHashEndif*/

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _pks_ProcessHashInclude ( pks_scanner *psc )
{
  pks_GetNextSymbol ( psc ); /* skip #include */
  if ( psc->nextsymbol != PKS_SYMB_STRING ) {
    psc->error_code = PRERROR_HASHINCLUDE;
    goto failure;
  }
#ifdef DEBUG
printf ( "#include %s\n", psc->nextname );
#endif
  if ( !pks_OpenInputFile ( psc, psc->nextname ) ) {
    psc->error_code = PRERROR_INPUT_FILE;
    goto failure;
  }
  pks_GetNextSymbol ( psc );
  return true;

failure:
  psc->nextsymbol = PKS_SYMB_ERROR;
  return false;
} /*_pks_ProcessHashInclude*/

/* ///////////////////////////////////////////////////////////////////////// */
static boolean _pks_ProcessHashSymbol ( pks_scanner *psc )
{
  int     hashname;

  pks_GetNextSymbol ( psc );  /* skip # */
  if ( psc->nextsymbol == PKS_SYMB_IDENT ) {
    hashname = pks_BinSearchNameTable ( HASH_SYMB_LENGTH, 0, hashnames,
                                        psc->nextname );
    switch ( hashname ) {
case HASH_CONST:
      return _pks_ProcessHashConst ( psc );

case HASH_DEF:
      return _pks_ProcessHashDef ( psc );

case HASH_DEFINE:
      return _pks_ProcessHashDefine ( psc );

case HASH_UNDEF:
      return _pks_ProcessHashUndef ( psc );

case HASH_IF:
      return _pks_ProcessHashIf ( psc );

case HASH_ELSEIF:
case HASH_ELSE:
      return _pks_ProcessHashElse ( psc );

case HASH_ENDIF:
      return _pks_ProcessHashEndif ( psc );

case HASH_INCLUDE:
      return _pks_ProcessHashInclude ( psc );

case HASH_ERROR:
case PKS_SYMB_ERROR:
default:
      psc->error_code = PRERROR_WRONG_TOKEN;
      psc->nextsymbol = PKS_SYMB_ERROR;
      return false;
    }
  }
  return true;
} /*_pks_ProcessHashSymbol*/

/* ///////////////////////////////////////////////////////////////////////// */
static int _pks_GetNextMacroSymbol ( pks_scanner *psc );

/* when macro expansion is begun, the nextsymbol is either the macro name,  */
/* or the right parenthesis closing the macro parameter list, the next call */
/* to pks_GetNextSymbol occurs after expanding any waiting macros           */
static boolean _pks_BeginMacroExpansion ( pks_scanner *psc, pks_macro *mcr )
{
  void          *sp;
  short         i;
  short         argpar, npar;
  char          *mbuf;
  int           bufp;
  pks_macro     *pmcr;  /* parameter macro */
  pks_macro     **mptab;
  pks_token_any end;
  char          paramid[100];

  sp = pkv_GetScratchMemTop ();
  if ( mcr->previous ) {   /* this macro is already being expanded, */
    psc->error_code = PRERROR_RECURSIVE_EXP;
    goto failure;         /* recursive expansion is an error */
  }
#ifdef DEBUG
printf ( "expanding %s\n", &((char*)mcr)[sizeof(pks_macro)] );
#endif
  if ( mcr->nparams > 0 ) {
    mbuf = pkv_GetScratchMem ( PKS_MACRO_BUF_LENGTH );
    if ( !mbuf ) {
      psc->error_code = PRERROR_NO_MEMORY;
      goto failure;
    }
    mptab = (pks_macro**)&((char*)mcr)[mcr->params];
    _pks_GetNextMacroSymbol ( psc );  /* skip the macro name */
          /* prepare macros for the parameters */
    if ( psc->nextsymbol != PKS_SYMB_LPAREN ) {
      psc->error_code = PRERROR_PARAM_MATCH;
      goto failure;
    }
    for ( i = 0; i < mcr->nparams; i++ ) {
      _pks_GetNextMacroSymbol ( psc );  /* skip the left parenthesis or a comma */
      pmcr = (pks_macro*)mbuf;
      memset ( pmcr, 0, sizeof(pks_macro) );
      bufp = sizeof(pks_macro);
      sprintf ( paramid, "%d of %s", i, &((char*)mcr)[sizeof(pks_macro)] );
      if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                             paramid, strlen(paramid)+1 ) ) {
        psc->error_code = PRERROR_BUFFER_OVERFLOW;
        goto failure;
      }
      pmcr->params = pmcr->tokens = bufp;
              /* get the tokens of the macro parameter */
      argpar = PKS_SYMB_NONE;  /* type and number of nested parentheses */
      npar = 0;
      for (;;) {
        if ( npar == 0 ) {
          switch ( psc->nextsymbol ) {
        case PKS_SYMB_LPAREN:
        case PKS_SYMB_LBRACKET:
        case PKS_SYMB_LBRACE:
            argpar = psc->nextsymbol + 1;
            npar = 1;
            goto next_param_token;
        case PKS_SYMB_RPAREN:
        case PKS_SYMB_COMMA:
            goto set_param_macro;
        case PKS_SYMB_EOF:
        case PKS_SYMB_ERROR:
            psc->error_code = PRERROR_WRONG_TOKEN;
            goto failure;
        default:
            goto next_param_token;
          }
        }
        else {
          switch ( psc->nextsymbol ) {
        case PKS_SYMB_LPAREN:
        case PKS_SYMB_LBRACKET:
        case PKS_SYMB_LBRACE:
            if ( psc->nextsymbol+1 == argpar )
              npar ++;
            break;
        case PKS_SYMB_RPAREN:
        case PKS_SYMB_RBRACKET:
        case PKS_SYMB_RBRACE:
            if ( psc->nextsymbol == argpar ) {
              npar --;
              if ( npar == 0 )
               argpar = PKS_SYMB_NONE;
            }
            break;
        case PKS_SYMB_EOF:
        case PKS_SYMB_ERROR:
            psc->error_code = PRERROR_WRONG_TOKEN;
            goto failure;
        default:
            break;
          }
next_param_token:
          _pks_AddTheMacroToken ( psc, mbuf, &bufp );
          _pks_GetNextMacroSymbol ( psc );
        }
      }
set_param_macro:
      if ( i == mcr->nparams-1 && psc->nextsymbol != PKS_SYMB_RPAREN ) {
        psc->error_code = PRERROR_PARAM_MATCH;
        goto failure;    /* too many parameters */
      }
      else if ( i < mcr->nparams-1 && psc->nextsymbol != PKS_SYMB_COMMA ) {
        psc->error_code = PRERROR_PARAM_MATCH;
        goto failure;    /* too little parameters */
      }
      end.symbol = PKS_SYMB_NONE;
      end.length = sizeof(pks_token_any);
      if ( !pkv_AppendData ( mbuf, PKS_MACRO_BUF_LENGTH, &bufp,
                             (char*)&end, end.length ) ) {
        psc->error_code = PRERROR_BUFFER_OVERFLOW;
        goto failure;
      }
      pmcr = malloc ( bufp );
      if ( !pmcr ) {
        psc->error_code = PRERROR_NO_MEMORY;
        goto failure;
      }
      memcpy ( pmcr, mbuf, bufp );
      mptab[i] = pmcr;
    }
  }
  mcr->nsymb = mcr->tokens;
  mcr->previous = (pks_macro*)psc->mstack;
  psc->mstack = (void*)mcr;
  pkv_SetScratchMemTop ( sp );
  return true;

failure:
  psc->nextsymbol = PKS_SYMB_ERROR;
  pkv_SetScratchMemTop ( sp );
  return false;
} /*_pks_BeginMacroExpansion*/

static void _pks_EndMacroExpansion ( pks_scanner *psc )
{
  pks_macro *mcr, **pmtab;
  int       i;

  mcr = (pks_macro*)psc->mstack;  /* pop the macro off the stack */
  psc->mstack = (void*)mcr->previous;
  mcr->previous = NULL;
#ifdef DEBUG
printf ( "end of %s\n", &((char*)mcr)[sizeof(pks_macro)] );
#endif
        /* destroy macro parameters, if there are any */
  if ( mcr->nparams > 0 ) {
    pmtab = (pks_macro**)&((char*)mcr)[mcr->params];
    for ( i = 0; i < mcr->nparams; i++ )
      if ( pmtab[i] ) {
        free ( pmtab[i] );
        pmtab[i] = NULL;
      }
  }
} /*_pks_EndMacroExpansion*/

static int _pks_GetNextMacroSymbol ( pks_scanner *psc )
{
  void      *sp;
  pks_macro *mcr, **pmtab;
  pks_token *token;
  int       symbol;
  char      *mcn;
  int       i;

  sp = pkv_GetScratchMemTop ();
again:
  mcr = (pks_macro*)psc->mstack;
  if ( !mcr )
    return pks_GetNextSymbol ( psc );
  pmtab = (pks_macro**)&((char*)mcr)[mcr->params];
  token = (pks_token*)&((char*)mcr)[mcr->nsymb];
  mcr->nsymb += token->any.length;
  switch ( token->any.symbol ) {
case PKS_SYMB_NONE:
    _pks_EndMacroExpansion ( psc );
    if ( psc->mstack )
      goto again;
    else {
      symbol = pks_GetNextSymbol ( psc );  /* skip the last macro name or parenthesis */
      if ( symbol == PKS_SYMB_IDENT )
        goto process_ident;
      else
        return psc->nextsymbol;
    }

case PKS_SYMB_IDENT:  /* this may be a next macro to expand */
process_ident:
    mcn = pkv_GetScratchMem ( sizeof(pks_macro)+strlen(token->name.name)+1 );
    if ( !mcn )
      return psc->nextsymbol = PKS_SYMB_ERROR;
    memset ( mcn, 0, sizeof(pks_macro) );
    if ( psc->mstack )
      strcpy ( &mcn[sizeof(pks_macro)], token->name.name );
    else
      strcpy ( &mcn[sizeof(pks_macro)], psc->nextname );
    if ( (mcr = (pks_macro*)AVLFind ( &psc->macros, (void*)mcn )) ) {
                      /* it is indeed */
      if ( _pks_BeginMacroExpansion ( psc, mcr ) )
        goto again;
      else {
        pkv_SetScratchMemTop ( sp );
        return psc->nextsymbol = PKS_SYMB_ERROR;
      }
    }
    else {
      pkv_SetScratchMemTop ( sp );
      if ( psc->mstack ) {
        psc->nextinteger = token->name.intvalue;
        psc->id_space = token->name.id_space;
        strcpy ( psc->nextname, token->name.name );
      }
      return psc->nextsymbol = PKS_SYMB_IDENT;
    }

case PKS_SYMB_MACROPAR:  /* begin expanding the parameter macro */
    i = token->integer.intvalue;
    if ( _pks_BeginMacroExpansion ( psc, pmtab[i] ) )
      goto again;
    else {
      pkv_SetScratchMemTop ( sp );
      return psc->nextsymbol = PKS_SYMB_ERROR;
    }

case PKS_SYMB_STRING:
    strcpy ( psc->nextname, token->name.name );
    return psc->nextsymbol = PKS_SYMB_STRING;

case PKS_SYMB_INTEGER:
case PKS_SYMB_CHAR:
    psc->nextfloat = psc->nextinteger = token->integer.intvalue;
    return psc->nextsymbol = token->any.symbol;

case PKS_SYMB_FLOAT:
    psc->nextfloat = token->real.floatvalue;
    return psc->nextsymbol = PKS_SYMB_FLOAT;

default:
    return psc->nextsymbol = token->any.symbol;
  }
} /*_pks_GetNextMacroSymbol*/

/* ///////////////////////////////////////////////////////////////////////// */
int pks_GetNextPSymbol ( pks_scanner *psc )
{
  void      *sp;
  int       symbol;
  pks_macro *mcr;
  char      *mcn;

  sp = pkv_GetScratchMemTop ();
  mcn = pkv_GetScratchMem ( sizeof(pks_macro)+psc->namebuflength+1 );
  if ( !mcn ) {
    psc->nextsymbol = PKS_SYMB_ERROR;
    goto way_out;
  }
  memset ( mcn, 0, sizeof(pks_macro) );
  for (;;) {
    if ( psc->mstack ) {  /* a macro is being expanded */
      symbol = _pks_GetNextMacroSymbol ( psc );
      if ( symbol == PKS_SYMB_HASH && !psc->mstack )
        goto hash;
      else
        goto way_out;
    }
    else {  /* get next symbol, find out if it is a token */
            /* or a special symbol for the preprocessor   */
      symbol = pks_GetNextSymbol ( psc );
hash:
      while ( symbol == PKS_SYMB_EOF ||
              (symbol == PKS_SYMB_HASH && isalpha ( psc->nextchar )) ) {
        if ( symbol == PKS_SYMB_HASH ) {
          if ( !_pks_ProcessHashSymbol ( psc ) )
            goto way_out;
        }
        if ( psc->nextsymbol == PKS_SYMB_EOF ) {  /* end of #include? */
#ifdef DEBUG
printf ( "end of %s\n", psc->infile->filename );
#endif
          pks_CloseInputFile ( psc );
          if ( psc->infile ) {
            symbol = pks_GetNextSymbol ( psc );
            goto hash;
          }
          else
            goto way_out;
        }
      }
      symbol = psc->nextsymbol;
      if ( symbol == PKS_SYMB_IDENT ) {
          /* if it is a macro name, begin its expansion, */
        strcpy ( &mcn[sizeof(pks_macro)], psc->nextname );
        if ( (mcr = (pks_macro*)AVLFind ( &psc->macros, (void*)mcn )) ) {
          if ( _pks_BeginMacroExpansion ( psc, mcr ) ) {
            symbol = _pks_GetNextMacroSymbol ( psc );
            if ( psc->nextsymbol == PKS_SYMB_HASH && !psc->mstack )
              goto hash;
            else
              goto way_out;
          }
          else
            continue;
        }
        else  /* else return the identifier */
          break;
      }
      else break;
    }
  }
way_out:
  pkv_SetScratchMemTop ( sp );
  return psc->nextsymbol;
} /*pks_GetNextPSymbol*/

